/* $Id: me.h,v 1.74 2003/08/10 10:03:35 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.74 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *
 *  Initially written by: Michael Elkins <elkins@aero.org>, 1995
 *****************************************************************************/

#include "melib.h"

/* args.c */

/* Returns array of files requested */
extern char ** parse_arguments P_((int argc, char *argv[], char ***to_whom));
extern char * wanted_charset;
extern int wanted_switchmode;

extern char * wanted_title_string;
extern char * wanted_icon_string;
extern char * with_title;
extern char * wanted_mailbox_title;

extern struct Attachments {
    mime_t *attach_files;    /* prepare attachments for sending */
    int attachment_count;
} attach_files;

CONST extern struct Attachments NULL_Attachments;
extern void add_Attachments  P_((struct Attachments *A,mime_t *x));
extern void free_Attachments P_((struct Attachments *A));

/* init.c */
#if ANSI_C
struct MailboxView;
#endif

void initialize P_((char **name_vector, struct MailboxView **main_mailbox));

/* file_util.c */

#if ANSI_C
struct current_storage;
#endif

int check_mailfile_size P_((struct folder_info *mfile));
int append P_((FILE *fd, char *filename, char *prefix_str));

/* find_alias.c */

extern int find_alias P_((char *word, int alias_type));

/* showmsg_c.c */

extern int showmsg_main_cmd P_((     struct MailboxView *mailbox,
				     int cmd, int *screen_mangled));
extern void put_cmd_name P_((char *command, int will_mangle));
extern void put_border P_((void));
extern void build_bottom P_((void));

typedef int showmsg_cmd P_((     struct MailboxView *mailbox,
				 int cmd, int *screen_mangled));

extern int process_showmsg_cmd P_((int command,      
				   struct MailboxView *mailbox,
				   showmsg_cmd cmd));
/* elm.c */

/* Return key if unknown */
extern int motion P_((int ch, struct MailboxView *mailbox,
		      int *redraw, int *nufoot));
extern void check_range P_((struct MailboxView *mailbox));
extern void forget_passphrase P_((void) );

extern int mailbox_command P_((struct MailboxView *mailbox, int ch, 
			       int *redraw, int *nufoot,
			       showmsg_cmd cmdX
			       ));
extern void new_mail_check P_((struct MailboxView *mailbox,
			       int *redraw));


/* help.c */

extern int display_helpfile P_((int section));
extern int help P_((int pager_help));

/* in_utils.c */

extern int want_to P_((char *question, int dflt, int where, 
		       int clear_and_center));
extern int read_number      P_((int ch, char *item, int current,
				int *redraw));

#define OE_APPEND_CURRENT   (1<<0)
#define OE_PASSWD           (1<<1)
#define OE_REDRAW_MARK      (1<<2)
#define OE_ADD_COMMA        (1<<3)
#define OE_ALT_SOLIDUS      (1<<4)
#define OE_ALLOW_MIMEENC    (1<<5)
#define OE_EDITOR_ESCAPE    (1<<6)
#define OE_TABACTION        (1<<7)

extern int optionally_enter2 P_((struct string **buffer, 
				 int x, int y, int flags, 
				 const char * format, const char *msg, ...));
extern int optionally_enter P_((char *, int, int, int, int));
extern int pattern_enter    P_((struct string **string, 
				struct string **alt_string, 
				int x, int y, 
				struct string *alternate_prompt, 
				struct string *prompt,
				int *redraw));
extern int GetPrompt        P_((void));


#if ANSI_C
struct enter_info;
#endif

enum enter_mode { em_redraw, em_enter, em_prev, em_next, em_wrap,
		  em_redraw_initial, em_editor_escape, em_tabaction,
		  em_bs_prev, em_left, em_right};
typedef struct string **give_buffer_r P_((struct enter_info *I,
					  enum enter_mode em));


struct enter_info {
    int counter;
    struct string **pvector;
    int px;
    int py;
    give_buffer_r *give_buffer;
    int flags;
    int ch_count;
    struct builtin_edit  *builtin;
    folder_dir           dir_p;
    struct mailer_info  *mailer_info;
};

extern int enter_helper P_((struct enter_info *info));

/* out_utils.c */

extern void error_wait P_((void));   /* call before exiting */
extern void sleep_message P_((void));
extern int  error P_((CONST char *s));
extern void out_util_setup P_((void));
extern void show_last_error P_((void));
extern void clear_error P_((void));
extern void set_error P_((char *s));
extern void lower_prompt P_((char *s));
extern void prompt P_((char *s));
extern void set_central_message P_((const char *format, 
				    const char *line, ...));
extern void display_central_message P_((void));
extern void clear_central_message P_((void));
extern void print_center P_((int line, struct string * buffer));
extern void print_format_center P_((int lineY,const char *format, 
				    const char *line, ...));

extern void Raw_OFF P_((int exiting));  /* used on init.c for exit */

/* Used by newmbox.c */
extern int need_refresh_low_line P_((void));


/* pattern.c */

extern int from_matches P_((struct header_rec *mss, struct string *pat));
extern int to_matches P_((struct header_rec *mss, struct string *pat));
extern int cc_matches P_((struct header_rec *mss, struct string *pat));
extern int subject_matches P_((struct header_rec *mss, struct string *pat));
extern int match_in_message P_((struct string *pat,
				struct MailboxView *mailbox));
extern int pattern_match P_((struct MailboxView *mailbox, int *redraw));
extern int meta_match P_((int function, struct MailboxView *mailbox));


/* addr_util.c */

extern int build_address_l P_((struct expanded_address *expanded,
			       struct mailer_info      *mailer_info));
extern int argv_to_expanded P_((struct expanded_address *result,
				char *argv[],
				struct mailer_info      *mailer_info));
extern void addr_to_expanded P_((struct expanded_address *result,
				 struct addr_item *addrs,
				 struct mailer_info      *mailer_info));
extern void expanded_to_edit_buffer P_((struct string ** buffer,   
					struct expanded_address expanded
					));
extern void update_expanded_from_edit_buffer P_((struct expanded_address 
						 *expanded,
						 CONST struct string *buffer,
						 struct mailer_info     
						 *mailer_info
						 ));


extern int check_8bit_addr P_((struct addr_item * addr));

extern int aliases_to_expanded P_((struct expanded_address *x));

/* alias.c */

#if ANSI_C
struct current_storage;
#endif

extern int alias_message_count;    
extern int alias_current;		/* current alias number  */
extern int alias_selected;

extern void alias                   P_((struct MailboxView *mailbox));
extern void main_state P_((void));
extern void open_alias_files P_((int are_in_aliases));
extern int delete_from_alias_text P_((char **name, int num_to_delete));
extern void install_aliases P_((void));

/* a_edit.c */

int edit_aliases_text P_((void));

/* a_sort.c */

void sort_aliases P_((int entries, int visible, int are_in_aliases));


/* a_screen.c */

extern struct string *build_alias_line P_((struct alias_rec *entry,
					   int message_number, int highlight));

extern void alias_screen P_((int modified));
extern void alias_title P_((int modified));
extern void show_alias_menu P_((void));

extern char *show_alias_status P_((struct alias_rec *entry));
extern void set_folder_window_title P_((struct folder_info *folder,
					char *title_s, char *icon_s));

/* a_quit.c */

extern void exit_alias P_((void));
extern int delete_aliases P_((int newaliases, int prompt));

/* exitprog.c */

extern int exit_prog P_((struct MailboxView **mailbox));

/* file.c */

extern int expand_filename P_((char *, int, int));
extern void init_helpmsg P_(( char *, char *, int, int ));
extern int save P_((int *redraw, int silently, int delete, int text_only,
		    struct MailboxView *mailbox));

/* fileutil.c */

extern FILE *open_end_update P_((char *name));

#ifdef USE_PGP
/* pgp.c */
extern int pgp_menu			P_((char *filename,
					    struct mailing_headers *headers));
extern int pgp_extract_public_key	P_((struct header_rec  *hdr,
					    FILE *infile));
extern int pgp_mail_public_key          P_((struct MailboxView *mailbox));
#endif

/* read_rc.c */

extern void directory_check     P_((void));



extern int metapager		P_((FILE *fp, 
				    struct header_rec *hdr, 
				    int do_headers,
				    int current,
				    int message_count));
extern int builtinplusplus	P_((struct stringbuffer * bout));

/* mime.c */

extern int have_metamail         P_((void));
extern void clear_mime_send_info P_((mime_send_t *mime_info,
				     struct mailer_info  *mailer_info));
extern void free_mime_send_info P_((mime_send_t *mime_info));
extern void add_paramater_1 P_((char **ptr, char *name, 
				char *value, int quoted));

/* Returns 1 if OK */
extern int convert_text P_((FILE *source, FILE *dest, 
			    mime_send_t *mime_info,
			    charset_t from_charset,
			    charset_t to_charset,
			    int do_pgp,
			    struct Attachments *attachments,
			    struct mailer_info  *mailer_info));

extern void add_parameter_1 P_((char **ptr, char *name, char *value,  
				int quoted));


/* strings.c */


extern char **argv_from_to      P_((char *to));
extern void split_word P_((
			   char *buffer, char *first, char *rest));
extern void Centerline P_((int line,
			   char *string));

extern char *argv_zero P_((const char *string));


/* string2.c */

extern void remove_possible_trailing_spaces P_((char *string));

/* NOTE: Must be 'int ch', no 'char ch' so that it matches to default promotion
         of arguments
*/
extern int occurances_of P_((int ch, char *string));

/* fileio.c */
extern int copy_message P_((struct folder_info *, struct header_rec *,
			     char *, FILE *, int, charset_t));
extern int copy_message_f P_((FILE *infile, struct header_rec *current_header,
			      char *prefix, FILE *dest_file,
			      int cm_options, charset_t file_set));

extern int copy_message_d P_((struct folder_info *infolder,
			      struct header_rec *current_header,
			      char * prefix,
			      struct folder_browser *dir, 
			      WRITE_STATE dest, int cm_options,
			      charset_t file_set));
extern int copy_message_df P_((FILE *infile,
			       struct header_rec *current_header,
			       char * prefix,
			       struct folder_browser *dir,
			       WRITE_STATE dest,
			       int cm_options,
			       charset_t file_set));

extern int copy_message_2 P_((FILE *infile, 
			      struct header_rec *current_header,
			      char *prefix, 
			      out_state_t *dest_file, 
			      int cm_options,
			      int env_flags));

extern int copy_body P_((FILE *infile, 
			 struct header_rec *current_header,
			 char *prefix, 
			 out_state_t *dest_file, 
			 int cm_options));

extern int copy_plain P_((char *,out_state_t *,int, struct header_rec *, 
			   FILE *));
extern int copy_mime P_((char *,out_state_t *,int, struct header_rec *, 
			  FILE *));

typedef int copy_decoder P_((char *,out_state_t *,int, struct header_rec *,
			     FILE *));
typedef copy_decoder *copy_decoder_t;
extern copy_decoder_t select_copy_decoder P_((struct header_rec *));

/* newmbox.c */

extern int open_folder_lock P_((int direction,struct folder_info *folder));
extern int reopen_folder_lock_sessionlock P_((struct folder_info *folder));
extern int is_pre_mime_content_type P_((mime_t *,char *));
extern int same_file P_((char *,char *)); 

extern int parse_header_routine P_((struct folder_info *folder,
				    READ_STATE read_state_ptr,
				    struct header_rec *current_header,
				    header_list_ptr parsed_headers 
				    ));
extern int parse_body_routine P_((struct folder_info *folder,
				  READ_STATE read_state_ptr,
				  struct header_rec *current_header,
				  header_list_ptr parsed_headers,
				  struct counter_data *counter
				  ));

#if ANSI_C
struct current_storage;
#endif
extern void newmbox_1 P_((struct folder_info *new_folder,
			 int adds_only, 
			 struct MailboxView **mailbox,
			 int append));
extern int read_new_mails P_((struct current_storage *storage));
extern void resort_mailbox P_((struct MailboxView *mailbox,
			       int add_new_only
			       ));

/* file_util.c */

extern long fsize P_((FILE *));
extern long bytes P_((char *));
extern int copy_to_folder P_((FILE * from, struct folder_info *to));
extern int copy1 P_((FILE *from, char *to, int isspool));

/* mailmsg1.c */


extern int a_sendmsg P_((int edit_message, int form_letter,
			 struct MailboxView *mailbox /* save_copy needs */ ));
extern int send_msg_argv P_((char *argv[],
			     char *given_subject,int options, int form,
			     struct MailboxView *mailbox  /* null ?? */
));

extern int send_msg_l P_((struct header_rec * current_header,
			  struct addr_item *given_to, 
			  struct addr_item *given_cc,
			  char *given_subject,
			  int   options, int form_letter,
			  FILE *infile,
			  struct MailboxView *mailbox /* for save_copy */
			  ));
extern void display_to P_((struct expanded_address address));
extern int get_to P_((struct expanded_address *to,
		      struct mailer_info  *mailer_info));

/* mailmsg2.c */

#if ANSI_C
struct run_state;           /* Needed for prototype */
#endif


extern int mail P_((struct header_rec *current_header, 
		    int options, int form, 
		    struct mailing_headers *headers,
		    struct mailer_info  *mailer_info,
		    FILE *infile,
		    struct MailboxView *mailbox));
extern int mail_form P_((struct header_rec *current_header, 
			 struct addr_item *address, 
			 char *subj,
			 FILE *infile,
			 struct MailboxView *mailbox));

/* id_phrase.c */


/* froms.c */

extern int mail_filled_in_form P_((struct header_rec * current_header,
				   struct addr_item *address,
				   char *subject,      
				   FILE *infile,
				   struct MailboxView *
				   mailbox /* for save_copy */ ));

/* aliaslib.c */

struct addr_item *get_alias_address_l P_((
					  char *name, /* name to expand as an alias */
					  int mailing, /* TRUE to fully expand group names & recursive aliases */
					  int *too_longp, /* error code if expansion overflows buffer             */
					  struct mailer_info  *mailer_info
					  ));


/* syscall.c */

extern int system_call    P_((char *, VOLATILE int,      
			      struct MailboxView *mailbox));
#ifdef ALLOW_SUBSHELL
extern int subshell       P_((struct MailboxView *mailbox));
#endif
extern int do_pipe        P_((struct MailboxView *mailbox));
extern int have_printout  P_((void));
extern int print_msg      P_((int pause_on_scroll,
			      struct MailboxView *mailbox));
extern int print_text     P_((int pause_on_scroll,      
			      struct MailboxView *mailbox));

extern int create_folder_state_file P_((struct MailboxView *mailbox));
extern int remove_folder_state_file P_((void));

/* lib/errno.c */

extern char *error_description P_((int));

/* savecopy.c */

struct copy_file {
    struct string         * copy_file;
    struct folder_browser * dir;
};


extern void zero_copy_file P_((struct copy_file *cf));
extern void clear_copy_file P_((struct copy_file *cf));
extern int name_copy_file P_((struct copy_file *cf));

extern int save_copy P_((struct mailing_headers * headers,
			 struct copy_file *cf,
			 int form,
			 mime_send_t *mime_info,
			 FILE * conv_file,
			 struct MailboxView *mailbox));

extern int append_copy_to_file P_((struct mailing_headers * headers,
				   struct folder_browser *cf,
				   int form,
				   mime_send_t *mime_info,
				   FILE * conv_file,
				   WRITE_STATE dest));

extern void write_header_info P_((out_state_t *mailer,
				  struct mailing_headers * headers,
				  int   form, int copy,
				  mime_send_t *mime_info));

extern int copy_message_across P_((mime_send_t *mime_info,
				   out_state_t *dest, int copy,
				   FILE *conv_file));

/* mime.c */

extern int check_for_multipart	P_((FILE *, mime_send_t *,
				    struct mailer_info  *));
extern int needs_encoding       P_((FILE *));

/* mime_encode.c */

extern int attach_message P_((mime_t *part, out_state_t *mailer, 
			      mime_send_t *mime_info, 
			      struct mime_send_part * X));
extern void base64_encode	    P_((FILE *, out_state_t *, 
					int, mime_send_t *));
extern void line_quoted_printable_encode P_((char *input, 
					     out_state_t *mailer, int len, 
					     int istext,
					     mime_send_t *mime_info));
extern void quoted_printable_encode  P_((FILE *srcfp, out_state_t *mailer, 
					 int istext,
					 mime_send_t *mime_info));
extern char *mime_generate_boundary	P_((char *, int));
extern void add_parameter		P_((char *,char *,char *,int, int));
extern void add_parameter_t             P_((mime_t *, char *, char *, int));

extern int update_encoding		P_((int *,int));
extern void write_encoded P_((FILE *srcfp, out_state_t *mailer, int encoding, 
			      int is_text, mime_send_t *mime_info));

extern void mime_write_part_headers P_((out_state_t *fp,
					mime_send_t *ptr,
					struct mime_send_part * part));
extern void mime_write_top_headers P_((out_state_t *fp,
				       mime_send_t *ptr));
extern void rfc1522_encode_text         P_((char *,int,CONST char *,int));

/* attach_menu.c */

extern void attach_menu		P_((FILE *F,
				    mime_t *T,
				    struct Attachments *A,
				    charset_t defcharset,
				    struct mailer_info  *mailer_info,
				    struct header_rec *hdr));
extern int Attach_it             P_((char *));
extern int Check_attachments     P_((void));

/* returnadd.c */
extern void kludge                      P_((char *buffer, int size));

/* lock.c */

extern int Grab_the_file P_((int flock_fd));
extern int Release_the_file P_((int flock_fd));
extern int unlock P_((int interrupt, struct folder_info *folder));
extern int lock P_((int direction,struct folder_info *folder));
extern int flush_mailfile P_((struct folder_info *folder));


/* utils.c */

extern void emergency_exit P_((int interrupt));
extern void leave P_((int interrupt,      struct MailboxView **mailbox));
extern void create_new_folders P_((void));
extern void create_new_elmdir P_((void));
extern void rm_temps_exit P_((     struct MailboxView **mailbox));
extern void silently_exit P_((     struct MailboxView **mailbox));
extern void leave_locked P_((     struct MailboxView **mailbox));
extern int get_page P_((int msg_pointer,   struct MailboxView *mailbox));

/* date.c */

extern void days_ahead P_((int days, char *buffer, int size));
extern char * elm_date_str P_((char *buf, time_t seconds, int size));
extern void make_menu_date P_((struct header_rec *entry));

/* expires.h */

extern void process_expiration_date P_((char *date,
					int  *message_status));


/* reply.c */

extern void get_return_name P_((char *address, char *name,
				int   trans_to_lowercase,
				int size));
extern int forward P_((struct header_rec *hdr,
		       FILE *infile,
		       struct MailboxView *mailbox));
extern int reply_to_everyone P_((struct header_rec *hdr,
				 FILE *infile,
				 struct MailboxView *mailbox 
				 /* for save_copy */));
extern int reply P_((struct header_rec *hdr,
		     FILE *infile,
		     struct MailboxView *mailbox /* for save_copy */));

/* leavembox.c */

int leave_mbox P_((int resyncing,  int quitting, int prompt,
		   struct folder_info **new_folder,
		   struct MailboxView *mailbox));
		
/* mkhdrs.c */

extern void generate_reply_to P_((struct header_rec * current_header,
				  struct mailing_headers *headers));

/* options.c */

extern int options P_((struct MailboxView *mailbox));

/* hdrconfig.c */

extern void show_presend_headers P_((struct mailing_headers *headers,
				     charset_t hdr_charset));

/* returns 0 if command not found
           1 if found 
	   REDRAW_MARK is redraw required 
*/
extern int presend_action P_((struct mailing_headers *headers,
			      struct mailer_info  *mailer_info, int c,
			      charset_t hdr_charset));
extern void edit_headers P_((struct mailing_headers *headers,
			     struct mailer_info  *mailer_info,
			     charset_t hdr_charset));

extern struct string *hdr_to_expval P_((struct expanded_address addrs));
extern void hdr_to_buffer P_((struct expanded_address addrs, 
			      struct string **ptr));
extern int buffer_to_header P_((struct expanded_address *addrs,
				struct string **ptr,
				int free_only,     
				struct mailer_info *mailer_info));

/* remail.c */

extern int remail P_((struct header_rec *hdr,
		      FILE *F));
#ifdef MMDF
extern void do_mmdf_addresses P_((FILE *dest_file,
				  struct mailing_headers * headers));
#endif

/* editmsg.c */

extern int edit_the_message P_((char *filename,
				int  already_has_text,
				struct mailing_headers * headers,
				char *editor,
				charset_t  file_set,
				struct mailer_info  *mailer_info,
				struct MailboxView *mailbox /* needed for
							       system_call() */
				));


/* delete.c */

extern void delete_msg   P_((int real_del, int update_screen,
			     struct MailboxView *mailbox));
extern void undelete_msg P_((int update_screen,
			     struct MailboxView *mailbox));
extern void show_msg_status P_((int msg,
				struct MailboxView *mailbox));
extern void show_msg_tag    P_((int msg,
				struct MailboxView *mailbox));
extern void show_new_status P_((int msg,
				struct MailboxView *mailbox));
extern int DeleteTagged     P_((struct MailboxView *mailbox));
extern int tag_message P_((int update_screen,
			   struct MailboxView *mailbox));
extern void unread_msg P_((int update_screen, 
			   struct MailboxView *mailbox));

/* encode.c */

extern void getkey P_((int send));
extern void get_key_no_prompt P_((void));
extern void encode P_((char *line));
extern void makekey P_((char *rkey));
extern void setup P_((void));

/* forms.c */

extern int format_form P_((char *filename));
extern int check_form_file P_((char *filename));

/* quit.c */

extern void quit P_((int prompt, struct MailboxView **mailbox));
extern int resync P_((struct MailboxView **mailbox));
extern int change_file P_((struct MailboxView **mailbox));


/* save_opts.c */

extern void save_options P_((void));
extern void save_user_options P_((FILE *elminfo_fd, FILE *newelmrc));

char *str_opt_nam P_((char *s, int f));


/* screen.c */

typedef void menu_func P_((void));
extern void showscreen P_((struct MailboxView *mailbox,
			   menu_func *menu));
extern void update_title P_((struct MailboxView *mailbox));
extern void show_menu P_((void));
extern void show_current P_((struct MailboxView *mailbox));
extern int show_headers P_((struct MailboxView *mailbox));
extern int on_page P_((int message,      
		       struct MailboxView *mailbox));

/* showmsg.c */

extern int show_msg P_((struct header_rec *current_header,
			FILE *infile, int current, int message_count));

extern int need_meta P_((struct header_rec *hdr));

/* sort.c */

extern void sort_mailbox P_((int entries, int visible, 
			     struct MailboxView *mailbox));

/* edit.c */

extern void edit_mailbox P_((struct MailboxView *mailbox));
extern int edit_a_file P_((const char *editfile));

/* limit.c */

extern int compute_visible P_((int message,
			       struct MailboxView *mailbox));
extern int next_message    P_((int iindex, int skipdel,
			       struct MailboxView *mailbox));
extern int prev_message P_((int iindex, int skipdel,
			    struct MailboxView *mailbox));

extern int visible_to_index P_((int message,
				struct MailboxView *mailbox));
extern int limit P_((struct MailboxView *mailbox));
extern void ClearSelection P_((struct MailboxView *mailbox));



/* calendar.c */

extern void scan_calendar P_((struct MailboxView *mailbox));

/* browser.c */

extern struct folder_info * folder_browser P_((struct folder_browser *p,
					       struct string **buffer,
					       int *redraw,
					       CONST char *format, 
					       CONST char *msg, ...));


enum word_sel { word_change, word_save, word_copy, word_save_copy, 
		word_read };

extern void gen_browser P_((struct folder_browser *p,
			    struct string **buffer,
			    int *redraw,
			    enum word_sel w,
			    struct string * prev_fold,
			    CONST char *format, 
			    CONST char *msg, ...));

extern int browser_expand P_((struct folder_browser * XXX,
			      struct string **name,
			      struct string ** prev_folder));

extern int file_browser P_((struct folder_browser *p,
			    struct string **buffer,
			    int * redraw, enum word_sel w,
			    const char *format, const char *msg, ...));

/* metapager.c */

extern void PressAnyKeyToContinue P_((void));

/* signals.c */

extern char * Stopped_Text;  /* Initialized by init.c */
extern char * Back_Text;     /* Initialized by init.c */
extern char * Segv_Text;     /* Initialized by init.c */
extern char * Bus_Text;      /* Initialized by init.c */
extern char * Fpe_Text;      /* Initialized by init.c */
extern char * Ill_Text;      /* Initialized by init.c */

extern void init_signals P_((void));
extern void init_signals1 P_((void));

/* src/messages/ */

struct current_storage {
    struct folder_info * current_folder;   
    int                  message_count;	/* max message number      */

    int  max_headers;	/* number of headers currently allocated */

    struct header_rec **headers; /* array of header structure pointers */

};

extern void free_storage P_((struct current_storage **storage));
extern struct current_storage *new_storage P_((struct folder_info *new_folder));
extern void realloc_headers P_((struct current_storage *storage, 
				int current_count));
extern void alloc_header P_((struct current_storage *storage, 
			     int current_count));

extern void header_zero P_((struct header_rec *));
extern void header_clear P_((struct header_rec *));  /* Free's fields */
extern void header_free P_((struct header_rec **h)); /* Free's record */


extern void header_parse_helper  P_((struct header_rec *current_header,
				     header_list_ptr parsed_headers));

extern void start_body_helper P_((struct header_rec *current_header,
				  long content_start,
				  header_list_ptr parsed_headers));

extern struct MailboxView    * current_mailbox;


/* Joins storage to new view */
extern struct MailboxView * malloc_mailbox_view P_((struct current_storage *
						    storage));

extern struct MailboxView * digest_to_mailbox_view P_((mime_t *list,
						       time_t received_time,
						       char * env_from,
						       FILE *F,
						       charset_t defcharset));
						       
extern struct MailboxView * partial_to_mailbox_view P_((struct MailboxView *
							parent));

extern void free_mailbox P_((struct MailboxView **mailbox));

/* Return 1 if redraw required */
extern int update_view P_((struct MailboxView *mailbox));
extern struct folder_info * get_main_folder P_((struct MailboxView *
						     mailbox));

/* Can be called from signal handler */
extern struct current_storage * get_storage P_((struct MailboxView *mailbox,
						int i));

/* Can be called from signal handler */
extern int get_storage_count P_((struct MailboxView *mailbox));




extern void add_storage P_((struct MailboxView *mailbox,
			    struct current_storage *storage));
extern void add_digest P_((struct MailboxView *mailbox,
			   mime_t *list,
			   time_t received_time,
			   char * env_from,
			   FILE *F,
			   charset_t defcharset
			   ));
			   

extern struct header_rec * give_header P_((struct MailboxView *mailbox,
					   int index));

extern void give_index_number P_((struct MailboxView *mailbox,
				  int index, struct folder_view *res));

/*  current == index+1     value is used only by caller */
extern int get_current P_((struct MailboxView *mailbox));
extern void set_current P_((struct MailboxView *mailbox, int cur));

extern int get_selected P_((struct MailboxView *mailbox));
extern void set_selected P_((struct MailboxView *mailbox, int sel));


extern int get_message_count P_((struct MailboxView *mailbox));

#if ANSI_C
struct sort_data;
#endif

extern struct header_rec * give_header_s P_((struct sort_data *s));
extern void give_index_number_s P_((struct sort_data *s,
				    struct folder_view *res));

typedef int hdr_compare_func P_((struct sort_data *s1,
				 struct sort_data *s2));

extern void sort_mailbox_view P_((struct MailboxView *mailbox,
				  hdr_compare_func   *func));

/* Return 1 if succeed, 0 is fails --
   seeks to {header}->offset
*/
extern int give_message_data P_((struct MailboxView *mailbox,
				 int index,
				 struct header_rec **ret_header,
				 FILE              **ret_F,
				 struct counter_data *counter,
				 parse_mime_callback *parse_mime));

extern int give_message_folder P_((struct MailboxView *mailbox,
				   int index,
				   struct header_rec **ret_header,
				   struct folder_info **ret_folder,
				   struct counter_data *counter,
				   parse_mime_callback *parse_mime));


extern void write_mailbox_info P_((FILE *fp, struct MailboxView *mailbox));

/* caller must string_free() result -- calculated if not set */
extern struct string * mailbox_title P_((struct MailboxView *mailbox));

/* v must not fred -- shared/stored to structure
   -- set_mailbox_title will free_string()
   it later */
extern void set_mailbox_title P_((struct MailboxView *mailbox,
				  struct string *v));

/* digest.c */

extern void ViewDigest P_((struct MailboxView *mailbox));

/* partial.c */

extern void ViewPartial P_((struct MailboxView *parent_mailbox));

extern int have_partial P_((struct MailboxView *parent_mailbox));


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
