static char rcsid[] = "@(#)$Id: charset.c,v 1.71 2003/05/17 16:13:40 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.71 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *
 *  Partially based on mime_parse.c, which is initially 
 *     written by: Michael Elkins <elkins@aero.org>, 1995
 *****************************************************************************/

#include "headers.h"
#include "s_me.h"
#include "cs_imp.h"
#ifdef USE_DLOPEN
#include "shared_imp.h"
#endif

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

DEBUG_VAR(Debug,__FILE__,"charset");

struct  charset_type * cs_first_type = &cs_unknown;

/* ---------------------------------------------------------------------- */

static struct extra_sets {
    struct charcode_info set;
    struct extra_sets    * next;
} * extra_sets = NULL, * extra_sets_tail = NULL;
static int extra_sets_count = 0;

union compare_info {
    CONST char *name;
    struct charcode_info * part;
};

typedef  int compare_set P_((struct charcode_info *set, 
			     union compare_info data));

static struct charcode_info * find_set P_(( compare_set *func,
					    union compare_info data ));

static struct charcode_info * find_set(func,data)
     compare_set *func;
     union compare_info data;
{
    int i;
    struct extra_sets *walk;

#define TEST(ITEM) \
    if (0 != (ITEM.flags & SET_valid) && func(&(ITEM),data)) { \
      DPRINT(Debug,9,(&Debug,"find_set: found=%p (%s)\n",&(ITEM),\
		       ITEM.MIME_name ? ITEM.MIME_name : "<no MIME name>")); \
    return &(ITEM); }

#define SCAN(SET) \
    for (i = 0; SET[i].charset_type; i++) {\
	DPRINT(Debug,72,(&Debug,"find_set: (scan#%d) %p -- flags=%d\n",\
			 i,&(SET[i]),SET[i].flags)); \
	TEST(SET[i]); } 


    DPRINT(Debug,70,(&Debug, 
		     "find_set: precompiled_sets\n"));
    SCAN(precompiled_sets);
    
    DPRINT(Debug,70,(&Debug, "find_set: extra_sets\n"));
    for (walk = extra_sets; walk; walk = walk -> next) {
	DPRINT(Debug,72,(&Debug, "find_set: (walk) %p -- flags=%d\n",
		    &(walk->set),walk->set.flags)); 
	TEST(walk->set);
    }

    DPRINT(Debug,40,(&Debug, "find_set: NOT FOUND\n"));
    return NULL;
}	
#undef SCAN
#undef SET

/*
  Assigned MIB enum Numbers
  -------------------------
  0-2		Reserved
  3-999		Set By Standards Organizations
  1000-1999	Unicode / 10646
  2000-2999	Vendor


  Divide array to 3 classes so memory usage is better ... 

*/

#define MIBenum_class_divide 1000

static struct MIBenum_class {
    struct charcode_info ** MIBenum_array;
    int                     MIBenum_array_len;
} MIBenum_CLASS[] = {
    { NULL, 0 }                /*    0 -  999 */,
    { NULL, 0 }                /* 1000 - 1999 */,
    { NULL, 0 }                /* 2000 - 2999 */,
}; 

static int array_initialized = 0;

static int add_MIBenum P_((struct charcode_info *ptr));
static int add_MIBenum(ptr)
     struct charcode_info *ptr;
{
    CONST int MIBenum = ptr->MIBenum;
    int MIBenum_class;
    int val;

    if (MIBenum <= 0)
	return 0;
    
    MIBenum_class = MIBenum / MIBenum_class_divide;
    val           = MIBenum % MIBenum_class_divide;

    if (MIBenum_class >= sizeof (MIBenum_CLASS) / sizeof (MIBenum_CLASS[0]))
	return 0;

    if (MIBenum_CLASS[MIBenum_class].MIBenum_array_len <= val) {
	int X = val+1;
	int i;

	if (X < 100) X = 100;



	MIBenum_CLASS[MIBenum_class].MIBenum_array =
	    safe_realloc (MIBenum_CLASS[MIBenum_class].MIBenum_array,
			  X * sizeof (MIBenum_CLASS[MIBenum_class].MIBenum_array[0])); 

	for (i = MIBenum_CLASS[MIBenum_class].MIBenum_array_len; i < X; i++)
	    MIBenum_CLASS[MIBenum_class].MIBenum_array[i] = NULL;
	
	MIBenum_CLASS[MIBenum_class].MIBenum_array_len = X;
    }

    if (MIBenum_CLASS[MIBenum_class].MIBenum_array[val] &&
	MIBenum_CLASS[MIBenum_class].MIBenum_array[val] != ptr &&
	MIBenum_CLASS[MIBenum_class].MIBenum_array[val] -> flags & SET_valid)
	return 0;

    if (MIBenum_CLASS[MIBenum_class].MIBenum_array[val])
	MIBenum_CLASS[MIBenum_class].MIBenum_array[val] -> MIBenum = 0;

    MIBenum_CLASS[MIBenum_class].MIBenum_array[val] = ptr; 
    
    return 1;
}

static void add_builtin_sets P_((void));
static void add_builtin_sets(void)
{
    int i;

    for (i = 0; precompiled_sets[i].charset_type; i++) {
	if (precompiled_sets[i].MIBenum) {
	    DPRINT(Debug,49,(&Debug,"(#%d) Setting MIBEnum for %s (%d)\n",
			     i,
			     precompiled_sets[i].MIME_name ? 
			     precompiled_sets[i].MIME_name  :
			     "<no MIME name>",
			     precompiled_sets[i].MIBenum));

	    if (!add_MIBenum(& (precompiled_sets[i]))) {
		panic("CHARSET PANIC",__FILE__,__LINE__,"add_builtin_set",
		      "Duplicate or bad MIBenum on builtin set",0);
	    }			 
	}
    }
    array_initialized++;
}


static struct charcode_info * add_set P_((struct charcode_info s));
static struct charcode_info * add_set(s)
     struct charcode_info s;
{
    struct extra_sets *tmp = safe_malloc( sizeof (struct extra_sets));
    
    tmp->set  = s;
    tmp->next = NULL;

    if (!array_initialized)
	add_builtin_sets();

    if (extra_sets_tail)
	extra_sets_tail->next = tmp;    
    extra_sets_tail = tmp;    
    if (!extra_sets)
	extra_sets = tmp;
    
    extra_sets_count++;

    if (tmp->set.MIME_name) {
	DPRINT(Debug,1,(&Debug, 
			"charset: Adding charset %s (type %s)\n",
		  tmp->set.MIME_name,
		  tmp->set.charset_type->type_name));	
    }

    if (tmp->set.MIBenum && !add_MIBenum(& (tmp->set))) {
	lib_error(CATGETS(elm_msg_cat, MeSet,
			  MeDuplicateMibEnum,
			  "Charset %s have duplicate or bad MIBenum %d"),
		  tmp->set.MIME_name ? tmp->set.MIME_name : "<no MIME name>",
		  tmp->set.MIBenum);

	tmp->set.MIBenum = 0;   /* Discard MIBEnum */
	

    }

    return &(tmp->set);
}

static int same_name P_((struct charcode_info *set, union compare_info data));
static int same_name (set,data)
     struct charcode_info *set;
     union compare_info data;
{
    DPRINT(Debug,71,(&Debug, 
		     "same_name -- set=%p (MIME_name=%s,codeset=%s,data.name=%s)\n",
		     set,
		     set->MIME_name ? set->MIME_name : "<NULL>",
		     set->codeset   ? set->codeset : "<NULL>",
		     data.name));
    if (set->codeset &&
	0 == istrcmp(data.name,set->codeset))
	return 1;
    
    if (!set->MIME_name)
	return 0;
    return 0 == istrcmp(data.name,set->MIME_name);
}

static struct charset_alias {
       char *                name;
       charset_t             set;
} *    charset_alias_list  = NULL;
static int charset_alias_count = 0;

static int locate_charset_alias P_((const char * name));
static int locate_charset_alias(name)
     CONST char * name;
{
    int i;

    for (i = 0; i < charset_alias_count; i++)
	if (0 == istrcmp(charset_alias_list[i].name,name))
	    return i;
    return -1;
}

static int add_charset_alias P_((const char * name, charset_t set));
static int add_charset_alias(name,set)
     CONST char * name;
     charset_t set;
{
    charset_alias_list = safe_realloc(charset_alias_list,
				      (charset_alias_count +1) *
				      sizeof (struct charset_alias));
    charset_alias_list[charset_alias_count].name = safe_strdup(name);
    charset_alias_list[charset_alias_count].set = set;
    return charset_alias_count++;
}

charset_t codeset_name_to_charset (name)
     CONST char *name;
{
    charset_t ret;
    union compare_info D;
    D.name =  name;
    
    ret = find_set(same_name,D);

    return ret;
}

charset_t MIME_name_to_charset (name,flag)
     CONST char *name;
     int flag;
{
    int create          = flag & CHARSET_create;
    CONST int noautodef = flag & CHARSET_noautodef;
    int i;
    charset_t ret;
    struct charcode_info new;
    int do_ascii = 0;

    int do_cs = 0;
    struct  charset_type * cs_type = NULL;
    struct  map_info     * cs_map  = NULL;

    union compare_info D;
    D.name =  name;
    
    ret = find_set(same_name,D);

    if (!ret && 0 == istrcmp(name,"US-ASCII")) {
	DPRINT(Debug,7,(&Debug, 
		   "MIME_name_to_charset: US-ASCII not found -- using builting constant...\n"));
	ret = ASCII_SET;
    }


    if (ret) {
	DPRINT(Debug,9,(&Debug, "MIME_name_to_charset(%s), found=%p",
			name,ret));

    found_it:
	if (ret->codeset) {
	    DPRINT(Debug,9,(&Debug, ", codeset=%s",ret->codeset));
	}
	DPRINT(Debug,9,(&Debug, "\n",ret->codeset));

	if (!ret->MIME_name && create) {
	    CONST char * X =  ret->codeset ? ret->codeset : name;   /* Should be unnecessary */
	    DPRINT(Debug,1,(&Debug,"*** No MIME name for charset %s, using %s as MIME name\n",
			    name,X));
	    ret->MIME_name = safe_strdup(X);
	}

	if (!ret->MIME_name) {
	    DPRINT(Debug,1,(&Debug,"*** No MIME name for charset %s (codeset %s)\n",
			    name,ret->codeset ? ret->codeset : "<?>"));

	    lib_error(CATGETS(elm_msg_cat, MeSet,
			      MeNoMIMEname,
			      "Charset %s exists, but that is not suitable name for MIME use"),
		      name);
	    return NULL;
	}

	/* Initialize associated map if not done already ... */       
	if (ret->map_info && !ret->map_info->map_initialized)
	    ret->map_info->map_init_it(ret->map_info);

	return ret;
    }

    /* Check aliases ... */
    if (0 <= (i = locate_charset_alias(name)) &&
	(ret = charset_alias_list[i].set) &&
	ret->MIME_name &&
	/* Ignore aliases which point invalidated charsets.
	   These charsets may me invalidated by setting
	   explicit compatcahrsets on user's elmrc
	*/
	(0 != (ret-> flags & SET_valid))) {

	DPRINT(Debug,9,(&Debug,
			"MIME_name_to_charset(%s), found=%p (alias of %s)",
			name,ret,
			ret->MIME_name ? ret->MIME_name : "<no MIME name>"));	
	goto found_it;

    }

    if (!ret && 0 == strincmp(name,"ISO-8859-",9) &&
	0 < (i = atoi(name+9))) {
	DPRINT(Debug,7,(&Debug, "MIME_name_to_charset: ISO-8859-X charset, "));
	if (noautodef) {
	    DPRINT(Debug,7,(&Debug, "CHARSET_noautodef set\n"));
	} else if (auto_iso_8859) {
	    do_ascii = 1;
	    create = 1;
	    DPRINT(Debug,7,(&Debug, " autocreating (as subtype of builtin ASCII constant)\n"));
	} else {
	    DPRINT(Debug,7,(&Debug, "auto-iso-8859 is not set.\n"));
	}
    }

#ifdef USE_DLOPEN
    if (!ret && !noautodef &&
	cs_auto_charset(name,&cs_type,&cs_map)) {
	do_cs = 1;
	create = 1;
	DPRINT(Debug,7,(&Debug, 
			"MIME_name_to_charset: %s charset, autocreating\n",
			name));
    }
#endif

    if (!create) {
	DPRINT(Debug,12,(&Debug,"MIME_name_to_charset(%s), not found\n",
			 name));
	return NULL;
    }    

    new.flags          = SET_nodata|SET_valid;
    new.map_info       = NULL;
    new.MIME_name      = safe_strdup(name);

    if (do_cs) {
	new.map_info       = cs_map;
	new.charset_type   = cs_type;
	new.subtype        = NULL;
    } else if (do_ascii) {
	new.charset_type   = &cs_ascii;
	new.subtype        = ASCII_SET;
    } else {
	new.charset_type   = &cs_unknown;
	new.subtype        = NULL;
    }
    new.iso2022_info   = NULL;
    new.MIBenum        = 0;
    new.codeset        = NULL;

    ret = add_set(new);

    DPRINT(Debug,12,(&Debug,"MIME_name_to_charset(%s), adding=%p\n",
		     name,ret));

    return ret;
}

static int add_it_ P_((charset_t ascii_ptr,
		       struct charcode_info *item,
		       struct charcode_info **list,
		       int *count, int max));


static int add_it_ (ascii_ptr,item,list,count,max)
     charset_t ascii_ptr;
     struct charcode_info *item;
     struct charcode_info **list;
     int *count; 
     int max;
{
    /* Pick only MIME charsets ... */

    if (item->MIME_name) 
	if (item->subtype == ascii_ptr) 		    
	    if (*count < max) 
		list[(*count)++] = item; 
	    else { 
		lib_error(FRM("charset: Internal error -- overflow")); 
		return 0;
	    }
    
    return 1;
}


static int array_walk_ P_((charset_t ascii_ptr,
			   struct charcode_info *arr,
			   struct charcode_info **list,
			   int *count, int max));


static int array_walk_ (ascii_ptr,arr, list, count, max)
     charset_t ascii_ptr;
     struct charcode_info *arr;
     struct charcode_info **list;
     int *count;
     int max;
{
    int i;

    
    for(i = 0; arr[i].charset_type; i++) 
	if (0 != (arr[i].flags & SET_valid))
	    if (!add_it_(ascii_ptr,&arr[i],list,count,max))
		return 0;

    return 1;
}

static int name_ok P_((CONST char *name));
static int name_ok(name)
     CONST char *name;
{
    if ('\0' == *name)
	return 0;

    return strlen(name) == 
	strspn(name,
	       "ABCDEFGHIJKLMNOPQRSTUVWXYZ-0123456789abcdefghijklmnopqrstuvwxyz");
}

int charset_compatfunc (value,enter)
     char **value;
     int enter;
{
    struct extra_sets *walk;
    int ret = 0;
    charset_t ascii_ptr = MIME_name_to_charset("US-ASCII",0);

    CONST int size = precompiled_size + extra_sets_count;

    struct charcode_info **need_modify_list =
	safe_malloc(size * sizeof(struct charcode_info *));
    int count = 0;

    if (!ascii_ptr)
	panic("CHARSET PANIC",__FILE__,__LINE__,"charset_compatfunc",
	      "US-ASCII not found",0);

#define HANDLE(SET) if (!array_walk_(ascii_ptr,SET,need_modify_list,&count,size)) goto fail;
#define ADD_IT(ITEM) if (!add_it_(ascii_ptr,ITEM,need_modify_list,&count,size)) goto fail;
        
    HANDLE(precompiled_sets);

    for (walk = extra_sets; walk; walk = walk -> next) 
	if (0 != (walk->set.flags & SET_valid))
	    ADD_IT(&(walk->set));

    if (enter) {
	char * temp = safe_strdup(*value);
	char *wrk,*next;
	int j;
	
	for (j = 0; j < count; j++)
	    need_modify_list[j]->flags |= SET_mark;

	for (wrk = temp; wrk && *wrk; wrk = next) {
	    charset_t charset_ptr = NULL;

	    next = qstrpbrk(wrk," ,;");
	retry:
	    if (next) {
		char *a = next;
		if (';' != *next)
		    next++;
		while (' ' == *next)
		    next++;
		if (';' == *next) {
		    next++;
		    while (' ' == *next)
			next++;
		    next = qstrpbrk(next," ,");
		    goto retry;
		}
		*a = '\0';
	    }

	    if (!load_charset_map_info(&charset_ptr,wrk)) 
		charset_ptr = MIME_name_to_charset(wrk,
						   CHARSET_create|
						   CHARSET_noautodef); 

	    if (0 != (charset_ptr -> flags & SET_mark)) {
		charset_ptr -> flags &=  ~SET_mark;
		charset_ptr -> flags &=  ~SET_nodata;
		if (&cs_unknown == charset_ptr -> charset_type)
		    charset_ptr -> charset_type  = &cs_ascii;

		DPRINT(Debug,39,(&Debug, 
			   "charset_compatfunc: modify=%p (%s), flags=%d, type=%s\n",
			   charset_ptr,
			   charset_ptr->MIME_name ? 
			   charset_ptr->MIME_name : "<no MIME name>",
			   charset_ptr->flags,
			   charset_ptr->charset_type->type_name));

	    } else {  /* new charset create information */
		
		charset_ptr -> charset_type  = &cs_ascii;
		charset_ptr -> flags         = SET_valid;
		charset_ptr -> subtype       = ascii_ptr;
		DPRINT(Debug,39,(&Debug, 
				 "charset_compatfunc: new=%p (%s), flags=%d, type=%s\n",
				 charset_ptr,
				 charset_ptr->MIME_name ? 
				 charset_ptr->MIME_name : "<no MIME name>",
				 charset_ptr->flags,
				 charset_ptr->charset_type->type_name));
	    }

	    if (charset_ptr->map_info && 
		charset_ptr->map_info->map_type != charset_ptr->charset_type) {

		lib_error(CATGETS(elm_msg_cat, MeSet,
				  MeCompatcharsetsDiscardMap,
				  "WARNING: %s redefined by compatcharsets, discarding map defination"),
			  charset_ptr->MIME_name ?
			  charset_ptr->MIME_name : "<no MIME name>");
		charset_ptr->map_info = NULL;		
	    }

	}
	     
	free(temp);

	/* invalidate rest */
	for (j = 0; j < count; j++)
	    if (0 != (need_modify_list[j]->flags & SET_mark)) {
		need_modify_list[j] -> flags &=  ~SET_valid;
		DPRINT(Debug,1,(&Debug,"charset: Invalidating charset %s\n",
				need_modify_list[j]->MIME_name));	
	    }

    } else {
	int j;
	
	/* static pointer to buffer accross invocations */
	static char * return_buffer = NULL;

	return_buffer = strmcpy(return_buffer,"");    /* Reset result */
	for (j = 0; j < count; j++) {
	    char * s = need_modify_list[j]->MIME_name;
	    
	    if (return_buffer[0])
		return_buffer = strmcat(return_buffer," ");
	    
	    if (name_ok(s))
		return_buffer = strmcat(return_buffer,
					need_modify_list[j]->MIME_name);
	    else {
		char * tmp = elm_message(FRM("%Q;!"),s);
		return_buffer = strmcat(return_buffer,tmp);
		free(tmp);
	    }
	}

	*value = return_buffer;
    }
    ret = 1;
    
	fail:
    free(need_modify_list);
    return ret;
}
#undef ADD_IT
#undef HANDLE

int charset_iso646func (value,enter)
     char **value;
     int enter;
{
    struct extra_sets *walk;
    int ret = 0;
    charset_t invariant_ptr = MIME_name_to_charset("INVARIANT",0);

    CONST int size =  precompiled_size + extra_sets_count;

    struct charcode_info **need_modify_list =
	safe_malloc(size * sizeof(struct charcode_info *));
    int count = 0;

    if (!invariant_ptr)
	invariant_ptr = INVARIANT_SET;

#define HANDLE(SET) if (!array_walk_(invariant_ptr,SET,need_modify_list,&count,size)) goto fail;
#define ADD_IT(ITEM) if (!add_it_(invariant_ptr,ITEM,need_modify_list,&count,size)) goto fail;
        
    HANDLE(precompiled_sets);

    for (walk = extra_sets; walk; walk = walk -> next) 
	if (0 != (walk->set.flags & SET_valid))
	    ADD_IT(&(walk->set));

    if (enter) {
	char * temp = safe_strdup(*value);
	char *wrk,*next;
	int j;
	
	for (j = 0; j < count; j++)
	    need_modify_list[j]->flags |= SET_mark;

	for (wrk = temp; wrk && *wrk; wrk = next) {
	    charset_t charset_ptr = NULL;

	    next = qstrpbrk(wrk," ,;");
	retry:
	    if (next) {
		char *a = next;
		if (';' != *next)
		    next++;
		while (' ' == *next)
		    next++;
		if (';' == *next) {
		    next++;
		    while (' ' == *next)
			next++;
		    next = qstrpbrk(next," ,");
		    goto retry;
		}
		*a = '\0';
	    }

	    if (!load_charset_map_info(&charset_ptr,wrk)) 
		charset_ptr = MIME_name_to_charset(wrk,
						   CHARSET_create|
						   CHARSET_noautodef); 

	    if (0 != (charset_ptr -> flags & SET_mark)) {
		charset_ptr -> flags &=  ~SET_mark;
		charset_ptr -> flags &=  ~SET_nodata;
		if (&cs_unknown == charset_ptr -> charset_type)
		    charset_ptr -> charset_type  = &cs_iso646;

		DPRINT(Debug,39,(&Debug, 
			   "charset_iso646func: modify=%p (%s), flags=%d, type=%s\n",
			   charset_ptr,
			   charset_ptr->MIME_name ? 
			   charset_ptr->MIME_name : "<no MIME name>",
			   charset_ptr->flags,
			   charset_ptr->charset_type->type_name));

	    } else {  /* new charset create information */
		charset_ptr -> charset_type  = &cs_iso646;
		charset_ptr -> flags         = SET_valid;
		charset_ptr -> subtype       = invariant_ptr;
		DPRINT(Debug,39,(&Debug, 
			   "charset_iso646func: new=%p (%s), flags=%d, type=%s\n",
			   charset_ptr,
			   charset_ptr->MIME_name ? 
			   charset_ptr->MIME_name : "<no MIME name>",
			   charset_ptr->flags,
			   charset_ptr->charset_type->type_name));
	    }

	    if (charset_ptr->map_info && 
		charset_ptr->map_info->map_type != charset_ptr->charset_type) {
		
		lib_error(CATGETS(elm_msg_cat, MeSet,
				  MeIso646charsetsDiscardMap,
				  "WARNING: %s redefined by iso646-charsets, discarding map defination"),
			  charset_ptr->MIME_name ?
			  charset_ptr->MIME_name : "<no MIME name>");
		charset_ptr->map_info = NULL;		
	    }

	}
	     
	free(temp);

	/* invalidate rest */
	for (j = 0; j < count; j++)
	    if (0 != (need_modify_list[j]->flags & SET_mark)) {
		need_modify_list[j] -> flags &=  ~SET_valid;
		DPRINT(Debug,1,(&Debug,"charset: Invalidating charset %s\n",
				need_modify_list[j]->MIME_name));	
	    }

    } else {
	int j;
	
	/* static pointer to buffer accross invocations */
	static char * return_buffer = NULL;

	return_buffer = strmcpy(return_buffer,"");    /* Reset result */
	for (j = 0; j < count; j++) {
	    char * s = need_modify_list[j]->MIME_name;
	    
	    if (return_buffer[0])
		return_buffer = strmcat(return_buffer," ");
	    
	    if (name_ok(s))
		return_buffer = strmcat(return_buffer,
					need_modify_list[j]->MIME_name);
	    else {
		char * tmp = elm_message(FRM("%Q;!"),s);
		return_buffer = strmcat(return_buffer,tmp);
		free(tmp);
	    }
	}

	*value = return_buffer;
    }
    ret = 1;
    
	fail:
    free(need_modify_list);
    return ret;
}
#undef ADD_IT
#undef HANDLE

#undef ARRSIZE

char * dequote_opt(source,len)
     CONST char *source; 
     int len;
{
    int size = len + 1;
    int ptr = 0;
    char * store = safe_malloc(size);
    CONST char *p;
    int q = 0;
    
    for (p = source; p - source < len && ptr < size-1; p++) {
	switch(*p) {
	case '"':
	    q = !q;
	    break;
	case '\\':
	    if (q) {
		p++;
		if (*p != '\0')
		    store[ptr++] = *p;
		else {
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeTrailingBackslash,
				      "Trailing backslash (\\): %.20s..."),
			      source);
		}
		break;
	    }
	    /* FALLTHRU */
	default:
	    if (q || *p != ' ')
		store[ptr++] = *p;
	    break;
	}    
    }
    store[ptr] = '\0';

    if (q)
	lib_error(CATGETS(elm_msg_cat, MeSet, 
			  MeUnbalancedQuote,
			  "Unbalanced quote (\"): %.20s..."),
		  source);


    DPRINT(Debug,20,(&Debug, 
		    "dequote_opt(%.*s (len=%d))=%s\n",
		    len,source,len,store));

    return store;
}

int load_charset_map_info(buffer,data)
     charset_t *buffer; 
     CONST char *data;
{
    char * temp = safe_strdup(data);
    char *c;
    char *store,*opt;
    struct charcode_info new_vals;
    int reallocation_needed = 0;
    char ** alias_vector = NULL;
    int alias_count      = 0;
    struct setlist     new_setlist;
    int         setcount          = 0;
    int         define_as_codeset = 0;

    iso2022_clear_setlist(&new_setlist);

    if (0 == strncmp(data,"codeset=",8)) {
	store = NULL;

	*buffer = NULL;

	reallocation_needed   = 0;   /* codeset= code will set this */
	new_vals.charset_type = &cs_unknown;
	new_vals.map_info     = NULL;
	new_vals.flags        = SET_valid;
	new_vals.MIME_name    = NULL;
	new_vals.subtype      = NULL;
	new_vals.iso2022_info = NULL;

	new_vals.MIBenum        = 0;
	new_vals.codeset        = NULL;

	define_as_codeset = 1;

	c = temp;
	goto codeset_hook;
    }


    c = qstrpbrk(temp,";");
    if (!c) {
	*buffer = NULL;
	
	DPRINT(Debug,7,(&Debug, 
		     "load_charset_map_info(\"%s\") = 0 (FAIL)\n", 
		     data));
	free(temp);
	return 0;
    }
    store = dequote_opt(temp,c - temp);
    DPRINT(Debug,11,(&Debug, 
		 "load_charset_map_info: charset=%s\n", 
		 store));

    DPRINT(Debug,11,(&Debug, 
		 "load_charset_map_info: info=%s\n", c));

    *buffer = MIME_name_to_charset(store,0);
    if (!*buffer) {
	reallocation_needed   = 1;
	new_vals.charset_type = &cs_unknown;
	new_vals.map_info     = NULL;
	new_vals.flags        = SET_valid;
	new_vals.MIME_name    = safe_strdup(store);
	new_vals.subtype      = NULL;
	new_vals.iso2022_info = NULL;

	new_vals.MIBenum        = 0;
	new_vals.codeset        = NULL;

    } else 
	new_vals = **buffer;
    
    c++;

    if ('\0' == *c || 0 == strcmp(c,"!")) {
	DPRINT(Debug,9,(&Debug,"charset: %s: no data\n",store));
	new_vals.flags       |= SET_nodata;
    } else { 
	int i;
	char * WALK;
       
	if (0 <= (i = locate_charset_alias(store)) &&
	    charset_alias_list[i].set == *buffer) {
	    DPRINT(Debug,1,(&Debug,
			    "DANGER: Defining charset %s which is also alias of %s\n",
			    store,
			    charset_alias_list[i].set->MIME_name ?
			    charset_alias_list[i].set->MIME_name :
			    "<no MIME name>"));
	    alias_vector = safe_realloc(alias_vector,
					(alias_count+1) *
					sizeof (char *));
	    alias_vector[alias_count++] = safe_strdup(store);

	    lib_error(CATGETS(elm_msg_cat, MeSet,
			      MeTryingCharsetAlias,
			      "Trying redefine alias %s as charset"),
		      store);
	}

    codeset_hook:
	WALK = NULL;

	for (opt = mime_parse_content_opts(c, &WALK); 
	     opt; 
	     opt = mime_parse_content_opts(NULL, &WALK)) {
	    char * q = strchr(opt,'=');
	    
	    DPRINT(Debug,20,(&Debug, "mime_parse_content_opts gives: %s\n",
			     opt));

	    if (q) {
		char *val; 
		*q++ = '\0';
		val = dequote_opt(q,strlen(q));

		if (0 == strcmp(opt,"codeset")) {		
		    charset_t X = codeset_name_to_charset(val);

		    if (X && X != *buffer && !define_as_codeset) {
			lib_error(CATGETS(elm_msg_cat, MeSet, 
					  MeCodesetRedefined,
					  "Codeset %s redefined -- invalidating old charset %s"),
				  X->codeset ? X->codeset : val,
				  X->MIME_name ? X->MIME_name : val);
			X -> flags &=  ~SET_valid;
		    }
		    
		    if (define_as_codeset) {
			*buffer = X;
			if (X) {
			    new_vals = *X;
			    free(val); val = NULL;
			} else {
			    reallocation_needed   = 1;
			    new_vals.codeset = val; val = NULL;
			}
			define_as_codeset = 0;
		    } else {
			reallocation_needed   = 1;
			new_vals.codeset = val; val = NULL;
		    }
		} else if (0 == strcmp(opt,"MIBenum")) {
  
		    new_vals.MIBenum = atoi(val);

		    if (new_vals.MIBenum < 1) {
			lib_error(CATGETS(elm_msg_cat, MeSet, 
					  MeBadMIBenum,
					  "Bad MIBenum %s"),
				  val);						
			new_vals.MIBenum = 0;
		    }

		    free(val); val = NULL;

		} else if (0 == strcmp(opt,"MIME-subset")) {
		    reallocation_needed = 1;
		    new_vals.flags  &= ~SET_nodata;

		    new_vals.subtype  = MIME_name_to_charset(val,0);
		    if (!new_vals.subtype) {
			lib_error(CATGETS(elm_msg_cat, MeSet, 
					  MeUnknownMIMEsubset,
					  "Unknown MIME-subset %s"),
				  val);				  
		    } 
		    
		    if (0 == istrcmp(val,"US-ASCII") &&
			&cs_unknown == new_vals.charset_type)
			new_vals.charset_type = &cs_ascii;
		    if (0 == istrcmp(val,"INVARIANT") &&
			&cs_unknown == new_vals.charset_type)
			new_vals.charset_type = &cs_iso646;
		    free(val); val = NULL;

		} else if (0 == strcmp(opt,"map")) {
		    struct  map_info * r;
		    if (new_vals.charset_type == &cs_unknown) {
			lib_error(CATGETS(elm_msg_cat, MeSet,
                                          MeCharsetTypeBeforeMap,
                                          "Charset type must be given before map -parameter"));
		    }
		    r = new_vals.charset_type->find_map_type(val);
		    if (!r) {
			lib_error(CATGETS(elm_msg_cat, MeSet,
                                          MeUnknownMap,
					  "No map %s on charset type %s"),
				  val,new_vals.charset_type->type_name);
		    } else {
			reallocation_needed = 1;
			new_vals.flags     &= ~SET_nodata;
			new_vals.map_info   = r;
		    }
		    free(val); /* ?? */ val = NULL;

		} else if (0 == strcmp(opt,"type")) {
		    struct  charset_type * walk;

		    for (walk = cs_first_type; walk; walk = walk -> next_type)
			if (0 == strcmp(val,walk->type_name))
			    break;
#ifdef USE_DLOPEN
		    if (!walk)
			walk = find_shared_CS_type(val);
#endif		    
		    if (!walk) {
			lib_error(CATGETS(elm_msg_cat, MeSet,
                                          MeUnknownCharsetType,
                                          "Unknown charset type %s"),
                                  val);
			walk = &cs_unknown;
		    }	
		    reallocation_needed   = 1;
		    new_vals.flags       &= ~SET_nodata;
		    new_vals.charset_type = walk;

		    free(val); val = NULL;
		} else if (0 == strcmp(opt,"alias")) {

		    if (new_vals.codeset &&
			0 == istrcmp(new_vals.codeset,val)) {
			DPRINT(Debug,1,(&Debug,
					"charset: Ignoring alias=\"%s\" for codeset=\"%s\"\n",
					val,new_vals.codeset));
			free(val); val = NULL;

		    } else if (new_vals.MIME_name &&
			0 == istrcmp(new_vals.MIME_name,val)) {
			DPRINT(Debug,1,(&Debug,
					"charset: Ignoring alias=\"%s\" for charset %s\n",
					val,new_vals.MIME_name));
			free(val); val = NULL;

					
		    } else {
			int i = locate_charset_alias(val);
			charset_t p1;
			
			if (i < 0 &&
			    (p1 = codeset_name_to_charset(val))) {

			    lib_error(CATGETS(elm_msg_cat, MeSet,
					      MeTryingAliasCodeset,
					      "Trying redefine codeset %s as alias"),
				      p1->codeset ? p1->codeset : 
				      p1->MIME_name ? p1->MIME_name :
				      val);
			    free(val); val = NULL;
			} else {
			    alias_vector = safe_realloc(alias_vector,
							(alias_count+1) *
							sizeof (char *));
			    alias_vector[alias_count++] = val; val = NULL;
			}
		    }
		} else if (parse_iso2022_specification(opt,val,
						       &setcount,
						       &new_setlist)) {
		    /* none */
		    free(val); /* ?? */ val = NULL;
		} else {
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeUnsupportedCharsetOption,
				      "Unsupported charset option %s (value %.20s...)"),
			      opt,val);				  
		    free(val); val = NULL;
		}
	    } else {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeUnsupportedCharsetOptionNoValue,
				  "Unsupported charset option %s (no value)"),
			  opt);				  
	    }
	}
    }

    if (setcount > 0) {

	if (new_vals.iso2022_info)
	    iso2022_merge_old_info(&new_setlist,setcount,new_vals.iso2022_info);

	if (new_vals.charset_type->cs_iso2022_info_set_it(&new_vals,&new_setlist,
							  setcount)) {
	    reallocation_needed   = 1;
	    new_vals.flags       &= ~SET_nodata;

	} else
	    goto iso2022_fail; /* Discard bank defination */


	/* TODO: Other processing? */

    }
 iso2022_fail:

    if (reallocation_needed) {
	/* Reset incompatble map types ... */
	if (new_vals.map_info && 
	    new_vals.map_info->map_type != new_vals.charset_type) {
	    DPRINT(Debug,1,(&Debug,
		      "charset: Resetting incompatible map %s (type %s) from charset %s (type %s)\n",
		      new_vals.map_info->map_name,
		      new_vals.map_info->map_type->type_name,
		      new_vals.MIME_name ? 
		      new_vals.MIME_name : "<no MIME name>",
		      new_vals.charset_type->type_name));
		   new_vals.map_info = NULL;
	}
	
	if (!new_vals.map_info && &cs_onebyte == new_vals.charset_type) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeNeedsMapParamater,
			      "Charset type %s requires map parameter"),
		      new_vals.charset_type->type_name);

	    if (new_vals.subtype == ASCII_SET) {
		DPRINT(Debug,1,(&Debug,
			  "charset: Charset %s: setting type to %s instead of %s\n",
			  (*buffer)->MIME_name ?
			  (*buffer)->MIME_name : "<no MIME name>",
			  cs_ascii.type_name,
			  new_vals.charset_type->type_name));
		new_vals.charset_type = &cs_ascii;
	    }
	    if (new_vals.subtype == INVARIANT_SET) {
		DPRINT(Debug,1,(&Debug,
			  "charset: Charset %s: setting type to %s instead of %s\n",
			  (*buffer)->MIME_name ?
			  (*buffer)->MIME_name : "<no MIME name>",
			  cs_iso646.type_name,
			  new_vals.charset_type->type_name));
		new_vals.charset_type = &cs_iso646;
	    }
	}

	if (*buffer) {
	    (*buffer) -> flags  &= ~SET_valid;
	    DPRINT(Debug,1,(&Debug,
		      "charset: Invalidating charset %s (and regenerating)\n",
		      (*buffer)->MIME_name ? 
		      (*buffer)->MIME_name : "<no MIME name>"));	
	}
	*buffer = add_set(new_vals);
	DPRINT(Debug,9,(&Debug,
			"charset: Adding charset %s (type %s)\n",
			(*buffer)->MIME_name ? 
			(*buffer)->MIME_name : "<no MIME name>",
			(*buffer)->charset_type->type_name));	
    }

    if (alias_vector) {
	int i;
	for (i = 0; i < alias_count; i++) {
	    int j =  locate_charset_alias(alias_vector[i]);
	    if (j < 0)
		j = add_charset_alias(alias_vector[i],*buffer);

	    /* In case of we redefine alias ... */
	    /* This is BUGGY on case of multiple aliases */ 
	    charset_alias_list[j].set->flags &= ~SET_havealias;

	    (*buffer) -> flags  |= SET_havealias;

	    DPRINT(Debug,9,(&Debug,
			    "load_charset_map_info: Setting %s as alias of %s\n",
			    charset_alias_list[j].name,
			    (*buffer)->MIME_name ?
			    (*buffer)->MIME_name : "<no MIME name>"));

	    charset_alias_list[j].set = *buffer;

	    free(alias_vector[i]);
	    alias_vector[i] = NULL;
	}
	free(alias_vector);
    }
    DPRINT(Debug,9,(&Debug,
		    "load_charset_map_info(\"%s\") = 1 (SUCCEED)\n", 
		    data));    
    free(temp);
    if (store)
	free(store);
    return 1;    
}

static void dump_alias_info P_((FILE *f,struct charcode_info *info));
static void dump_alias_info(f,info)
     FILE *f;
     struct charcode_info *info;
{
    int i;
    for (i = 0; i < charset_alias_count; i++)
	if (info == charset_alias_list[i].set &&
	    charset_alias_list[i].name) {
	    if (name_ok(charset_alias_list[i].name))
		elm_fprintf(f,FRM(";alias=%s"),charset_alias_list[i].name);
	    else
		elm_fprintf(f,FRM(";alias=%Q"),charset_alias_list[i].name);
	}   
}

void dump_map_info(f,info)
     FILE *f;
     struct charcode_info *info;
{
   
    if (!info->MIME_name) {
	if (!info->codeset)
	    fputs("# <No MIME name> ",f);
    } else if (name_ok(info->MIME_name)) {
	fputs(info->MIME_name,f);
    } else {
	elm_fprintf(f,FRM("%Q"),info->MIME_name);
    }
    
    if (info->codeset) {
	elm_fprintf(f,FRM("%scodeset=%Q"),
		    info->MIME_name ? ";" : "",
		    info->codeset);
    } else if (0 != (info->flags & SET_nodata)) {
	fputs(";!",f);
	return;
    }

    if (&cs_unknown != info->charset_type)
	elm_fprintf(f,FRM(";type=%s"),info->charset_type->type_name);
    if (info->map_info && name_ok(info->map_info->map_name))
	elm_fprintf(f,FRM(";map=%s"),info->map_info->map_name);
    else if (info->map_info)
	elm_fprintf(f,FRM(";map=%Q"),info->map_info->map_name);

    if (info->subtype && 
	info->subtype->MIME_name &&
	name_ok(info->subtype->MIME_name))
	elm_fprintf(f,FRM(";MIME-subset=%s"),info->subtype->MIME_name);
    else if (info->subtype && 
	     info->subtype->MIME_name)
	elm_fprintf(f,FRM(";MIME-subset=%Q"),info->subtype->MIME_name);

    if (info->MIBenum) {
	fprintf(f,";MIBenum=%d",info->MIBenum);
    }

    if (info->iso2022_info)
	print_setlist(f,info->iso2022_info);

    if (0 != (info->flags & SET_havealias)) {
	dump_alias_info(f,info);
    }
}

int get_charset_map_info (buffer,data,size)
     char *buffer;
     CONST char *data; 
     int size;
{
    charset_t set;

    if (!load_charset_map_info(&set,data) ||
	!set->MIME_name) {
	buffer[0] = '\0';

	DPRINT(Debug,7,(&Debug, 
		     "get_charset_map_info(\"%s\") = 0 (FAIL)\n", 
		     data));

	return 0;
    }

    strfcpy(buffer,set->MIME_name,size);

    DPRINT(Debug,11,(&Debug, 
		 "get_charset_map_info(\"%s\") = 1 (SUCCEED)\n", 
		 data));
    return 1;
}

static int charset_superset_of_p P_((charset_t charset, 
				     charset_t subset,
				     int level));

static int charset_superset_of_p(charset,subset,level)
     charset_t charset,subset;
     int level;
{

    if (level > 10) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeCharsetSubsetLoop,
			  "Charset subset defination loop detected for %s"),
		  charset->MIME_name ? charset->MIME_name : "<no MIME name>");
	return 0;
    }

    if (charset->subtype == subset)
	return 1;

    if (charset->subtype &&
	charset_superset_of_p(charset->subtype,subset,level+1))
	return 1;
    return 0;
}

int charset_superset_of(charset,subset)
     charset_t charset,subset;
{
    int res = charset_superset_of_p(charset,subset,0);

    DPRINT(Debug,6,(&Debug, 
		    "charset_superset_of(%p (%s),%p (%s))=%d\n",
	      charset,
	      charset->MIME_name ? charset->MIME_name :"<no MIME name>",
	      subset,subset->MIME_name ? subset->MIME_name :"<no MIME name>",
	      res));
    return res;
}

int charset_ok_p(ptr)
     charset_t ptr;
{
    charset_t ascii_ptr = MIME_name_to_charset("US-ASCII",0);

    if (!ascii_ptr)
	panic("CHARSET PANIC",__FILE__,__LINE__,"charset_ok_p",
	      "US-ASCII not found",0);
    return charset_superset_of(ptr,ascii_ptr);
}

struct locale_map_item  * load_locale_map(filename,errors) 
     CONST char *filename;
     int *errors;
{
    struct locale_map_item *result;
    int result_len = 0;
    FILE * f = fopen(filename,"r");
    int max_result = 0;
    int c,l1;
    char buf[LONG_STRING];

    if (!f) {
	int err = errno;
	DPRINT(Debug,2,(&Debug, 
			"load_locale_map: %s: %s\n",
			filename,error_description(err)));
	return NULL;
    }

    while(EOF != (c = fgetc(f)))
	if ('\n' == c)
	    max_result++;

    DPRINT(Debug,11,(&Debug, 
		     "load_locale_map: %s, max_result=%d\n",
		     filename,max_result));

    if (!max_result) {
	fclose(f);
	return NULL;
    }
    rewind(f);

    result = safe_malloc((max_result +1) * sizeof (struct locale_map_item));

    while (result_len < max_result &&
	   (l1 = mail_gets(buf,sizeof buf, f)) > 0) {
	char * c = buf,*c1;
	
	if ('\n' == buf[l1 -1]) 
	    buf[l1 - 1] = '\0';
	else {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLine,
			      "%30s: Too long line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;
	}
	l1--;

	while (l1-- > 0 && whitespace(buf[l1]))
	    buf[l1] = '\0';
	
	c = buf;
	while (*c && whitespace (*c)) /* skip leading whitespace */
	    c++;
	if ('#' == *c)
	    continue;
	if (!*c)
	    continue;

	c1 = strpbrk(c," \t");

	if (!c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	*c1 = '\0';

	c1++;

	while (*c1 && whitespace (*c1)) /* skip leading whitespace */
	    c1++;
	if (!*c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}

	if (0 == strcmp(c,"-")) {
	    result[result_len].match = NULL;
	    if (!load_charset_map_info(&result[result_len].charset,c1)) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeNotCharsetDeclaration,
				  "%30s: Not charset declaration: %.30s..."),
			  filename,c1);
		(*errors) ++;
		break;	    
	    }
	    result_len++;
	} else {
	    result[result_len].match = safe_strdup(c);

	    if (0 == strcmp(c1,"(unsupported)")) {
		/* Adding nameless unsupported charset */
		struct charcode_info new;
		charset_t ret;

		new.flags          = 0;
		new.map_info       = NULL;
		new.MIME_name      = NULL;
		new.charset_type   = &cs_unknown;
		new.subtype        = NULL;
		new.iso2022_info   = NULL;

		new.MIBenum        = 0;
		new.codeset        = NULL;

		ret = add_set(new);

		DPRINT(Debug,11,(&Debug,"Adding (unsupported) =%p\n",
				 ret));

		result[result_len].charset = ret;
	    } else if (!load_charset_map_info(&result[result_len].charset,c1))
		result[result_len].charset = 
		    MIME_name_to_charset(c1,CHARSET_create); 
	    
	    result_len++;
	}
	DPRINT(Debug,11,(&Debug,"Added map[%d] %-15s   %-15s  (flags %X)\n",
			 result_len-1,
			 result[result_len-1].match ?
			 result[result_len-1].match : "- (define)",
			 result[result_len-1].charset->MIME_name ?
			 result[result_len-1].charset->MIME_name :
			 "(unsupported -- no MIME name)",
			 result[result_len-1].charset->flags));

    }
    result[result_len].match   = NULL;
    result[result_len].charset = NULL;

    fclose(f);

    DPRINT(Debug,11,(&Debug,"load_locale_map: %s, result_len=%d\n",
	       filename,result_len));
    
    return result;
}

void dump_by_MIBenum(FILE *f)
{
    int class;

    fprintf(f,"# Start of MIBenum listing -------------------------------\n");
    for (class = 0; class < sizeof (MIBenum_CLASS) / sizeof (MIBenum_CLASS[0]); class++) {
	int j;
	fprintf(f,"# MIBenum %d-%d\n",
		class * MIBenum_class_divide ,
		class * MIBenum_class_divide + MIBenum_CLASS[class].MIBenum_array_len -1);

	for (j = 0; j < MIBenum_CLASS[class].MIBenum_array_len; j++) {

	    if (MIBenum_CLASS[class].MIBenum_array[j]) {
		fprintf(f,"-\t");
		dump_map_info(f,MIBenum_CLASS[class].MIBenum_array[j]);
		fputc('\n',f);
	    }
	}
    }
    fprintf(f,"# End of MIBenum listing ---------------------------------\n");
}

void dump_locale_map(f,map)
     FILE *f; 
     struct locale_map_item *map;
{
    int i;
    struct locale_map_item *ptr;

    for (ptr = map; ptr && ptr->charset; ptr++) {
	ptr->charset->flags &= ~SET_mark;
	ptr->charset->flags &= ~SET_printed;
	DPRINT(Debug,25,(&Debug, 
			"dump_locale_map: map[%d] %p - '%s' - flags=%X\n",
			 ptr-map,ptr->charset,
			 ptr->charset->MIME_name ? ptr->charset->MIME_name :"<none>",
			 ptr->charset->flags));
    }

    /* For precompiled sets we print only alias definations ... */
    for (i = 0; precompiled_sets[i].charset_type; i++) {
	precompiled_sets[i].flags |= SET_mark;
	DPRINT(Debug,25,(&Debug, 
			 "dump_locale_map: pre[%d] %p - '%s' - flags=%X\n",
			 i,&(precompiled_sets[i]),
			 precompiled_sets[i].MIME_name ? 
			 precompiled_sets[i].MIME_name :"<none>",
			 precompiled_sets[i].flags));
    }
    for (ptr = map; ptr && ptr->charset; ptr++) {
	if (0 != (ptr->charset->flags & SET_mark) &&
	    0 != (ptr->charset->flags & SET_havealias) &&
	    0 == (ptr->charset->flags & SET_printed)) {
	    fputs("-\t",f);
	    if (!ptr->charset->MIME_name) {
		if (ptr->charset->codeset)
		    elm_fprintf(f,FRM("codeset=%Q"),
				ptr->charset->codeset);
		else
		    fputs("# <No MIME name> ",f);
	    } else if (name_ok(ptr->charset->MIME_name)) {
		fputs(ptr->charset->MIME_name,f);
	    } else {
		elm_fprintf(f,FRM("%Q"),ptr->charset->MIME_name);
	    }
	    dump_alias_info(f,ptr->charset);
	    ptr->charset->flags |= SET_printed;
	    fputc('\n',f);
	    DPRINT(Debug,25,(&Debug, 
			     "dump_locale_map: map[%d] %p - '%s' - flags=%X\n",
			     ptr-map,ptr->charset,
			     ptr->charset->MIME_name ? ptr->charset->MIME_name :"<none>",
			     ptr->charset->flags));
	}
    }

    for (ptr = map; ptr && ptr->charset; ptr++) {
	if (ptr->charset->MIME_name ||
	    ptr->charset->codeset) {
	    if (0 == (ptr->charset->flags & SET_mark) &&
		0 == (ptr->charset->flags & SET_nodata) &&
		0 == (ptr->charset->flags & SET_printed)) {
		fputs("-\t",f);
		dump_map_info(f,ptr->charset);
		ptr->charset->flags |= SET_printed;
		fputc('\n',f);
	    }
	    ptr->charset->flags |= SET_mark;
	    DPRINT(Debug,25,(&Debug,"dump_locale_map: map[%d] %p - '%s' - flags=%X\n",
		       ptr-map,ptr->charset,
		       ptr->charset->MIME_name ? ptr->charset->MIME_name :"<none>",
		       ptr->charset->flags));
	}
    }

    for (ptr = map; ptr && ptr->charset; ptr++) {
	if (ptr->match) {
	    fputs(ptr->match,f);
	    fputc('\t',f);
	    if (!ptr->charset->MIME_name) {
		if (ptr->charset->codeset)
		    elm_fprintf(f,FRM("codeset=%Q"),
				ptr->charset->codeset);
		else
		    fputs("(unsupported)",f);
	    } else if ((0 != (ptr->charset->flags & SET_mark) ||
		      0 != (ptr->charset->flags & SET_nodata) ||
		      0 != (ptr->charset->flags & SET_printed)) && 
		     ptr->charset->MIME_name &&
		     NULL == strpbrk(ptr->charset->MIME_name,";\""))
		fputs(ptr->charset->MIME_name,f);
	    else
		dump_map_info(f,ptr->charset);
	    ptr->charset->flags |= SET_printed;
	    fputc('\n',f);
	    DPRINT(Debug,25,(&Debug,
			     "dump_locale_map: map[%d] %p - '%s' - flags=%X\n",
			     ptr-map,ptr->charset,
			     ptr->charset->MIME_name ? ptr->charset->MIME_name :"<none>",
			     ptr->charset->flags));

	}
    }
}

char * mime_parse_content_opts (str, walk)
     char *str;
     char **walk;
{
    /* A routine for parsing the options in a Content-Type: - like field.  
     * The important point here is to skip the semi-colon if it appears
     * inside of quotes.  This works sort of like strtok, except that
     * the token is already known.
     */
    char *ptr  = *walk;
    char *ret;
    int in_quote = 0;
    
    /* This is the initialization call */
    if (str) {
	DPRINT(Debug,20,(&Debug, 
			"mime_parse_content_opts: string=%s\n",
		   str));
	ptr = str;
    }    
    /* skip leading spaces */
    while (*ptr && isascii(*ptr) && isspace (*ptr))
	ptr++;

    if (*ptr == '\0') {
	DPRINT(Debug,20,(&Debug, "mime_parse_content_opts=NULL: EOS\n"));
	*walk = ptr;
	return NULL;
    }
    
    ret = ptr;
    while (*ptr) {
	if (*ptr == '\\' && in_quote) {		  
	    /* \ escapes next character  
	     * (not allowed outside of quotes) */
	    ptr++;
	    if (*ptr == '\0')
		break;
	} else if (*ptr == '\"') {
	    if (in_quote)
		in_quote = 0;
	    else
		in_quote = 1;
	}
	else if (! in_quote) {
	    if (*ptr == ';') {
		*ptr++ = '\0';
		DPRINT(Debug,20,(&Debug, "mime_parse_content_opts=%s (token)\n",
				 ret));
		*walk = ptr;
		return (ret);
	    }
	} 
	ptr++;
    }
    DPRINT(Debug,20,(&Debug, "mime_parse_content_opts=%s (token)\n",
		     ret));
    *walk = ptr;
    return (ret);
}

/* Used to get charset argument */
int mime_get_param (name, value, opts, size)
     char *value;
     CONST char *opts, *name;
     int size;
{
    char *c, tmp[VERY_LONG_STRING];
    int i = 0, quoted = FALSE, found = FALSE;
    
    char * WALK = NULL;

    value[0] = '\0';
    
    if (!opts) {
	DPRINT(Debug,20,(&Debug, 
			 "mime_get_param: name=\"%s\", opts=NULL\n",
			 name));
	return 0;
    }

    DPRINT(Debug,20,(&Debug, "mime_get_param: name=\"%s\", opts=\"%s\"\n",
		     name,opts));
    
    /* Make sure not to harm opts */
    strfcpy (tmp, opts, sizeof (tmp));
    rfc822_reap_comments (tmp, NULL, 0);
    
    c = tmp;
    while ((c = mime_parse_content_opts (c, &WALK)) != NULL && !found) {
	char * d  = strchr(c,'=');
	char * d2 = d;
	if (!d) {
	    c = NULL;
	    continue;    /* bad paramater */
	}
	while (d2 > c && (whitespace (*(d2-1))))
	    d2--;
	*d2 = '\0';
	
	while (*c && whitespace(*c))
	    c++;
	
	if (istrcmp (c, name) == 0) {
	    found = TRUE;
	    
	    c = d+1;
	    while (*c && whitespace(*c))
		c++;
	    if (*c == '"') {
		c++;
		quoted = TRUE;
	    }
	    /* Either look for a trailing quoted (if quoted==TRUE) or a SPACE */
	    while (*c && ((quoted && *c != '"') || (!quoted && *c != ' '))) {
		if (*c == '\\' && quoted) {
		    /* \ escapes next character */
		    c++;
		    if (!*c)
			break;
		}
		if (i >= size-1)
		    break;    /* Avoid buffer overflow */
		value[i++] = *c++;
	    }
	    value[i] = '\0';
	    break;
	}
	c = NULL;
    }
    
    DPRINT(Debug,20,(&Debug, 
		     "mime_get_param: found=%d, value=%s\n",found,value));
    
    return found;
}

/* Removes comments from string */
void rfc822_reap_comments (ptr, comments, size) 
     char *ptr, *comments;
     int size;
{
    char *w_ptr = ptr, *c_ptr = comments;
    int comment_level = 0, saved_level = 0;
    int in_quote = 0;
  
    while (*ptr) {
	if (*ptr == '\\' && (in_quote || comment_level > 0)) {		  
	    /* \ escapes next character  
	     * (not allowed outside of quotes or comments) */
	    ptr++;
	    if (*ptr == '\0')
		break;
	    if (comment_level > 0 && comments) {
		if (c_ptr < comments + size - saved_level -3) {
		    *c_ptr++ = '\\';
		    *c_ptr++ = *ptr;
		}
	    }
	    if (comment_level == 0) {
		*w_ptr++ = '\\';
		*w_ptr++ = *ptr;
	    }
	    ptr++;
	    continue;
	} else if (comment_level > 0) {
	    if (*ptr == ')')
		comment_level --;
	    if (*ptr == '(')
		comment_level ++;
	    if (comments && c_ptr < comments + size - saved_level -3) {
		*c_ptr++ = *ptr;
		saved_level = comment_level;
	    }
	} else if (*ptr == '\"') {
      if (in_quote)
	  in_quote = 0;
      else
	  in_quote = 1;
	} else if (!in_quote && *ptr == '(') {
	    comment_level ++;
	    if (comments && c_ptr < comments + size - saved_level -4) {
		if (c_ptr != comments)
		    *c_ptr++ = ' ';
		*c_ptr++ = *ptr;
		saved_level = comment_level;
	    }
	    *w_ptr++ = ' ';  /* RFC 822 (STD 11) says that 
				comments represents one space */
	}
	if (comment_level == 0 && (in_quote || *ptr != ')'))
	    *w_ptr++ = *ptr;
	ptr++;
    }
    while (comments && saved_level > 0) {
	*c_ptr++ = ')';
	saved_level--;
    }
    
    if (comments)
	*c_ptr = '\0';
    *w_ptr = '\0';
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
