static char rcsid[] = "@(#)$Id: id_phrase.c,v 1.3 2002/04/01 09:48:33 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.3 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"

DEBUG_VAR(Debug,__FILE__,"header");

void zero_id_phrase(P)
     struct id_phrase *P;
{
    P->id   = NULL;
    P->text = NULL;
}

void free_id_phrase(P)
     struct id_phrase *P;
{
    if (P->id)
	free(P->id);
    P->id = NULL;
    if (P->text)
	free_string(&(P->text));    
}

int check_8bit_id_phrase(P)
     struct id_phrase *P;
{
    if (P->id && check_8bit_str(P->id))
	return TRUE;
    if (P->text && !can_ascii_string(P->text))
	return TRUE;
    return FALSE;
}

int check_8bit_string(P)
     struct string *P;
{
    if (!P)
	return FALSE;
    if (!can_ascii_string(P))
	return TRUE;
    return FALSE;
}

int check_8bit_str (str)
     char *str;
{
  char *s;

  for (s = str;	*s; s++)
    if (*s & 0x80)
      return HAVE_8BIT;
  return 0;
}

void write_id_phrase_header (mailer,hdr_name,P, top_encoding, enmime, 
			     hdr_charset)
     out_state_t *mailer;
     char *hdr_name; 
     struct id_phrase * P;
     int top_encoding;
     int enmime;
     charset_t hdr_charset;
{
    long POS = out_state_ftell(mailer);

#define LEN (out_state_ftell(mailer)-POS)
#define WRAP print_EOLN(mailer,top_encoding), \
  (POS = out_state_ftell(mailer)), state_putc(' ',mailer)

    char *c;

    if (!P -> id  && !P -> text) {
	DPRINT(Debug,9,
	       (&Debug,"write_id_phrase_header: (%s) -- NONE\n",
		hdr_name));
	return;
    }

    state_puts(hdr_name,mailer);
    state_puts(": ",mailer);

    DPRINT(Debug,9,(&Debug,
		    "write_id_phrase_header: (enmime=%d) %s: \n",
		    enmime,hdr_name));

    DPRINT(Debug,9,(&Debug,
		    "write_id_phrase_header: %s:",hdr_name));

    if (P->id) {
	state_puts(P->id,mailer);
	DPRINT(Debug,9,(&Debug,
			"%s",P->id));
    }

    if (P->text) {
	int X = 0;

	if (string_len(P->text) &&
	    0x0020 == give_unicode_from_string(P->text,0)) {
	    
	    if ((LEN > 60)) {
		WRAP;
		DPRINT(Debug,9,(&Debug,
				"[WRAP]\n          "));
	    } else {
		state_putc(' ',mailer);
		DPRINT(Debug,9,(&Debug," "));
	    }
	    X = 1;
	}

	while (string_len(P->text) > X) {
	    int x = X;
	    struct string * temp = clip_from_string(P->text,&X,900);
	    char * s = string_to_hdr(HDR_PHRASE,temp,
				     hdr_charset,enmime);
	    DPRINT(Debug,9,(&Debug,"{phrase quoted=%s}\n",s));
	    DPRINT(Debug,9,(&Debug,"\n          [%d..%d] ",x,X));
	    
	    for (c = s; *c != '\0'; c++) {
		if (*c == '\n' || 
		    (whitespace(*c) && (LEN > 75))) {
		    WRAP;
		    DPRINT(Debug,9,(&Debug,"[WRAP]\n          "));
		} else {
		    state_putc(*c,mailer);
		    DPRINT(Debug,9,(&Debug,"%c",*c));
		}
	    }
	    
	    free(s);
	    free_string(&temp);
	}
    }
    print_EOLN(mailer,top_encoding);    
    DPRINT(Debug,9,(&Debug,"[EOLN]\n"));
}

void write_string_header (mailer,hdr_name,P, top_encoding, enmime, hdr_charset)
     out_state_t *mailer;
     char *hdr_name; 
     struct string * P;
     int top_encoding;
     int enmime;
     charset_t hdr_charset;
{
    long POS = out_state_ftell(mailer);

    char *c;
    int X = 0;

    state_puts(hdr_name,mailer);
    state_puts(": ",mailer);

    DPRINT(Debug,9, (&Debug, "write_string_header: (enmime=%d) %s: \n",
		     enmime,hdr_name));

    DPRINT(Debug,9, (&Debug, "write_string_header: %s:",hdr_name));

    while (string_len(P) > X) {
	int x = X;
	struct string * temp = clip_from_string(P,&X,900);
	char * s = string_to_hdr(HDR_TEXT,temp,
				 hdr_charset,enmime);
	DPRINT(Debug,9, (&Debug,"{text encoded=%s}\n",s));
	DPRINT(Debug,9, (&Debug,"\n          [%d..%d] ",x,X));
	
	for (c = s; *c != '\0'; c++) {
	    if (*c == '\n' || 
		(whitespace(*c) && (LEN > 75))) {
		WRAP;
	        DPRINT(Debug,9, (&Debug,"[WRAP]\n          "));
	    } else {
		state_putc(*c,mailer);
		DPRINT(Debug,9, (&Debug,"%c",*c));
	    }
	}
	
	free(s);
	free_string(&temp);
    }
    print_EOLN(mailer,top_encoding);    
    DPRINT(Debug,9, (&Debug,"[EOLN]\n"));
}

#undef LEN
#undef WRAP


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
