static char rcsid[] = "@(#)$Id: precompiled_sets.c,v 1.21 2003/08/31 18:05:20 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.21 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "defs.h"
#include "elmlib.h"
#include "cs_imp.h"

#define INVARIANT &precompiled_sets[0]
#define ASCII &precompiled_sets[1]
#define LATIN1 &precompiled_sets[2]
#define LATIN2 &precompiled_sets[3]       /* ISO-8859-2 */
#define LATIN5 &precompiled_sets[10]      /* ISO-8859-9 */

/* MIBenum and codeset from http://www.iana.org/assignments/character-sets  */

struct charcode_info precompiled_sets[] = {
    { &cs_iso646, &map_invariant, SET_valid,  "INVARIANT",  NULL, NULL, 29,   NULL },
    { &cs_ascii, &map_ascii, SET_valid,  "US-ASCII",
      INVARIANT, &(sets_iso_8859_X[0]),                                  3,   "ANSI_X3.4-1968" },
    { &cs_ascii, &map_latin1, SET_valid,  "ISO-8859-1", 
      ASCII, &(sets_iso_8859_X[1]),                                      4,   "ISO_8859-1:1987" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-2", 
      ASCII, &(sets_iso_8859_X[2]),                                      5,   "ISO_8859-2:1987" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-3", 
      ASCII, &(sets_iso_8859_X[3]),                                      6,   "ISO_8859-3:1988" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-4", 
      ASCII, &(sets_iso_8859_X[4]),                                      7,   "ISO_8859-4:1988" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-5", 
      ASCII, &(sets_iso_8859_X[5]),                                      8,   "ISO_8859-5:1988" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-6", 
      ASCII, &(sets_iso_8859_X[6]),                                      9,   "ISO_8859-6:1987" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-7",
      ASCII, &(sets_iso_8859_X[7]),                                     10,   "ISO_8859-7:1987" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-8", 
      ASCII, &(sets_iso_8859_X[8]),                                     11,   "ISO_8859-8:1988" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-9", 
      ASCII, &(sets_iso_8859_X[9]),                                     12,   "ISO_8859-9:1989" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-10", 
      ASCII, &(sets_iso_8859_X[10]),                                    13,   NULL },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-11", 
      ASCII, &(sets_iso_8859_X[11]),                                  2259,  "TIS-620" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-13", 
      ASCII, &set_latin7,                                              109,   NULL },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-14", ASCII, &set_latin8,  110,   NULL },
    { &cs_ascii, &map_885915, SET_valid,  "ISO-8859-15", 
      ASCII, &set_latin9,                                              111,   NULL },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-16", 
      ASCII, &set_iso885916,                                           112,   NULL },
    { &cs_ascii, &map_koi8r, SET_valid,  "KOI8-R", ASCII, NULL,       2084,  NULL },
    { &cs_ascii, &map_koi8u, SET_valid,  "KOI8-U", ASCII, NULL,       2088,  NULL },
    { &cs_ascii, &map_cp1251, SET_valid, "windows-1251", ASCII, NULL, 2251,  NULL },
    { &cs_ascii, &map_cp1252, SET_valid, "windows-1252", LATIN1,NULL, 2252,  NULL },
    { &cs_utf8,  &map_utf8,  SET_valid,  "UTF-8",  NULL, &set_utf8,    106,  NULL },
    { &cs_utf7,  &map_utf7,  SET_valid,  "UTF-7",  NULL, NULL,        1012,  NULL },

    /* Following variants are not really handled ....
       (different directonalities are not supported) */

    { &cs_ascii, NULL, SET_valid,  "ISO-8859-6-E", 
      ASCII, &(sets_iso_8859_X[6]),                                     81, "ISO_8859-6-E" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-6-I", 
      ASCII, &(sets_iso_8859_X[6]),                                     82, "ISO_8859-6-I" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-8-E", 
      ASCII, &(sets_iso_8859_X[8]),                                     84, "ISO_8859-8-E" },
    { &cs_ascii, NULL, SET_valid,  "ISO-8859-8-I", 
      ASCII, &(sets_iso_8859_X[8]),                                     85, "ISO_8859-8-E" },

    /* Following should be some kind special case */

    { &cs_unknown,  NULL,  SET_valid, "UNKNOWN-8BIT",  NULL,  NULL,   2079, NULL },
    
    
    /* Arbitary names from http://www.iana.org/assignments/character-sets    */

    /* Multibyte character set so cs_ascii does not work .... */
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               27,  "ISO-10646-UTF-1" },
    { &cs_iso646,   NULL,  SET_valid,  NULL,  NULL, NULL,  
                                                                        28,  "ISO_646.basic:1983" },
    { &cs_iso646,   NULL,  SET_valid,  "IRV",  NULL, NULL,              30,  "ISO_646.irv:1983" },    
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-GB", NULL,  NULL,   
                                                                        20,  "BS_4730" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               31,  "NATS-SEFI" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               32,  "NATS-SEFI-ADD" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               33,  "NATS-DANO" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,        NULL, NULL,         34,  "NATS-DANO-ADD" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-FI", NULL,  NULL,  
                                                                        35,  "SEN_850200_B" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-SE2", NULL,  NULL,  
                                                                        21,  "SEN_850200_C" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,        NULL, NULL,         36,  "KS_C_5601-1987" },
    { &cs_iso2022,  &map_ISO2022_ascii,  SET_valid,  "ISO-2022-KR", NULL, 
      &set_ISO2022KR,  37,  NULL }, /* RFC 1557 */
    { &cs_euc,      &map_EUC_ascii,      SET_valid,  "EUC-KR",    NULL,
      &set_EUCKR,  38,  NULL }, 
    { &cs_iso2022,  &map_ISO2022_ascii,  SET_valid,  "ISO-2022-JP", NULL, 
      &set_ISO2022JP,      39,  NULL }, /* RFC 1468 */
    { &cs_iso2022,  &map_ISO2022_ascii_latin1,  SET_valid,  "ISO-2022-JP-2", 
      NULL, &set_ISO2022JP2,    40,  NULL }, /* RFC 1554 */
    { &cs_iso2022,  &map_ISO2022_ascii,  SET_valid,  "ISO-2022-CN", NULL,  
      &set_ISO2022CN,     104,  NULL }, /* RFC 1922 */
    { &cs_iso2022,  &map_ISO2022_ascii,  SET_valid,  "ISO-2022-CN-EXT", NULL,
      &set_ISO2022CNEXT, 105,  NULL }, /* RFC 1922 */
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               41,  "JIS_C6220-1969-jp" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-JP", NULL,  &set_ISO646JP,        42,  "JIS_C6220-1969-ro" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-IT", NULL,  NULL,        22,  "IT" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-PT", NULL,  NULL,        43,  "PT" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-ES", NULL,  NULL,        23,  "ES" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,        NULL,  NULL,        44,  "greek7-old" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               45,  "latin-greek" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-DE", NULL,  NULL,        24,  "DIN_66003" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-FR1", NULL,  NULL,       46,  "NF_Z_62-010_(1973)" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               47,  "Latin-greek-1" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               48,  "ISO_5427" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               49,  "JIS_C6226-1978" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               50,  "BS_viewdata" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               51,  "INIS" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               52,  "INIS-8" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               53,  "INIS-cyrillic" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               54,  "ISO_5427:1981" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               55,  "ISO_5428:1980" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-CN", NULL,  NULL,        56,  "GB_1988-80" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               57,  "GB_2312-80" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-NO", NULL,  NULL,        25,  "NS_4551-1" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-NO2", NULL,  NULL,       58,  "NS_4551-2" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-FR", NULL,  NULL,        26,  "NF_Z_62-010" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               59,  "videotex-suppl" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-PT2", NULL,  NULL,       60,  "PT2" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-ES2", NULL,  NULL,       61,  "ES2" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-HU", NULL,  NULL,        62,  "MSZ_7795.3" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               63,  "JIS_C6226-1983" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               64,  "greek7" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               65,  "ASMO_449" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               66,  "iso-ir-90" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               67,  "JIS_C6229-1984-a" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-JP-OCR-B", NULL,  NULL,  68,  "JIS_C6229-1984-b" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               69,  "JIS_C6229-1984-b-add" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               70,  "JIS_C6229-1984-hand" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               71,  "JIS_C6229-1984-hand-add" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               72,  "JIS_C6229-1984-kana" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               73,  "ISO_2033-1983" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               74,  "ANSI_X3.110-1983" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               75,  "T.61-7bit" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               76,  "T.61-8bit" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               77,  "ECMA-cyrillic" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-CA", NULL,  NULL,        78,  "CSA_Z243.4-1985-1" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-CA2", NULL,  NULL,       79,  "CSA_Z243.4-1985-2" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               80,  "CSA_Z243.4-1985-gr" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               83,  "T.101-G2" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               86,  "CSN_369103" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-YU", NULL,  NULL,        87,  "JUS_I.B1.002" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               14,  "ISO_6937-2-add" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               88,  "IEC_P27-1" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               89,  "JUS_I.B1.003-serb" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               90,  "JUS_I.B1.003-mac" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               91,  "greek-ccitt" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-CU", NULL,  NULL,        92,  "NC_NC00-10:81" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               93,  "ISO_6937-2-25" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               94,  "GOST_19768-74" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               95,  "ISO_8859-supp" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               96,  "ISO_10367-box" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               97,  "latin-lap" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               98,  "JIS_X0212-1990" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-DK", NULL,  NULL,        99,  "DS_2089" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,              100,  "us-dk" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,              101,  "dk-us" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               15,  "JIS_X0201" },
    { &cs_iso646,   NULL,  SET_valid,  "ISO646-KR", NULL,  NULL,       102,  "KSC5636" },
    { &cs_ascii,    NULL,  SET_valid,  "DEC-MCS",  ASCII, NULL,            2008,  NULL },
    { &cs_ascii,    NULL,  SET_valid,  "hp-roman8" ,  ASCII, NULL,            2004,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2027,  "macintosh" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2030,  "IBM273" },
    { &cs_ascii,    NULL,  SET_valid,  "CP437", ASCII, NULL,          2011,  "IBM437" },
    { &cs_ascii,    NULL,  SET_valid,  "CP775", ASCII, NULL,          2087,  "IBM775" },
    { &cs_ascii,    NULL,  SET_valid,  "CP850", ASCII, NULL,          2009,  "IBM850" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2045,  "IBM851" },
    { &cs_ascii,    NULL,  SET_valid,  "CP852", ASCII, NULL,          2010,  "IBM852" },
    { &cs_ascii,    NULL,  SET_valid,  "CP855", ASCII, NULL,          2046,  "IBM855" },
    { &cs_ascii,    NULL,  SET_valid,  "CP857", ASCII, NULL,          2047,  "IBM857" },
    { &cs_ascii,    NULL,  SET_valid,  "CP860", ASCII, NULL,          2048,  "IBM860" },
    { &cs_ascii,    NULL,  SET_valid,  "CP861", ASCII, NULL,          2049,  "IBM861" },
    { &cs_ascii,    NULL,  SET_valid,  "CP862", ASCII, NULL,          2013,  "IBM862" },
    { &cs_ascii,    NULL,  SET_valid,  "CP863", ASCII, NULL,          2050,  "IBM863" },
    { &cs_ascii,    NULL,  SET_valid,  "CP864", ASCII, NULL,          2051,  "IBM864" },
    { &cs_ascii,    NULL,  SET_valid,  "CP865", ASCII, NULL,          2052,  "IBM865" },
    { &cs_ascii,    NULL,  SET_valid,  "CP866", ASCII, NULL,          2086,  "IBM866" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2053,  "IBM868" },
    { &cs_ascii,    NULL,  SET_valid,  "CP869", ASCII, NULL,          2054,  "IBM869" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2058,  "IBM891" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2059,  "IBM903" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2060,  "IBM904" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2063,  "IBM1026" },

    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2080,  "MNEMONIC" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2081,  "MNEM" },
    { &cs_unknown,  NULL,  SET_valid,  "VISCII",  NULL, NULL,         2082,   NULL },  /* rfc 1456 */
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2083,  "VIQR" }, /* rfc 1456 */

    { &cs_unknown,  NULL,  SET_valid,  "IBM00858",  NULL, NULL,       2089,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2101,  "Big5-HKSCS" },
    { &cs_utf7,     NULL,  SET_valid,  "UNICODE-1-1-UTF-7",NULL, NULL, 103,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               16,  "JIS_Encoding" },
    { &cs_unknown,  NULL,  SET_valid,  "Shift_JIS",  NULL, NULL,        17,  NULL },
    { &cs_euc,  &map_EUC_ascii,  SET_valid,  "EUC-JP",  
      NULL, &set_EUCJP,           18,             
      "Extended_UNIX_Code_Packed_Format_for_Japanese" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,               19,  
      "Extended_UNIX_Code_Fixed_Width_for_Japanese" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2005,  "Adobe-Standard-Encoding" },
    { &cs_ascii,    NULL,  SET_valid,  NULL,  ASCII, NULL,            2006,  "Ventura-US" },
    { &cs_ascii,    NULL,  SET_valid,  NULL,  ASCII, NULL,            2007,  "Ventura-International" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2012,  "PC8-Danish-Norwegian" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2014,  "PC8-Turkish" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2015,  "IBM-Symbols" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2016,  "IBM-Thai" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2017,  "HP-Legal" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2018,  "HP-Pi-font" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2019,  "HP-Math8" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2020,  "Adobe-Symbol-Encoding" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2021,  "HP-DeskTop" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2022,  "Ventura-Math" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2023,  "Microsoft-Publishing" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2024,  "Windows-31J" },

    /* This is really EUC-CN */ 
    { &cs_euc,      &map_EUC_ascii,  SET_valid,  "GB2312",  NULL, 
      &set_EUCCN,         2025,  "GB2312-1980" }, /* ASCII + GB 2312-80 */

    { &cs_unknown,  NULL,  SET_valid,  "Big5",  NULL, NULL,           2026,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1250", ASCII, NULL,     2250,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1253", ASCII, NULL,     2253,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1254", ASCII ,NULL,     2254,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1255", ASCII ,NULL,     2255,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1256", ASCII ,NULL,     2256,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1257", ASCII ,NULL,     2257,  NULL },
    { &cs_ascii,    NULL,  SET_valid, "windows-1258", ASCII ,NULL,     2258,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2085,  "HZ-GB-2312" },

    /* How this compares to windows-1252 ... ?? */
    { &cs_ascii, NULL, SET_valid, NULL, LATIN1, NULL,        2000, "ISO-8859-1-Windows-3.0-Latin-1" },
    { &cs_ascii, NULL, SET_valid, NULL, LATIN2, NULL,        2001, "ISO-8859-2-Windows-Latin-2" },
    { &cs_ascii, NULL, SET_valid, NULL, LATIN5, NULL,        2003, "ISO-8859-9-Windows-Latin-5" },

    /* Following are ebcdic character sets and therefore not suitable for ElmME+ use */
    
    { &cs_unknown,  NULL,  SET_valid,  "IBM037",  NULL, NULL,         2028,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM038",  NULL, NULL,         2029,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM274",  NULL, NULL,         2031,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM275",  NULL, NULL,         2032,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM277",  NULL, NULL,         2033,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM278",  NULL, NULL,         2034,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM280",  NULL, NULL,         2035,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM281",  NULL, NULL,         2036,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM284",  NULL, NULL,         2037,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM285",  NULL, NULL,         2039,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM297",  NULL, NULL,         2040,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM420",  NULL, NULL,         2041,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM423",  NULL, NULL,         2042,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM424",  NULL, NULL,         2043,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM500",  NULL, NULL,         2044,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM870",  NULL, NULL,         2055,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM871",  NULL, NULL,         2056,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM880",  NULL, NULL,         2057,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM905",  NULL, NULL,         2061,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM918",  NULL, NULL,         2062,   NULL },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2064,  "EBCDIC-AT-DE" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2065,  "EBCDIC-AT-DE-A" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2066,  "EBCDIC-CA-FR" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2067,  "EBCDIC-DK-NO" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2068,  "EBCDIC-DK-NO-A" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2069,  "EBCDIC-FI-SE" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2070,  "EBCDIC-FI-SE-A" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2071,  "EBCDIC-FR" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2072,  "EBCDIC-IT" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2073,  "EBCDIC-PT" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2074,  "EBCDIC-ES" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2075,  "EBCDIC-ES-A" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2076,  "EBCDIC-ES-S" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2077,  "EBCDIC-UK" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             2078,  "EBCDIC-US" },
    { &cs_unknown,  NULL,  SET_valid,  "IBM00924",  NULL, NULL,       2090,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01140",  NULL, NULL,       2091,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01141",  NULL, NULL,       2092,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01142",  NULL, NULL,       2093,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01144",  NULL, NULL,       2095,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01145",  NULL, NULL,       2096,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01146",  NULL, NULL,       2097,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01147",  NULL, NULL,       2098,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01148",  NULL, NULL,       2099,  NULL },
    { &cs_unknown,  NULL,  SET_valid,  "IBM01149",  NULL, NULL,       2100,  NULL },

    /* Following are NOT valid charsets for MIME */

    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1000,  "ISO-10646-UCS-2" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1001,  "ISO-10646-UCS-4" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1010,  "UNICODE-1-1" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1011,  "SCSU" },  /* depends ... */
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1013,  "UTF-16BE" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1014,  "UTF-16LE" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1015,  "UTF-16" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1002,  "ISO-10646-UCS-Basic" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1003,  "ISO-10646-Unicode-Latin1" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             0,     "ISO-10646-J-1" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1005,  "ISO-Unicode-IBM-1261" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1006,  "ISO-Unicode-IBM-1268" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1007,  "ISO-Unicode-IBM-1276" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1008,  "ISO-Unicode-IBM-1264" },
    { &cs_unknown,  NULL,  SET_valid,  NULL,  NULL, NULL,             1009,  "ISO-Unicode-IBM-1265" },

    { NULL, NULL, 0,       NULL, NULL, NULL }

};

CONST charset_t ASCII_SET     = ASCII;
CONST charset_t INVARIANT_SET = INVARIANT;
charset_t display_charset = ASCII;  /* current charset used by terminal */
charset_t system_charset  = ASCII;  /* the charset, the locale supports */

/* 'Raw' byte buffer -- treated as ascii -- used on buffering !!!!!!  */

static struct charcode_info raw_buffer = {
    &cs_ascii, &map_ascii, SET_valid,  "*RAW*", ASCII, &(sets_iso_8859_X[0]),  0, NULL
};
CONST charset_t RAW_BUFFER = &raw_buffer;

#define ARRSIZE(SET) sizeof SET / sizeof (struct charcode_info)


CONST int precompiled_size = ARRSIZE(precompiled_sets);
