static char rcsid[] = "@(#)$Id: terminal.c,v 1.37 2003/08/19 14:34:12 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.37 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "s_me.h"
#include "cs_imp.h"
#include "cs_terminal.h"

DEBUG_VAR(Debug,__FILE__,"charset");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

enum terminal_map_keyword { terminal_bad = -1, terminal_iso2022         = 0,
                            terminal_private = 1, terminal_iso2022_mb   = 2,
			    terminal_flag    = 3, terminal_iso2022_like = 4
};

enum terminal_flags { terminal_bad_flag    = -1,  
		      terminal_xterm_title = 1,
		      terminal_xwsh_title  = 2
};

struct terminal_map_item {
    char                    * match;
    enum terminal_map_keyword keyword;
    union {
	struct cset {
	    charset_t  charset;  
	    int        is_def;
	} iso2022;
        struct privset {
	    struct term_cond  * condition;
            charset_t           charset;
            unsigned char *     bytes;
	} private;
	enum terminal_flags flag;
	struct cset2 {
	    charset_t           charset;
	    struct setlist    * iso2022_info;   
	} iso2022_like;
    } value;
};

static enum terminal_map_keyword get_keyword P_((const char *word));

static enum terminal_map_keyword get_keyword(word)
     CONST char *word;
{
    if (0 == istrcmp(word,"ISO-2022"))
	return terminal_iso2022;
    if (0 == istrcmp(word,"ISO-2022-LIKE"))
	return terminal_iso2022_like;
    if (0 == istrcmp(word,"ISO-2022/DW"))
	return terminal_iso2022_mb;
    if (0 == istrcmp(word,"PRIVATE"))
	return terminal_private;
    if (0 == istrcmp(word,"FLAG"))
	return terminal_flag;
    
    return terminal_bad;
}

static char * sStr P_((unsigned char *p));
static char * sStr (p)
     unsigned char *p;
{
    return (char *)p;
}

struct term_cond {
    enum tc_type { tt_value, tt_simple_var, tt_equal } X;


    struct term_cond        * p1;
    struct term_cond        * p2;

    char                    * value;

};

static void zero_term_cond P_((struct term_cond *X));
static void zero_term_cond(X)
     struct term_cond *X;
{
    X->X      = tt_value;
    X->p1     = NULL;
    X->p2     = NULL;
    X->value  = NULL;
}

/* Not needed? */
static void free_term_cond P_((struct term_cond **X));
static void free_term_cond(X)
     struct term_cond **X;
{
    struct term_cond *A = *X; 
    if (A->p1)
	free_term_cond(&(A->p1));
    if (A->p2)
	free_term_cond(&(A->p2));

    if (A->value) {
	free(A->value);
	A->value = NULL;
    }

    free(A);
    A = NULL;
	
    *X = A;
}


enum tc_token_type { tc_left = '[', tc_rigth = ']', tc_equal = '=',
		     tc_quoted = '"', tc_plain = 'a', tc_simple_var = '$',
		     tc_none = 0 };

static enum tc_token_type tc_get_token P_((char **value, char **ptr));
static enum tc_token_type tc_get_token (value,ptr)
     char **value;
     char **ptr;
{
    enum tc_token_type r = tc_none;
    char *C = *ptr;

    *value = NULL;

    while (*C && whitespace (*C)) /* skip leading whitespace */
	C++;

    switch (*C) {
	int L;
    case '\0':    goto fail;
    case '[':     r = tc_left; C++;  break;
    case ']':     r = tc_rigth; C++; break;
    case '=':     r =  tc_equal; C++; break;
    case '"':
	L = len_next_part(C);
	*value = dequote_opt(C,L);
	C += L;
	r = tc_quoted;
	break;
    case '$':	
	L = strcspn(C+1," []=\"$/");
	if (1 == L) {
	    r = tc_none;
	    goto fail;
	}
	*value = safe_malloc(L+1);
	memcpy(*value,C+1,L);
	(*value)[L] = '\0';
	C += L +1;
	r = tc_simple_var;
	break;
    default:
	L = strcspn(C," []=\"$");
	*value = safe_malloc(L+1);
	memcpy(*value,C,L);
	(*value)[L] = '\0';
	C += L;
	r = tc_plain;
	break;
    }

 fail:
    *ptr = C;
    return r;
}


static struct term_cond * tc_get_expression P_((char              **ptr));
static struct term_cond * tc_get_expression(ptr)     
     char              **ptr;
{
    struct term_cond *r  = NULL;
    char *value = NULL;

    switch (tc_get_token(&value,ptr)) {
    case tc_plain:
	lib_error(CATGETS(elm_msg_cat, MeSet, MeTreatedAsQuoted,
			  "Unregonized token %s treated as quoted value"),
		  value); 
	/* FALLTHRU */
    case tc_quoted:
	r = safe_malloc(sizeof (*r));
	zero_term_cond(r);
	r->value = value;
	r->X     = tt_value;
	value = NULL;
	break;
    case tc_simple_var:
	r = safe_malloc(sizeof (*r));
	zero_term_cond(r);
	r->value = value;
	r->X     = tt_simple_var;
	value = NULL;
	break;

    }

 err:

    return r;
}

static struct term_cond * parse_condition P_((char              **ptr));
static struct term_cond * parse_condition(ptr)     
     char              **ptr;
{
    struct term_cond *r  = NULL;
    struct term_cond *r1 = NULL;
    struct term_cond *r2 = NULL;

    char *value = NULL;

    if (tc_left != tc_get_token(&value,ptr))
	goto err;
    r1 = tc_get_expression(ptr);
    if (!r1)
	goto err;
    if (tc_equal != tc_get_token(&value,ptr))
	goto err;
    r2 = tc_get_expression(ptr);
    if (!r2)
	goto err;
    if (tc_rigth != tc_get_token(&value,ptr))
	goto err;

    r = safe_malloc(sizeof (*r));
    zero_term_cond(r);

    r->X = tt_equal;
    r->p1 = r1;
    r->p2 = r2;

    r1 = NULL; r2 = NULL;

 err:
    if (value)
	free(value);
    if (r1) 
	free_term_cond(&r1);
    if (r2) 
	free_term_cond(&r2);

    return r;
}


static void tc_dump_expression P_((FILE *f, struct term_cond * C));

static char * tc_eval_expression P_((struct term_cond * C));
static char * tc_eval_expression(C)
     struct term_cond * C;
{
    char * r = NULL;
    

    switch (C->X) {
	char buffer[1024];
	int v;
    case tt_value:
	/* expand_path converts $VAR between / characters */
	v = expand_path(buffer,C->value,sizeof buffer);
	if (v > 0) {
	    DPRINT(Debug,9,(&Debug,"tc_eval_expression : \"%s\" expanded\n",
			    C->value));
	    r = safe_strdup(buffer);
	} else if (v == 0) {
	    DPRINT(Debug,9,(&Debug,"tc_eval_expression : \"%s\" copied\n",
			    C->value));
	    r = safe_strdup(C->value);       
	} else {
	    DPRINT(Debug,9,(&Debug,"tc_eval_expression : \"%s\" failed\n",
			    C->value));
	}

	DPRINT(Debug,9,(&Debug,"tc_eval_expression \"%s\" = %s\n",
			C->value,
			r ? r : "NULL"));

	break;
    case tt_simple_var:
	r = getenv(C->value);
	if (r)
	    r = safe_strdup(r);

	DPRINT(Debug,9,(&Debug,"tc_eval_expression $%s = %s\n",
			C->value,
			r ? r : "NULL"));

	break;
    default:
	DPRINT(Debug,8,(&Debug,"tc_eval_expression %d (bad)\n",C->X));
    }

    return r;
}

static void dump_condition P_((FILE *f, struct term_cond * C));

static int eval_condition P_((struct term_cond * C));
static int eval_condition(C)
     struct term_cond * C;
{
    int r = 0;
    char * v1 = NULL;
    char * v2 = NULL;

#ifdef DEBUG
    if (Debug.active >= 8) {

	FILE *F;

	DPRINT(Debug,8,(&Debug,"eval_condition   "));
	F = debug_to_FILE(&Debug);
	if (F) {
	    dump_condition(F,C);

	    fputc('\n',F);
	    fclose(F);
	}	
    } 
#endif

    if (C->p1)
	v1 = tc_eval_expression(C->p1);
    if (C->p2)
	v2 = tc_eval_expression(C->p2);


    switch (C->X) {
    case tt_equal:
	
	if (!v1 || !v2)
	    break;
	r = 0 == strcmp(v1,v2);

	DPRINT(Debug,8,(&Debug,"eval_condition  \"%s\" = \"%s\"\n",v1,v2));

	break;
    }

    if (v1)
	free(v1);
    if (v2)
	free(v2);
    DPRINT(Debug,8,(&Debug,"eval_condition=%d\n",r));
    return r;
}

static void tc_dump_expression(f,C)
     FILE *f;
     struct term_cond * C;
{
    switch (C->X) {
    case tt_value:
	elm_fprintf(f,FRM("%Q"),C->value);
	break;
    case tt_simple_var:
	fprintf(f,"$%s",C->value);
	break;
    default:
	fputc('?',f);
    }
}

static void dump_condition(f,C)
     FILE *f;
     struct term_cond * C;
{
    fputc('[',f);
    fputc(' ',f);
    
    if (C->p1)
	tc_dump_expression(f,C->p1);

    if (tt_equal == C->X)
	fputc('=',f);
    else
	fputc('?',f);

    if (C->p2)
	tc_dump_expression(f,C->p2);

    fputc(' ',f);
    fputc(']',f);
}

static struct {
    char                *flag;
    enum terminal_flags  flag_value;
} terminal_flag_list[] = {
    { "xterm-title" , terminal_xterm_title },
    { "xwsh-title"  , terminal_xwsh_title }

};
    

struct terminal_map_item * load_terminal_map(filename,errors)
     CONST char *filename;
     int *errors;
{
    struct terminal_map_item *result;
    int result_len = 0;
    FILE * f = fopen(filename,"r");
    int max_result = 0;
    int c;
    char buf[LONG_STRING];

    if (!f) {
 	int err = errno;
	DPRINT(Debug,2,(&Debug,"load_terminal_map: %s: %s\n",
			filename,error_description(err)));
	return NULL;
    }

    while(EOF != (c = fgetc(f)))
	if ('\n' == c)
	    max_result++;

    DPRINT(Debug,10,(&Debug,"load_terminal_map: %s, max_result=%d\n",
		     filename,max_result));
    
    if (!max_result) {
	fclose(f);
	return NULL;
    }
    rewind(f);
    
    result = safe_malloc((max_result +1) * sizeof (struct terminal_map_item));
    
    while (result_len < max_result &&
	   fgets(buf,sizeof buf, f) != NULL) {
	char * c = buf,*c1,*c2;
	int l1 = strlen(buf);
	enum terminal_map_keyword kw;

	if ('\n' == buf[l1 -1]) 
	    buf[l1 - 1] = '\0';
	else {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeTooLongLine,
			      "%30s: Too long line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;
	}
	l1--;

	while (l1-- > 0 && whitespace(buf[l1]))
	    buf[l1] = '\0';
	
	c = buf;
	while (*c && whitespace (*c)) /* skip leading whitespace */
	    c++;
	if ('#' == *c)
	    continue;
	if (!*c)
	    continue;

	c1 = strpbrk(c," \t");

	if (!c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	*c1 = '\0';
	
	c1++;

	while (*c1 && whitespace (*c1)) /* skip leading whitespace */
	    c1++;
	if (!*c1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	    

	c2 = strpbrk(c1," \t");

	if (!c2) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	*c2 = '\0';
	
	c2++;

	while (*c2 && whitespace (*c2)) /* skip leading whitespace */
	    c2++;
	if (!*c2) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	    
	kw = get_keyword(c1);

	if (terminal_bad == kw) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
			      "%30s: Bad line: %.30s..."),
		      filename,buf);
	    (*errors) ++;
	    break;	    
	}
	
	switch(kw) {
	    char * c3;
	    int size;
	    int i;
	case terminal_iso2022:
	case terminal_iso2022_mb:
	    result[result_len].keyword = kw;
	    
	    if (load_charset_map_info(&result[result_len].
				      value.iso2022.charset,c2))
		result[result_len].value.iso2022.is_def = 
		    ((result[result_len].value.iso2022.charset->flags &
		      SET_nodata) != 0);
	    else {
		result[result_len].value.iso2022.charset = 
		    MIME_name_to_charset(c2,0); 
		result[result_len].value.iso2022.is_def = 0;
	    }
	    
	    if (!result[result_len].value.iso2022.charset) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeUnknownTerminalCharset,
				  "Unknown terminal charset %s (file %s)"),
			  c2,filename);
		(*errors) ++;
		continue;
	    }

	    if (!result[result_len].value.iso2022.charset->map_info) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeCharsetNeedsMap,
				  "Charset %s needs map= defination (file %s)"),
			  result[result_len].value.iso2022.charset->MIME_name ?
			  result[result_len].value.iso2022.charset->MIME_name :
			  "<no MIME name>",
			  filename);
		(*errors) ++;
		continue;
	    }

	    if (!result[result_len].value.iso2022.charset->iso2022_info) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeCharsetNeedsIso2022,
				  "Charset %s needs iso2022 defination (file %s)"),
			  result[result_len].value.iso2022.charset->MIME_name ?
			  result[result_len].value.iso2022.charset->MIME_name :
			  "<no MIME name>",
			  filename);
		(*errors) ++;
		continue;
	    }

	    if (terminal_iso2022_mb == result[result_len].keyword &&
		0 == (CS_printable_len & charset_properties(result[result_len].value.iso2022.charset))) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeCharsetNotMB,
				  "Charset %s does not support /DW flag (file %s)"),
			  result[result_len].value.iso2022.charset->MIME_name ?
			  result[result_len].value.iso2022.charset->MIME_name :
			  "<no MIME name>",
			  filename);
		(*errors) ++;
		continue;

	    }

	    result[result_len].match = safe_strdup(c);
	    result_len++;
	    break;
	case terminal_iso2022_like: {
	    struct setlist     new_setlist;
	    int         setcount          = 0;

	    char * opt;
	    char * WALK;
	
	    iso2022_clear_setlist(&new_setlist);

	    result[result_len].keyword = kw;

	    c3 = strpbrk(c2," \t");

	    if (!c3) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		break;	    
	    }
	    *c3 = '\0';
	
	    c3++;

	    while (*c3 && whitespace (*c3)) /* skip leading whitespace */
		c3++;
	    if (!*c3) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		continue;    
	    }

	    result[result_len].value.iso2022_like.charset = 
		MIME_name_to_charset(c2,0); 
	    if (!result[result_len].value.iso2022_like.charset) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeUnknownTerminalCharset,
				  "Unknown terminal charset %s (file %s)"),
			  c2,filename);
		(*errors) ++;
		continue;
	    }

	    if (';' == *c3)
		c3++;

	    for (opt = mime_parse_content_opts(c3, &WALK); 
		 opt; 
		 opt = mime_parse_content_opts(NULL, &WALK)) {
		
		char * q = strchr(opt,'=');
		char *val; 

		if (!q) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				      "%30s: Bad line: %.30s..."),
			      filename,buf);
		    (*errors) ++;
		    continue;    		    
		}
		*q++ = '\0';
		val = dequote_opt(q,strlen(q));
		
		if (! parse_iso2022_specification(opt,val,
						  &setcount,
						  &new_setlist)) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				      "%30s: Bad line: %.30s..."),
			      filename,buf);
		    (*errors) ++;
		}

		free(val);
	    }
	    
	    if (0 == setcount) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		continue;
	    }
		
	    result[result_len].value.iso2022_like.iso2022_info =
		loc_setlist(new_setlist);
	    result[result_len].match = safe_strdup(c);
	    result_len++;

	}
	    break;
	case terminal_private:
	    result[result_len].keyword = kw;

	    c3 = strpbrk(c2," \t");

	    if (!c3) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		break;	    
	    }
	    *c3 = '\0';
	
	    c3++;

	    while (*c3 && whitespace (*c3)) /* skip leading whitespace */
		c3++;
	    if (!*c3) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				  "%30s: Bad line: %.30s..."),
			  filename,buf);
		(*errors) ++;
		continue;    
	    }
	    result[result_len].value.private.condition         = NULL;

	    if ('[' == *c3) {

		result[result_len].value.private.condition = parse_condition(&c3);
		if (!result[result_len].value.private.condition)
		    goto err_3;

		while (*c3 && whitespace (*c3)) /* skip leading whitespace */
		    c3++;
		if (!*c3) {
		err_3:
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeBadLine,
				      "%30s: Bad line: %.30s..."),
			      filename,buf);
		    (*errors) ++;
		    continue;    
		}
	    }

	    result[result_len].value.private.charset = 
		MIME_name_to_charset(c2,0); 
	    if (!result[result_len].value.private.charset) {
		lib_error(CATGETS(elm_msg_cat, MeSet, 
				  MeUnknownTerminalCharset,
				  "Unknown terminal charset %s (file %s)"),
			  c2,filename);
		(*errors) ++;
		continue;
	    }
	    size = strlen(c3);
	    result[result_len].value.private.bytes = safe_malloc(size+1);

	    if (!parse_gen_spec(c3,result[result_len].
				value.private.bytes,
				size)) {
		(*errors) ++;
		continue;
	    }
	    result[result_len].value.private.bytes[size] = '\0';
	    result[result_len].match = safe_strdup(c);
	    result_len++;
	    break;
	case terminal_flag:
	    result[result_len].keyword = kw;
	    result[result_len].value.flag    = terminal_bad_flag;

	    for (i = 0; 
		 i < sizeof terminal_flag_list / sizeof (terminal_flag_list[0]);
		 i++) {
		if (0 == strcmp(terminal_flag_list[i].flag,c2)) {
		    result[result_len].value.flag   = 
			terminal_flag_list[i].flag_value;
		    goto ok1;
		}
	    }

	    lib_error(CATGETS(elm_msg_cat, MeSet, 
			      MeUnknownTerminalFlag,
			      "Unknown terminal flag %s (file %s)"),
		      c2,filename);
	    (*errors) ++;
	    continue;
	    
	ok1:
	    result[result_len].match = safe_strdup(c);
	    result_len++;
	    break;
	}
    }

    result[result_len].match   = NULL;
    result[result_len].keyword = terminal_bad;

    fclose(f);

    DPRINT(Debug,10,(&Debug,
		     "load_terminal_map: %s, result_len=%d\n",
		     filename,result_len));
    
    return result;
}

static int name_ok P_((CONST char *name));
static int name_ok(name)
     CONST char *name;
{
    if ('\0' == *name)
	return 0;

    return strlen(name) == strspn(name,"ABCDEFGHIJKLMNOPQRSTUVXYZ-0123456789");
}


void dump_terminal_map(f,map)
     FILE *f; 
     struct terminal_map_item *map;
{
    struct terminal_map_item *ptr;
    
    for (ptr = map; ptr && ptr->match; ptr++) {
	switch(ptr->keyword) {
	    char *s;
	    int i;

	case terminal_iso2022:
	case terminal_iso2022_mb:
	    fputs(ptr->match,f);
	    fputc('\t',f);
	    if (terminal_iso2022_mb == ptr->keyword)
		fputs("ISO-2022/DW",f);
	    else
		fputs("ISO-2022",f);
	    fputc('\t',f);
	    if (ptr->value.iso2022.is_def)
		dump_map_info(f,ptr->value.iso2022.charset);
	    else {
		if (!ptr->value.iso2022.charset->MIME_name) {
		    fputs("# <No MIME name> ",f);
		} else if (name_ok(ptr->value.iso2022.charset->MIME_name)) {
		    fputs(ptr->value.iso2022.charset->MIME_name,f);
		} else {
		    elm_fprintf(f,FRM("%Q;!"),
				ptr->value.iso2022.charset->MIME_name);
		}
	    }
	    break;

	case terminal_iso2022_like:
	    fputs(ptr->match,f);
	    fputc('\t',f);
	    fputs("ISO-2022-LIKE",f);
	    fputc('\t',f);
	    fputs(ptr->value.iso2022_like.charset->MIME_name,f);
	    fputc('\t',f);
	    print_setlist(f,ptr->value.iso2022_like.iso2022_info);
	    break;

	case terminal_private:
	    fputs(ptr->match,f);
	    fputc('\t',f);
	    fputs("PRIVATE",f);
	    fputc('\t',f);
	    fputs(ptr->value.private.charset->MIME_name,f);
	    if (ptr->value.private.condition) {
		fputc('\t',f);
		dump_condition(f,ptr->value.private.condition);
	    }
	    fputc('\t',f);
	    s = iso2022_codestr(ptr->value.private.bytes,
				strlen(sStr(ptr->value.private.bytes)));
	    if (s) {
		fputs(s,f);
		free(s);
	    }
	    break;
	case terminal_flag:
	    fputs(ptr->match,f);
	    fputc('\t',f);
	    fputs("FLAG",f);
	    fputc('\t',f);
	    for (i = 0; 
		 i < sizeof terminal_flag_list / sizeof (terminal_flag_list[0]);
		 i++) {
		if (ptr->value.flag ==  terminal_flag_list[i].flag_value) {
		    fputs(terminal_flag_list[i].flag,f);
		    break;
		}
	    }
	    break;
	}
	fputc('\n',f);
    }
}

/* Returns first flags found for terminal from list or terminal_bad_flag, 
   flags[] is terminated with terminal_bad_flag

   Hopefully quite safe to be called from signal handler 
*/
static enum terminal_flags find_terminal_flag P_((const char *terminal,
						  enum terminal_flags flags[]));

static enum terminal_flags find_terminal_flag(terminal,flags)
     CONST char *terminal;
     enum terminal_flags flags[];
{

    int i,j;

    SIGDPRINT(Debug,9,(&Debug,"** Looking terminal %s (flags",
		       terminal));
    for (j = 0; terminal_bad_flag != flags[j]; j++) {
	SIGDPRINT(Debug,9,(&Debug," %d",flags[j]));
    }
    SIGDPRINT(Debug,9,(&Debug,")\n"));
    
    for (i =  0; i < 2; i++) {
	struct terminal_map_item *ptr = user_terminal_map;

	if (i)
	    ptr = system_terminal_map;

	while (ptr && ptr->match) {
	    char * c = NULL;

	    if (0 == strcmp(ptr->match,terminal) ||
		( (c = strchr(ptr->match,'*')) &&
		  *(c+1) == '\0' &&
		  0 == strncmp(ptr->match,terminal,c - ptr->match))) {


		if (ptr->keyword == terminal_flag) {
		    int j;

		    for (j = 0; terminal_bad_flag != flags[j]; j++) {

			if (ptr->value.flag == flags[j]) {
			    SIGDPRINT(Debug,9,(&Debug,
					       "-- Found flag match %s (flag %d)\n",
					       ptr->match,flags[j]));

			    return flags[j];
			}

		    }		   
		}

	    }
	    ptr++;
	}
    }

    return terminal_bad_flag; 
}
						  
						  

/* Hopefully quite safe to be called from signal handler */
static struct terminal_map_item * loc_info P_((const char *terminal, 
					       charset_t set));

static struct terminal_map_item * loc_info(terminal,set)
     CONST char *terminal; 
     charset_t set;
{
    int i;
    
    SIGDPRINT(Debug,9,(&Debug,"** Looking terminal %s (set %s)\n",
		       terminal,
		       set->MIME_name ? set->MIME_name : "<no MIME name>"));

    for (i =  0; i < 2; i++) {
	struct terminal_map_item *ptr = user_terminal_map;

	if (i)
	    ptr = system_terminal_map;

	while (ptr && ptr->match) {
	    char * c = NULL;

	    if (0 == strcmp(ptr->match,terminal) ||
		( (c = strchr(ptr->match,'*')) &&
		  *(c+1) == '\0' &&
		  0 == strncmp(ptr->match,terminal,c - ptr->match))) {

		if (ptr->keyword == terminal_iso2022 &&
		    ptr->value.iso2022.charset == set) {
		    SIGDPRINT(Debug,9,(&Debug,"-- Found ISO2022 match %s\n",
				       ptr->match));

		    return ptr;	    
		}

		if (ptr->keyword == terminal_iso2022_mb &&
		    ptr->value.iso2022.charset == set) {
		    
		    SIGDPRINT(Debug,9,(&Debug,"-- Found ISO2022 MB match %s\n",
				       ptr->match));

		    return ptr;	    
		}

		if (ptr->keyword == terminal_iso2022_like &&
		    ptr->value.iso2022_like.charset == set) {
		    
		    SIGDPRINT(Debug,9,(&Debug,"-- Found ISO2022-LIKE match %s\n",
				       ptr->match));

		    return ptr;	    
		}

		if (ptr->keyword == terminal_private &&
		    ptr->value.private.charset == set) {
		    if (ptr->value.private.condition) {
			if (!eval_condition(ptr->value.private.condition)) {
			    goto no_this;
			}
		    }

		    SIGDPRINT(Debug,9,(&Debug,"-- Found provate match %s\n",
				       ptr->match));

		    return ptr;	    
		}
	    }
	no_this:
	    ptr++;
	}
    }
    return NULL;
}

void terminal_can_switch(terminal,storage,n,max)
     CONST char *terminal;
     charset_t *storage;
     int *n;
     int max;
{

    int i;

    /* Can not switch to other sets if can not
       switch back to display_charset
    */
    if (!loc_info(terminal,display_charset))
	return;

    for (i =  0; i < 2; i++) {
	struct terminal_map_item *ptr = user_terminal_map;
	
	if (i)
	    ptr = system_terminal_map;
	
	while (ptr && ptr->match && *n < max) {
	    char * c = NULL;
	    int j;

	    SIGDPRINT(Debug,9,(&Debug,"-- Looking %s ...\n",
			       ptr->match));

	    if (0 == strcmp(ptr->match,terminal) ||
		( (c = strchr(ptr->match,'*')) &&
		  *(c+1) == '\0' &&
		  0 == strncmp(ptr->match,terminal,c - ptr->match))) {
	
		switch (ptr->keyword) {
		case terminal_iso2022:
		case terminal_iso2022_mb:
		    SIGDPRINT(Debug,8,(&Debug,
				       "-- Adding %s (%p) to possible sets\n",
				       ptr->value.iso2022.charset->MIME_name ?
				       ptr->value.iso2022.charset->MIME_name :
				       "<no MIME name>",
				       ptr->value.iso2022.charset
				       ));
		    storage[(*n)++] = ptr->value.iso2022.charset;
		    break;

		case terminal_iso2022_like:
		    SIGDPRINT(Debug,8,(&Debug,
				       "-- Adding %s (%p) to possible sets\n",
				       ptr->value.iso2022_like.charset->MIME_name ?
				       ptr->value.iso2022_like.charset->MIME_name :
				       "<no MIME name>",
				       ptr->value.iso2022_like.charset
				       ));
		    storage[(*n)++] = ptr->value.iso2022_like.charset;
		    break;

		case terminal_private:
		    if (ptr->value.private.condition) {
			if (!eval_condition(ptr->value.private.condition)) {
			    SIGDPRINT(Debug,8,(&Debug,
					       "-- NOT adding %s (%p) to possible sets (condition fail)\n",
					       ptr->value.private.charset->MIME_name ?
					       ptr->value.private.charset->MIME_name :
					       "<no MIME name>",
					       ptr->value.private.charset
					       ));

			    goto no_this;
			}
		    }

		    SIGDPRINT(Debug,8,(&Debug,
				       "-- Adding %s (%p) to possible sets\n",
				       ptr->value.private.charset->MIME_name ?
				       ptr->value.private.charset->MIME_name :
				       "<no MIME name>",
				       ptr->value.private.charset
				       ));

		    storage[(*n)++] = ptr->value.private.charset;
		    break;

		}

		for (j = 0; j < (*n) -1; j++)
		    if (storage[j] == storage[(*n)-1]) {
			SIGDPRINT(Debug,8,(&Debug,
					   "-- Removing duplicate %s from possible sets\n",
					   storage[(*n)-1]->MIME_name ?
					   storage[(*n)-1]->MIME_name :
					   "<no MIME name>"));
			(*n)--;  /* Remove duplicate */  
			break;
		    }
	    }
	no_this:
	    ptr++;
	}
    }
}

int terminal_can_switch_to(terminal,set,silent)
     CONST char *terminal; 
     charset_t set;
     int silent;
{

    struct terminal_map_item * a = loc_info(terminal,set);

    if (!a) {
	if (!silent)
	    lib_error(CATGETS(elm_msg_cat, MeSet, 
			      MeNoInformationToSwitchTerminal,
			      "No information to switch %s terminal to %s charset!"), 
		      terminal,
		      set->MIME_name ? set->MIME_name : "<no MIME name>");
		      
	return 0;
    }
    
    a = loc_info(terminal,display_charset);

    if (!a) {
	if (!silent)
	    lib_error(CATGETS(elm_msg_cat, MeSet, 
			      MeNoInformationToSwitchBack,
			      "No information to switch %s terminal back to %s charset!"), 
		      terminal,
		      display_charset->MIME_name ? display_charset->MIME_name :
		      "<no MIME name>");
	return 0;
    }
    return 1;
}

void debug_display_settings(p,sig,stream)
     struct display_settings *p;
     int sig;
     int stream;
{
    int i;
    int corrupted = 0;

    if (stream) {
	SIGDPRINT(Debug,8,(&Debug,
			   "Stream sets: (%d sets)\n",
			   p->set_count));
    } else {
	SIGDPRINT(Debug,8,(&Debug,
			   "Terminal sets: (%d sets)\n",
			   p->set_count));
    }
    for (i = 0; i < p->set_count; i++) {
	if (i >= sizeof (p->sets) /
	    sizeof (p->sets[0])) {
	    corrupted++;
	    break;
	}

	SIGDPRINT(Debug,8,(&Debug,"    set [%d]:  bank=",i));
	switch(p->sets[i].bank) { 
	case bank_unspecified:  SIGDPRINT(Debug,8,(&Debug, "(none)")); break;
	default:                SIGDPRINT(Debug,8,(&Debug, "G%d   ",
						p->sets[i].bank)); break;
	}

	SIGDPRINT(Debug,8,(&Debug," type="));
	switch(p->sets[i].type) { 
	case iso2022_other: SIGDPRINT(Debug,8,(&Debug,"other")); break;
	case iso2022_94:    SIGDPRINT(Debug,8,(&Debug,"94   ")); break;
	case iso2022_96:    SIGDPRINT(Debug,8,(&Debug,"96   ")); break;
	case iso2022_94x94: SIGDPRINT(Debug,8,(&Debug,"94x94")); break;	
	case iso2022_96x96: SIGDPRINT(Debug,8,(&Debug,"96x96")); break;	
	default:            SIGDPRINT(Debug,8,(&Debug,"?????")); 
	    corrupted++; break;
	}
	{
	    int k;
	    SIGDPRINT(Debug,8,(&Debug," bytes="));

	    for (k = 0; k < sizeof p->sets[i].bytes && p->sets[i].bytes[k]; k++) {
		int val = (unsigned char) p->sets[i].bytes[k];

		if (k > 0) {
		    SIGDPRINT(Debug,8,(&Debug," "));
		}

		SIGDPRINT(Debug,8,(&Debug,"%d/%d",val/16,val%16));
	    }
	}
	SIGDPRINT(Debug,8,(&Debug," width=%d\n",p->width[i]));
    }

    for (i = 0; i < 4; i++) {
	SIGDPRINT(Debug,8,(&Debug,"    bank G%d=",i));;
	switch(p->bank[i]) { 
	case -1:    SIGDPRINT(Debug,8,(&Debug,"(none)")); break;
	default:    SIGDPRINT(Debug,9,(&Debug,"set %d",p->bank[i]));
	    if (p->bank[i] < 0 || p->bank[i] >= p->set_count) {
		corrupted++;
		SIGDPRINT(Debug,8,(&Debug,"   ????"));
	    }
	    break;
	}
	SIGDPRINT(Debug,8,(&Debug,"\n"));
    }
    SIGDPRINT(Debug,8,(&Debug,"       left="));;
    switch(p->current_L) {
    case bank_unspecified:  SIGDPRINT(Debug,8,(&Debug,"(none)")); break;
    default:                SIGDPRINT(Debug,8,(&Debug,"G%d   ",
					    p->current_L)); break;
    }
    SIGDPRINT(Debug,8,(&Debug,"\n"));
    SIGDPRINT(Debug,8,(&Debug,"      right="));;
    switch(p->current_R) {
    case bank_unspecified:  SIGDPRINT(Debug,8,(&Debug,"(none)")); break;
    default:                SIGDPRINT(Debug,8,(&Debug,"G%d   ",
					    p->current_R)); break;
    }
    SIGDPRINT(Debug,8,(&Debug,"\n"));

    if (corrupted) {
	if (stream)
	    panic("STRING PANIC",__FILE__,__LINE__,
		  "debug_display_settings",
#ifdef DEBUG
		  Debug.active >= 8 ? 
		  "Corrupted stream sets information -- see debugfile" :
#endif
		  "Corrupted stream sets information",sig);
	else
	    panic("STRING PANIC",__FILE__,__LINE__,
		  "debug_display_settings",
#ifdef DEBUG
		  Debug.active >= 8 ? 
		  "Corrupted terminal sets information -- see debugfile" :
#endif
		  "Corrupted terminal sets information",sig);
    }
}

static void debug_charset_sets P_((struct setlist  *p));
static void debug_charset_sets(p)
     struct setlist  *p;
{
    int i;

    DPRINT(Debug,8,(&Debug,"(output) charset sets: \n"));
    for (i = 0; 
	 i < sizeof (p->sets)  / sizeof (p->sets[0]) && p->sets[i]; 
	 i++) {
	char *s;
	DPRINT(Debug,8,(&Debug,"    set [%d]:  bank=",i));
	switch(p->sets[i]->bank) { 
	case bank_unspecified:  DPRINT(Debug,8,(&Debug,"(none)")); break;
	default:                DPRINT(Debug,8,(&Debug,"G%d   ",
						p->sets[i]->bank)); break;
	}
	DPRINT(Debug,8,(&Debug," type="));
	switch(p->sets[i]->type) { 
	case iso2022_other: DPRINT(Debug,8,(&Debug,"other")); break;
	case iso2022_94:    DPRINT(Debug,8,(&Debug,"94   ")); break;
	case iso2022_96:    DPRINT(Debug,8,(&Debug,"96   ")); break;
	case iso2022_94x94: DPRINT(Debug,8,(&Debug,"94x94")); break;	
	case iso2022_96x96: DPRINT(Debug,8,(&Debug,"96x96")); break;	
	default:            DPRINT(Debug,8,(&Debug,"?????")); break;
	}
	s = iso2022_codestr(p->sets[i]->bytes, sizeof p->sets[i]->bytes);
	DPRINT(Debug,8,(&Debug," bytes=%s\n",s));
	free(s);
    }    
    DPRINT(Debug,8,(&Debug," left="));
    switch(p->initial_L) { 
    case bank_unspecified:  DPRINT(Debug,8,(&Debug,"(none)")); break;
    default:                DPRINT(Debug,8,(&Debug,"G%d   ",
					    p->initial_L)); break;
    }
    DPRINT(Debug,8,(&Debug," rigth="));
    switch(p->initial_R) { 
    case bank_unspecified:  DPRINT(Debug,8,(&Debug,"(none)")); break;
    default:                DPRINT(Debug,8,(&Debug,"G%d   ",
					    p->initial_R)); break;
    }
    DPRINT(Debug,8,(&Debug,"\n"));
    for (i = 0; 
	 i < sizeof (p->initial_bank)  / sizeof (p->initial_bank[0]); 
	 i++) {
	DPRINT(Debug,8,(&Debug," initial bank G%d=",i));
	if (-1 == p->initial_bank[i]) {
	    DPRINT(Debug,8,(&Debug,"(none)\n"));
	} else {
	    DPRINT(Debug,8,(&Debug,"set %d",p->initial_bank[i]));
	    if (p->initial_bank[i] < 0 ||
		p->initial_bank[i] >=  sizeof (p->sets)  / sizeof (p->sets[0]) || 
		!p->sets[p->initial_bank[i]]) {
		DPRINT(Debug,8,(&Debug,"   ?????"));
	    }
	    DPRINT(Debug,8,(&Debug,"\n"));
	}
    }
}

static void reset_display_settings P_((struct display_settings *p));
static void reset_display_settings(p)
     struct display_settings *p;
{
    int i;
    
    /* defined in hdrs/defs.h */
    bzero((void *)p,sizeof (struct display_settings));
    
    p->magic              = DISPLAY_STATE_magic;
    p->set_count          = 0;    
    p->bank[0]            = -1;
    p->bank[1]            = -1;
    p->bank[2]            = -1;
    p->bank[3]            = -1;
    p->current_L          = bank_unspecified;
    p->current_R          = bank_unspecified;

    if (sizeof (p->width) / sizeof (p->width[0]) != 
	sizeof (p->sets) / sizeof (p->sets[0]))
	panic("STRING PANIC",__FILE__,__LINE__,
	      "reset_display_settings",
	      "width and sets array do not have save number of elemens",0);
    
    for (i = 0; i < sizeof (p->width) / sizeof (p->width[0]); i++)
	p->width[i] = -1;
}

void free_terminal_info(p)
     struct display_settings **p;
{
    if (*p) {
	if (DISPLAY_STATE_magic != (*p)->magic) 
	    panic("STRING PANIC",__FILE__,__LINE__,
		  "free_terminal_info",
		  "Bad magic number",0);

	(*p)->magic = 0;   /* Not valid */
	
	free(*p);
	*p = NULL;
    }
}

/* May be called from signal handler -- on that situation use
   buffer specified as argument -- otherwise result is malloced
*/

char * iso2022_change_helper_1(terminal_info,ID,setnum,buffer,size)
     screen_info_p terminal_info;
     struct iso2022_setid ID;
     int setnum;
     char * buffer; 
     int size;
{
    char * ret = NULL;
    int maybe_signal = buffer != NULL;

    if (maybe_signal) {
	SIGDPRINT(Debug,9,(&Debug,
			   "iso2022_change_helper_1: buffer=%p size=%d\n",
			   buffer,size));
    }

    if (ID.bank != bank_unspecified) {
	if (-1 == terminal_info->bank[ID.bank]) {	       	    
	    terminal_info->bank[ID.bank] = setnum;
	    
	    if (maybe_signal) {
		if (!ret) 
		    ret = iso2022_setid_stream(ID,buffer,size);
		else {
		    char buffer1[128];
		    char *c = iso2022_setid_stream(ID,buffer1,sizeof buffer1);
		    ret = strfcat(buffer,c,size);
		}
	    } else {
		char *c = iso2022_setid_stream(ID,NULL,0);
		ret = strmcat(ret,c);
		free(c);
	    }

	} else if (terminal_info->bank[ID.bank] != setnum) {
	    SIGDPRINT(Debug,4,(&Debug,
			    "terminal_switch_to [iso2022_change_helper_1]: Several assignments for bank G%d\n",
			    ID.bank));
	    SIGDPRINT(Debug,8,(&Debug,
			    "     Assigned to set %d -- another candinate set %d\n",
			    terminal_info->bank[ID.bank],setnum));
	}


	/* Initial bank settinsg do not set left and right so
	   set them here if them are not set ion iso2022_info */

	if (terminal_info->bank[ID.bank] == setnum) {

	    if (terminal_info->current_L  == bank_unspecified &&
		ID.bank    == bank_G0) {
		terminal_info->current_L  = bank_G0;
		
		if (maybe_signal) {
		    if (!ret)
			ret = lock_shift(0,terminal_info->current_L,buffer,
					 size);
		    else {
			char buffer1[10];
			char * c = lock_shift(0,terminal_info->current_L,
					      buffer1,sizeof buffer1);
			ret = strfcat(buffer,c,size);
		    }
		    
		} else {
		    char *c = lock_shift(0,terminal_info->current_L,NULL,0);
		    ret = strmcat(ret,c);
		    free(c);
		}
	    }
	    
	    if (/* For ISO-8859-*           charsets */
		terminal_info->current_R  == bank_unspecified &&
		ID.type    == iso2022_96                      ||
		/* For EUC-*                charsets */
		terminal_info->current_R  == bank_unspecified &&
		ID.type    == iso2022_94x94                   &&
		ID.bank    == bank_G1
		) {

		terminal_info->current_R  = ID.bank;

		if (maybe_signal) {
		    if (!ret)
			ret = lock_shift(1,terminal_info->current_R,
					 buffer,size);
		    else {
			char buffer1[10];
			char *c =  lock_shift(1,terminal_info->current_R,
					      buffer1,sizeof buffer1);
			ret = strfcpy(buffer,c,size);
		    }
		} else {
		    char *c = lock_shift(1,terminal_info->current_R,NULL,0);
		    ret = strmcat(ret,c);
		    free(c);
		}
	    }	    	    	   
	}

    } else if (ID.type == iso2022_other) {
	ret = iso2022_setid_stream(ID,buffer,size);

    }

    if (maybe_signal && ret && ret != buffer)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "iso2022_change_helper_1","buffer != ret when called from signal",1);

    return ret;
}

/* Result is MALLOCed */
char * terminal_charset_post_init(terminal_info,iso2022_info)
     screen_info_p terminal_info;
     struct setlist          *   iso2022_info;
{
    char * ret = NULL;
    int max,k;
    int i;

    /* FIXME: Usage of this function on string.c unclear!!! */

    static screen_info_p    LAST_COMPLAIN   = NULL;
    static struct setlist * LAST_COMPLAIN_b = NULL;        


    if (DISPLAY_STATE_magic != terminal_info->magic)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "terminal_charset_post_init",
	      "Bad terminal_info (bad magic)",0);

    /* Only check that terminal_info includes same sets
       (on beginning of list) than iso2022_info */
    
    for (max = 0; max < sizeof (iso2022_info->sets)  / 
	     sizeof (iso2022_info->sets[0]) &&
	     iso2022_info->sets[max];
	 max++) {
	
	if (max >= terminal_info->set_count ||
	    terminal_info->sets[max].type != iso2022_info->sets[max]->type ||
	    0 != strncmp(sStr(terminal_info->sets[max].bytes),
			 sStr(iso2022_info->sets[max]->bytes),
			 sizeof (terminal_info->sets[max].bytes))  ||
	    (bank_unspecified           != iso2022_info->sets[max]->bank &&
	     terminal_info->sets[max].bank != iso2022_info->sets[max]->bank)) {
	    if (LAST_COMPLAIN   != terminal_info ||
		LAST_COMPLAIN_b != iso2022_info) {
		LAST_COMPLAIN   = terminal_info;
		LAST_COMPLAIN_b = iso2022_info;
		DPRINT(Debug,8,(&Debug,
				"terminal_charset_post_init: Output charset not current terminal charset\n"));
		DPRINT(Debug,8,(&Debug,
				"            (perhaps subset) -- can't really post init terminal charset\n"));
		if (max >= terminal_info->set_count) {
		    DPRINT(Debug,8,(&Debug,
				    "             Only %d sets on terminal... \n",
				    terminal_info->set_count));
		} else {
		    DPRINT(Debug,8,(&Debug,
				    "             [%d] -- bank G%d on terminal set not match to output charset\n",
				    max, terminal_info->sets[max].bank));
		}
		debug_display_settings(terminal_info,0,0);
		debug_charset_sets(iso2022_info);
	    }
	    return NULL;
	}
    }
    LAST_COMPLAIN   = NULL;
    LAST_COMPLAIN_b = NULL;

    /* Remove assignments which does not belong to iso2022_info */
    for (k = 0; k < 4; k++) {
	if (terminal_info->bank[k] >= max) {
	    if (k == terminal_info->current_L)
		terminal_info->current_L = bank_unspecified;
	    if (k == terminal_info->current_R)
		terminal_info->current_R = bank_unspecified;
	    terminal_info->bank[k] = -1;	   
	}
    }
    
    if (bank_unspecified != iso2022_info->initial_L &&
	terminal_info->current_L != iso2022_info->initial_L) {
	char *c = lock_shift(0,terminal_info->current_L,NULL,0);
	ret = strmcat(ret,c);
	free(c);
    }

    if (bank_unspecified != iso2022_info->initial_R &&
	terminal_info->current_R != iso2022_info->initial_R) {
	char *c = lock_shift(1,terminal_info->current_R,NULL,0);
	ret = strmcat(ret,c);
	free(c);
    }

    for (i = 0; 
	 i < sizeof (iso2022_info->initial_bank) / 
	     sizeof (iso2022_info->initial_bank[0]);
	 i++) {
	int x = iso2022_info->initial_bank[i];
		

	if (-1 != x) {
	    struct iso2022_setid ID;

	    if (x < 0 ||
		x >= sizeof (iso2022_info->sets)  /
		sizeof (iso2022_info->sets[0]) ||
		! iso2022_info->sets[x]) 
		panic("STRING PANIC",__FILE__,__LINE__,
		      "terminal_charset_post_init",
		      "Bad initial_bank (set)",
		      0);
	       
	    ID = * (iso2022_info->sets[x]);
	    
	    if (ID.bank != i)
		panic("STRING PANIC",__FILE__,__LINE__,
		      "terminal_charset_post_init",
		      "Bad initial_bank (bank number)",
		      0);
	    
	    set_initial_bank(&ret,ID,terminal_info,NULL,0,0);

	}
    }


    for (k = 0; k < max; k++) {
	char * c;
	struct iso2022_setid ID = terminal_info->sets[k];

	c=iso2022_change_helper_1(terminal_info,ID,k,NULL,0);
	if (c) {
	    ret = strmcat(ret,c);
	    free(c);
	}
    }

    if (ret)
	debug_display_settings(terminal_info,0,0);

    return ret;
}

screen_info_p create_terminal_info(void)
{
    screen_info_p terminal_info;

    terminal_info = safe_malloc(sizeof (struct display_settings));

    reset_display_settings(terminal_info);

    return terminal_info;
}

void iso2022_setid_select_bank (ID,terminal_info)
     struct iso2022_setid *ID;
     screen_info_p terminal_info;
{
    if (ID->bank == bank_unspecified) {
	switch(ID->type) {
	case iso2022_94:
	    ID->bank = bank_G0;
	    break;

	case iso2022_94x94:         
	    ID->bank = bank_G1;          
	    break;
	    /* We use bank G2 to compatible of VT3xx terminals */
	case iso2022_96:  
	case iso2022_96x96:         /* not perhaps exists */
	    if (-1 == terminal_info->bank[bank_G2])
		ID->bank = bank_G2;
	    else
		ID->bank = bank_G3;
	}
    }
}

static void change_helper_2 P_((char **ret,struct iso2022_setid ID,
				screen_info_p terminal_info,   
				char * buffer,
				int size, 
				int maybe_signal,
				int setpos
				));

static void change_helper_2(ret,ID,terminal_info,buffer,size,maybe_signal,setpos)
     char **ret;
     struct iso2022_setid ID;
     screen_info_p terminal_info;   
     char * buffer; 
     int size;
     int maybe_signal;
     int setpos;
{   
    if (maybe_signal) { 
	if (!*ret)
	    *ret = 
		iso2022_change_helper_1(terminal_info,ID,
					setpos, 
					buffer,size);
	else {
	    char buffer1[100];
	    char *c = 
		iso2022_change_helper_1(terminal_info,ID,
					setpos, 
					buffer1,sizeof buffer1);
	    if (c) 
		*ret = strfcat(buffer,c,size);
	}

    } else {
	char *c = 
	    iso2022_change_helper_1(terminal_info,ID,
				    setpos, 
				    NULL,0);
	if (c) {
	    *ret = strmcat(*ret,c);
	    free(c);
	}
    }
}

int iso2022_give_setpos(ID,output_state)
     struct iso2022_setid   *ID; 
     screen_info_p  output_state;
{
    int setpos = -1;
    int i;

    iso2022_setid_select_bank (ID,output_state);

    for (i = 0; i < output_state->set_count; i++) {
	if (same_setid(output_state->sets[i],*ID))
	    setpos = i;
    }

    return setpos;
}

/* Return setnumber from terminal info */
int set_initial_bank(ret,ID,terminal_info,buffer,size,maybe_signal)
     char **ret;
     struct iso2022_setid ID;
     screen_info_p terminal_info;
     char * buffer;
     int size;
     int maybe_signal;
{  
    int setpos = -1;

    if (ID.bank < 0 || ID.bank >= ISO2022_BANK_NUM)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "set_initial_bank",
	      "Bad bank",maybe_signal);

    setpos = iso2022_give_setpos(&ID,terminal_info);

    if (-1 == setpos) {
	if (terminal_info->set_count <
	    sizeof (terminal_info->sets) /
	    sizeof (terminal_info->sets[0])) {
	    setpos = terminal_info->set_count++;
	} else {
	    setpos = ID.bank;
	    DPRINT(Debug,7,(&Debug,
			    "set_initial_bank: Too many sets on output state, replacing set %d\n",
			    setpos));
	} 
    
	terminal_info->sets[setpos]  = ID;
	terminal_info->width[setpos] = -1;           /* Initially no with */
    }

    terminal_info->bank[ID.bank] = setpos;
    if (maybe_signal) {
	if (!*ret) 
	    *ret = iso2022_setid_stream(ID,buffer,size);
	else {
	    char buffer1[128];
	    char *c = iso2022_setid_stream(ID,buffer1,sizeof buffer1);
	    *ret = strfcat(buffer,c,size);
	}
    } else {
	char *c = iso2022_setid_stream(ID,NULL,0);
	*ret = strmcat(*ret,c);
	free(c);
    }

    return setpos;
}


static void strXcat P_((char **ret,char *val, 
			int maybe_signal, char *buffer, int size));
static void strXcat(ret,val,maybe_signal,buffer,size)
     char **ret;
     char *val; 
     int maybe_signal; 
     char *buffer; 
     int size;
{
    
    if (maybe_signal) {
	if (!*ret)
	    *ret = strfcpy(buffer,val,size);
	else
	    *ret = strfcat(buffer,val,size);
    } else
	*ret = strmcat(*ret,val);
}

static void change_helper_3 P_((char **RET,struct setlist * iso2022_info,
				screen_info_p terminal_info,   
				char * buffer,
				int size, 
				int maybe_signal,
				int MB));

static void change_helper_3(RET,iso2022_info,terminal_info,buffer,size,
			    maybe_signal,MB)
     char **RET;
     struct setlist * iso2022_info;
     screen_info_p terminal_info;   
     char * buffer;
     int size; 
     int maybe_signal;
     int MB;
{
    char * ret = *RET;
    int i;    
	    
    if (bank_unspecified != iso2022_info->initial_L) {
	terminal_info->current_L = iso2022_info->initial_L;

	if (maybe_signal) {
	    if (!ret)
		ret = lock_shift(0,terminal_info->current_L,buffer,
				 size);
	    else {
		char buffer1[10];
		char * c = lock_shift(0,terminal_info->current_L,
				      buffer1,sizeof buffer1);
		ret = strfcat(buffer,c,size);
	    }
		    
	} else {
	    char *c = lock_shift(0,terminal_info->current_L,NULL,0);
	    ret = strmcat(ret,c);
	    free(c);
	}	       
    }
    
    if (bank_unspecified != iso2022_info->initial_R) {
	terminal_info->current_R = iso2022_info->initial_R;
		
	if (maybe_signal) {
	    if (!ret)
		ret = lock_shift(1,terminal_info->current_R,
				 buffer,size);
	    else {
		char buffer1[10];
		char *c =  lock_shift(1,terminal_info->current_R,
				      buffer1,sizeof buffer1);
		ret = strfcpy(buffer,c,size);
	    }
	} else {
	    char *c = lock_shift(1,terminal_info->current_R,NULL,0);
	    ret = strmcat(ret,c);
	    free(c);
	}
    }
	    
    for (i = 0; 
	 i < sizeof (iso2022_info->initial_bank) / 
	     sizeof (iso2022_info->initial_bank[0]);
	 i++) {
	int x = iso2022_info->initial_bank[i];
		
	if (-1 != x) {
	    int n;
	    struct iso2022_setid ID;

	    if (x < 0 ||
		x >= sizeof (iso2022_info->sets)  /
			sizeof (iso2022_info->sets[0]) ||
			! iso2022_info->sets[x]) 
		panic("STRING PANIC",__FILE__,__LINE__,
		      "change_helper_3",
		      "Bad initial_bank (set)",
		      maybe_signal);
	       
	    ID = * (iso2022_info->sets[x]);
		    
	    if (ID.bank != i)
		panic("STRING PANIC",__FILE__,__LINE__,
		      "change_helper_3",
		      "Bad initial_bank (bank number)",
		      maybe_signal);
		    
	    n = set_initial_bank(&ret,ID,terminal_info,buffer,size,
				 maybe_signal);

	    /* Do not support variable width inside of bank */
	    if (MB &&
		(iso2022_94x94 == ID.type ||
		 iso2022_96x96 == ID.type))
		terminal_info->width[n]  = 2;
	    else
		terminal_info->width[n]  = 1;
	}
    }

    for (i = 0; 
	 i < sizeof (iso2022_info->sets)  / 
	     sizeof (iso2022_info->sets[0]) &&
	     iso2022_info->sets[i];
	 i++) {
	int setpos;
	struct iso2022_setid ID = * (iso2022_info->sets[i]);

	/* iso2022_give_setpos does iso2022_setid_select_bank() */
	setpos = iso2022_give_setpos(&ID,terminal_info);

	if (-1 == setpos) {
	    if ((terminal_info)->set_count >=
			sizeof ((terminal_info)->sets) /
			sizeof ((terminal_info)->sets[0]))
			panic("STRING PANIC",__FILE__,__LINE__,
			      "change_helper_3",
			      "Too many iso 2022 sets on display charset",
			      maybe_signal);

		    setpos = terminal_info->set_count++;
		    terminal_info->sets[setpos] = ID;
	}

	change_helper_2(&ret,ID,terminal_info,buffer,size,
			maybe_signal,setpos);
		
	/* Do not support varaiable width inside of bank */
	if (MB &&
	    (iso2022_94x94 == ID.type ||
	     iso2022_96x96 == ID.type))
	    terminal_info->width[setpos]  = 2;
	else
	    terminal_info->width[setpos]  = 1;
    }

    debug_display_settings(terminal_info,maybe_signal,0);
	   
    *RET = ret;
}

/* May be called from signal handler -- on that situation use
   buffer specified as argument -- otherwise result is malloced
*/
char * terminal_switch_to(terminal,set,terminal_info,buffer,size)
     char *terminal; 
     charset_t set;
     screen_info_p *terminal_info;   
     char * buffer; 
     int size;
{
    int maybe_signal = buffer != NULL;

    char *ret = NULL;
    struct terminal_map_item * a = loc_info(terminal,set);
    
    if (!a)
	return NULL;

    if (maybe_signal) {
	SIGDPRINT(Debug,8,(&Debug,
			   "terminal_switch_to: buffer=%p size=%d\n",
			   buffer,size));
    }

    if (!*terminal_info) {
	panic("STRING PANIC",__FILE__,__LINE__,
	      "terminal_switch_to",
	      "*terminal_info not set",maybe_signal);
    } 

    if (DISPLAY_STATE_magic != (*terminal_info)->magic)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "terminal_switch_to",
	      "Bad terminal_info (bad magic)",
	      maybe_signal);

    if ((*terminal_info)->set_count > 0 &&     
	(*terminal_info)->sets[0].type == iso2022_other) {
	SIGDPRINT(Debug,8,(&Debug,
			   "Terminal was on no ISO2022 mode ... returning\n"));
	if (maybe_signal)
	    ret = iso2022_setid_stream(return_to_iso2022,buffer,size);
	else
	    ret = iso2022_setid_stream(return_to_iso2022,NULL,0);
    }
    
    reset_display_settings(*terminal_info);

    switch(a->keyword) {
    case terminal_iso2022:
    case terminal_iso2022_mb:
    if (set->iso2022_info) {
	SIGDPRINT(Debug,8,(&Debug,
			   "Using ISO 2022 code for assignment\n"));
	
	change_helper_3(&ret,set->iso2022_info,*terminal_info,
			buffer,size,maybe_signal,
			a->keyword == terminal_iso2022_mb);
    }
    break;
    
    case terminal_iso2022_like:

	SIGDPRINT(Debug,8,(&Debug,
			   "Using ISO 2022 LIKE code for assignment\n"));
	
	change_helper_3(&ret,a->value.iso2022_like.iso2022_info,
			*terminal_info,
			buffer,size,maybe_signal,0);
	
	break;

    case terminal_private:
	SIGDPRINT(Debug,8,(&Debug,
			   "Using terminal private code for assignment\n"));
	
	
	strXcat(&ret,sStr(a->value.private.bytes),
		maybe_signal,buffer,size);

	break;
	

    }

    if (maybe_signal && ret && ret != buffer)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "terminal_switch_to",
	      "buffer != ret when called from signal",1);
    
    return ret;
}


/* May be called from signal handler -- on that situation use
   buffer specified as argument -- otherwise result is malloced
*/
char * terminal_set_title(terminal,window_title, icon_title,buffer,size)
     char *terminal; 
     char *window_title;
     char *icon_title;
     char * buffer; 
     int size;
{
    /* window_title and icon_title are assumed to be US-ASCII ... */

    int maybe_signal = buffer != NULL;

    char *ret = NULL;

    static enum terminal_flags methods[] = {
	terminal_xterm_title,
	terminal_xwsh_title,
	terminal_bad_flag
    };

    enum terminal_flags method = find_terminal_flag(terminal,methods);

    if (maybe_signal) {
	SIGDPRINT(Debug,8,(&Debug,
			   "terminal_set_title: buffer=%p size=%d\n",
			   buffer,size));
    }


    switch (method) {
    case terminal_xterm_title:
	if (window_title == icon_title) {  /* Optimize */

	    if (window_title) {
		strXcat(&ret,"\033]0;",
			maybe_signal,buffer,size);
		strXcat(&ret,window_title,
			maybe_signal,buffer,size);
		strXcat(&ret,"\007",
			maybe_signal,buffer,size);
	    }
	} else {
	    if (icon_title) {
		strXcat(&ret,"\033]1;",
			maybe_signal,buffer,size);
		strXcat(&ret,icon_title,
			maybe_signal,buffer,size);
		strXcat(&ret,"\007",
			maybe_signal,buffer,size);
	    }

	    if (window_title) {
		strXcat(&ret,"\033]2;",
			maybe_signal,buffer,size);
		strXcat(&ret,window_title,
			maybe_signal,buffer,size);
		strXcat(&ret,"\007",
			maybe_signal,buffer,size);
	    }

	}
	break;
    case terminal_xwsh_title:

	if (icon_title) {
	    strXcat(&ret,"\033P3.y",
		    maybe_signal,buffer,size);
	    strXcat(&ret,icon_title,
		    maybe_signal,buffer,size);
	    strXcat(&ret,"\033\\",
		    maybe_signal,buffer,size);
	}

	if (window_title) {
	    strXcat(&ret,"\033P1.y",
		    maybe_signal,buffer,size);
	    strXcat(&ret,window_title,
		    maybe_signal,buffer,size);
	    strXcat(&ret,"\033\\",
		    maybe_signal,buffer,size);
	}

    case terminal_bad_flag:
	SIGDPRINT(Debug,8,(&Debug,
			   "terminal_set_title: No way to set title for %s\n",
			   terminal));
	break;
    }


    if (maybe_signal && ret && ret != buffer)
	panic("STRING PANIC",__FILE__,__LINE__,
	      "terminal_set_title",
	      "buffer != ret when called from signal",1);
    
    return ret;
}



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */


