static char rcsid[] = "@(#)$Id: smtp.c,v 1.40 2003/05/17 16:13:42 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.40 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "elmsmtp.h"

#include "../../hdrs/s_elm.h"

#ifdef REMOTE_MBX

DEBUG_VAR(Debug,__FILE__,"smtp");

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

#define SMTP_TIMEOUT  240

/* callbacks for connection ----------------------------------------------- */

/* SMTP_flags */
#define SMTP_submission         1       
#define SMTP_noverify           2       
#define SMTP_reserve            4

/* mailer info keeps connection information ... */
struct mailer_info_1 {
    struct remote_account RA;
    struct run_state RS;
    struct Read_Buffer  read_buffer;
    struct Write_Buffer write_buffer;

    enum smtp_state {  smtp_wait_reply, smtp_reply_read,
		       smtp_reply_incomplete,
		       smtp_idle, smtp_command, 
		       smtp_need_data, smtp_data, 
		       smtp_error, smtp_closing
    } state;
    char response_code[3];

    char ** response_text;
    int     response_text_len;

    int     SMTP_flags;

    char    *incomplete_command;
    int      incomplete_command_len;

    char   **ESMPT_keywords;
    int      ESMPT_keyword_len;   /* -1 == EHLO not issued yet */

};

S_(ss_action_routine smtp_read_action)
static int smtp_read_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    int n;
    struct mailer_info_1 *I = data;
    int i;
    int linelen;
    int close_it = 0;

    char * ERROR = NULL;
   
    if (!data ||
	I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_read_action",
	      "Bad data argument !!!",0);

    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_read_action: mailer info, I=%p\n",I));

    if (smtp_error == I->state) {
        DPRINT(Debug,13,(&Debug,
                         "smtp_read_action=0 (stream need closing)\n"));
        return 0;
    }

    n = ReadFromStream(ss, &(I->read_buffer),-1);

    if (n < 0) {      
	ERROR = RemoveStreamError(ss);
	DPRINT(Debug,13,(&Debug,
			 "smtp_read_stream: read error %s\n",
			 ERROR ? ERROR : "(none)"));
    }

    DPRINT(Debug,50,(&Debug,
		     "Buffer content="));
    for (i = 0; i < I->read_buffer.read_len; i++) {
	DPRINT(Debug,50,(&Debug,
			 " %d",
			 I->read_buffer.read_buffer[i]));
    }
    DPRINT(Debug,50,(&Debug,
		     "\n (message?)="));
    for (i = 0; i < I->read_buffer.read_len; i++) {
	DPRINT(Debug,50,(&Debug,
			 " %c",
			 isascii(I->read_buffer.read_buffer[i]) &&
			 isprint(I->read_buffer.read_buffer[i]) ? 
			 I->read_buffer.read_buffer[i] :
			 '.'));
    }
    DPRINT(Debug,50,(&Debug,
		     "\n"));

    while (0 < (linelen = find_crlf(&(I->read_buffer),1))) {
	char code[3];
	int final = 1;
	int i;
	char * message = NULL;
	enum smtp_state old_state = I->state;  /* Handle CLOSING case */

	DPRINT(Debug,13,(&Debug,
                         "smtp_read_action: Parsing response line (len =%d)\n",
			 linelen));

	for (i = 0; i < linelen && i < 3; i++)
	    code[i] = I->read_buffer.read_buffer[i];

	if (i < 3) {
	    int z = i;

	    for (; i < 3; i++)
		code[i] = '\0';
	
	    DPRINT(Debug,1,(&Debug,
			     "smtp_read_action: Line not include full repsonse code: %.3s (only %d chars)\n",
			    code,z));
	}

	if (linelen > 3)
	    final = I->read_buffer.read_buffer[3] != '-';

	if (linelen > 4)
	    message = & (I->read_buffer.read_buffer[4]);

	DPRINT(Debug,14,(&Debug,
                         "smtp_read_action: Parsing response line, code=%.3s final=%d message=%s\n",
			 code,final,
			 message ? message : "<NULL>"));


	if (0 == memcmp(code,"421",3)) {       /* CLOSING */
	    I->state = smtp_closing;
	    lib_transient(FRM("%s server: %s"),
			  (I->SMTP_flags & SMTP_submission) ? 
			  "Submission" : "SMTP",
			  message);		
	} 

	/* Parse response ... */
	if (smtp_wait_reply       == old_state ||
	    smtp_reply_incomplete == I->state) {

	    DPRINT(Debug,13,(&Debug,
			    "smtp_read_action: Waited response\n"));

	    if (smtp_wait_reply       == old_state) {
		if (smtp_wait_reply   == I->state)
		    I->state = smtp_reply_incomplete;
		memcpy(I->response_code,code,3);
		
		if (code[0] < '2' || code[0] > '5') {
		    I->state = smtp_error;
		    lib_error(FRM("Bad response from %s server: %s"),
			      (I->SMTP_flags & SMTP_submission) ? 
			      "submission" : "SMTP",
			      I->read_buffer.read_buffer);
		} 
	    }

	    if (smtp_reply_incomplete == I->state) {
		if (0 != memcmp(I->response_code,code,3)) {
		    I->state = smtp_error;
		    lib_error(FRM("Bad response from %s server: %s"),
			      (I->SMTP_flags & SMTP_submission) ? 
			      "submission" : "SMTP",
			      I->read_buffer.read_buffer);
		}

		if (final) {
		    if ('3' == code[0])
			I->state = smtp_need_data;
		    else
			I->state = smtp_reply_read;
		    
		    DPRINT(Debug,13,(&Debug,
				     "smtp_read_action : %sresponse=%.3s %s\n",
				     smtp_reply_read == I->state ?
				     "command ready, " : "",
				     I->response_code,
				     message ? message : "(no message text)"));
		}
	    }
	    
	    if (message) {
		I->response_text = safe_realloc(I->response_text,
						(I->response_text_len+1) *
						sizeof (*(I->response_text)));
		I->response_text[I->response_text_len++] = 
		    safe_strdup(message);
	    }
	    
	    cut_line(&(I->read_buffer),linelen);
	    message = NULL;          /* No longer valid */
	} else {
	    DPRINT(Debug,16,(&Debug,
			     "smtp_read_action: Quiting parsing loop (%d bytes left, unaccepted lines was %d bytes)\n",
			     I->read_buffer.read_len,
			     linelen));
	    break;
	}
    }

    if (I->read_buffer.read_len) {
	switch(I->state) {
	    int i;
	case smtp_reply_read:
	case smtp_need_data:
	case smtp_data:
	case smtp_idle:
	case smtp_command:
	    DPRINT(Debug,13,(&Debug,
			     "smtp_read_action: %d bytes data left! (bad state):\n",
			     I->read_buffer.read_len));
	    DPRINT(Debug,13,(&Debug,
			     "smtp_read_action: extra data="));
	    for (i =0; i < I->read_buffer.read_len; i++) {
		DPRINT(Debug,13,(&Debug,
				 " %d",
				 I->read_buffer.read_buffer[i]));
	    }
	    DPRINT(Debug,13,(&Debug,
			     "\nsmtp_read_action (message?)="));
	    for (i =0; i < I->read_buffer.read_len; i++) {
		DPRINT(Debug,13,(&Debug,
				 "%c",
				 isascii(I->read_buffer.read_buffer[i]) &&
				 isprint(I->read_buffer.read_buffer[i]) ?
				 I->read_buffer.read_buffer[i] :
				 '.'));
	    }
	    DPRINT(Debug,13,(&Debug,
			     "\n"));
	    
	    lib_error(FRM("Unexpected response from %s server"),
		      (I->SMTP_flags & SMTP_submission) ? 
		      "submission" : "SMTP");
	    close_it = 1;
	}
    }

 check_read:
    if (n == 0) {
	lib_error(FRM("Connection closed to %s server"),
		  (I->SMTP_flags & SMTP_submission) ? "submission" : "SMTP");
	close_it = 1;
    }
    if (n < 0 && ERROR) {
	lib_error(FRM("Error reading from %s server: %s"),
		  (I->SMTP_flags & SMTP_submission) ? "submission" : "SMTP",
		  ERROR);
	close_it = 1;
    }
    if (ERROR) {
	free(ERROR);
	ERROR = NULL;
    }
    if (close_it) {
	I->state = smtp_error;
	DPRINT(Debug,13,(&Debug,
			 "smtp_read_action=0 (stream need closing)\n"));
	return 0;
    }
    DPRINT(Debug,13,(&Debug,
		     "smtp_read_action=1\n"));

    return 1;
}

S_(ss_action_routine smtp_write_action)
static int smtp_write_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    char * ERROR = NULL;
    struct mailer_info_1 *I = data;
    int n;

    if (!data ||
	I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_write_action: mailer info, I=%p\n",I));

    if (!I->write_buffer.write_len) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action: NO DATA TO WRITE!\n"));
    }

    if (smtp_error   == I->state ||
	smtp_closing == I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 (stream need closing)\n"));
	return 0;
    }

    if (smtp_command != I->state &&
	smtp_data    != I->state) {

	panic("MBX PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "Bad state !!!",0);
    }

    n = WriteToStream(ss, & (I->write_buffer));

    if (n > 0) {
	int i;
	
	DPRINT(Debug,70,(&Debug,
			 "Write content="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug," %d",
			     I->write_buffer.write_buffer[i]));
	}
	DPRINT(Debug,70,(&Debug,"\n    (text?)="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug," %c",
			     isascii(I->write_buffer.write_buffer[i]) 
			     &&
			     isprint(I->write_buffer.write_buffer[i]) 
			     ? 
			     I->write_buffer.write_buffer[i] :
			     '.'));
	}
	DPRINT(Debug,70,(&Debug,"\n"));
	
	cut_Write_Buffer(& (I->write_buffer),n);

    } else if (n < 0) {
	ERROR = RemoveStreamError(ss);
	
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action: write error %s\n",
			 ERROR ? ERROR : "(none)"));
    }

    if (!I->write_buffer.write_len) {
	if (smtp_command == I->state)
	    I->state = smtp_wait_reply;
	else
	    I->state = smtp_need_data;

	if (ERROR) {
	    free(ERROR);
	    ERROR = NULL;
	}
	       
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 -- data written\n"));
	return 0;
    }
    
    if (ERROR) {
	lib_error(FRM("Error writing to %s server: %s"),
		  (I->SMTP_flags & SMTP_submission) ? "submission" : "SMTP",
		  ERROR);
	I->state = smtp_error;
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 (stream need closing)\n"));

	free(ERROR);
	ERROR = NULL;
        return 0;	
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_write_action=1\n"));
    return 1;

}

static int smtp_start_command P_((struct mailer_info_1 *I, char *command));
static void smtp_command_ready P_((struct mailer_info_1 *I, int data_end));

S_(ss_action_routine smtp_timeout_action)
static int smtp_timeout_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    struct mailer_info_1 *I = data;
    int status;

    if (!data ||
	I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_timeout_action",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_timeout_action: mailer info, I=%p\n",I));

    if (smtp_error   == I->state ||
	smtp_closing == I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_timeout_action=0 (stream need closing)\n"));
	return 0;
    }

    if (smtp_idle != I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_timeout_action=1 (stream not idle)\n"));
	return 1;
    }

    status = smtp_start_command(I,"NOOP");

    if (status) 
	smtp_command_ready(I,0);
    /* DO not wait completion of command */

    DPRINT(Debug,13,(&Debug,
                     "smtp_timeout_action=%d\n",
                     status));

    return status;
}

static void free_response_text P_((struct mailer_info_1 *I));
static void free_response_text(I)
     struct mailer_info_1 *I;
{
    if (I->response_text) {
	int i;
	for (i = 0; i < I->response_text_len; i++) {
	    if (I->response_text[i]) {
		free(I->response_text[i]);
		I->response_text[i] = NULL;
	    }
	}
	free(I->response_text);
    }            
    I->response_text     = NULL;
    I->response_text_len = 0;
}

static int smtp_start_command(I,command)
     struct mailer_info_1 *I; 
     char *command;
{
    int status = 0;
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command: mailer info, I=%p\n",
		     I));
    
    if (smtp_error == I->state   ||
	smtp_closing == I->state) {
    failure:
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    if (I->state != smtp_idle) {
	DPRINT(Debug,7,(&Debug,
			"smtp_start_command: wrong state -- perhaps idle timer -- waiting...\n"));

	while (smtp_reply_read != I->state &&
	       smtp_need_data    != I->state &&
	       smtp_error        != I->state &&
	       smtp_closing      != I->state &&
	       NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	DPRINT(Debug,7,(&Debug,
			 "smtp_start_command: ... wait ended\n"));

	/* DO NOT accpet smtp_need_data here -- it means that
	   data (not command) is expected
	*/
	if (smtp_error == I->state ||
	    smtp_closing == I->state) 
	    goto failure;

	if (0 != I->write_buffer.write_len) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command: write buffer not empty (%d bytes left)\n",
			     I->write_buffer.write_len));
	    status = 0;
	    goto clean;
	}       	

	free_response_text(I);
    }

    if (!I->RA.stream) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command: stream not open\n"));
	status = 0;
	goto clean;
    }

    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_start_command",
	      "incomplete_command not NUL !",0);
    I->incomplete_command_len = 0;

    I->state = smtp_command;

    DPRINT(Debug,15,(&Debug,
		     "smtp_start_command: Command %.4s...\n",command));

    I->incomplete_command     = safe_strdup(command);
    I->incomplete_command_len = strlen(command);

    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command=%d\n",
		     status));

    return status;
}

static int smtp_push_data P_((struct mailer_info_1 *I,
			      const char *data,
			      int len,
			      int is_data_line));

static int smtp_push_data(I,data,len,is_data_line)
     struct mailer_info_1 *I; 
     CONST char *data;
     int len;
     int is_data_line;
{
    int status = 0;
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data: mailer info, I=%p\n",
		     I));

    if (smtp_error == I->state   ||
	smtp_closing == I->state) {
    failure:
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_push_data: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }
    
    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data",
	      "incomplete_command not NUL !",0);
    I->incomplete_command_len = 0;


    if (smtp_need_data != I->state  &&
	smtp_data == I->state) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data",
	      "Wrong state !",0);
    }
    I->state = smtp_data;

    if (is_data_line) {
	char * d;
	/* Dot-Stuff data line */
	if ('.' == data[0]) {
	    char *x = safe_strdup(".");
	    add_to_Write_Buffer(& (I->write_buffer),&x,1); /* May free x */
	}

	d = malloc(len);
	memcpy(d,data,len);
	add_to_Write_Buffer(& (I->write_buffer),&d,len);   /* May free d */

	ConfigStream(I->RA.stream,
		     smtp_read_action,smtp_write_action,smtp_timeout_action,
		     SMTP_TIMEOUT,I);

	/* We push large amout of data so start immediately without
	   buffering all of data
	*/
	WaitStreamFor(I->RA.stream,SS_write_act);
    } else {
	I->incomplete_command = malloc(len);
	memcpy(I->incomplete_command,data,len);
	I->incomplete_command_len = len;
    }
    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data=%d\n",status));
    return status;
}

static void smtp_command_arg P_((struct mailer_info_1 *I, const char *str));
static void smtp_command_arg(I, str)
     struct mailer_info_1 *I;
     CONST char *str;
{
    int L = strlen(str);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_arg: mailer info, I=%p\n",
		     I));

    if (I->state != smtp_command &&
	I->state != smtp_data) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_arg",
	      "Bad state",0);
    }

    I->incomplete_command = safe_realloc(I->incomplete_command,
					 I->incomplete_command_len+1+L);
    I->incomplete_command[I->incomplete_command_len++] = ' ';

    memcpy(I->incomplete_command+I->incomplete_command_len,str,L);
    I->incomplete_command_len += L;
}

static void smtp_command_ready(I, data_end)
     struct mailer_info_1 *I;
     int data_end;
{
    char *s;
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready: mailer info, I=%p\n",
		     I));
    
    if (I->state != smtp_command &&
	I->state != smtp_data) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready",
	      "Bad state",0);
    }

    if (data_end && I->state != smtp_data) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready",
	      "Bad state -- not smtp data",0);
    }

    /* May free I->p.p1->incomplete_command */
    add_to_Write_Buffer(& (I->write_buffer),
			&(I->incomplete_command),
			I->incomplete_command_len);

    s = safe_strdup("\r\n");
    /* May free s */
    add_to_Write_Buffer(& (I->write_buffer),&s,2);

    ConfigStream(I->RA.stream,
		 smtp_read_action,smtp_write_action,smtp_timeout_action,
		 SMTP_TIMEOUT,I);

    if (data_end) {
	I->state = smtp_command;
	/* We push large amout of data so start immediately without
	   buffering all of data
	*/
	WaitStreamFor(I->RA.stream,SS_write_act);
    }

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready: done%s\n",
		     smtp_command == I->state ? " (state: smtp_command)" : ""
		     ));
}


/* May return command OK before all response lines are returned */
static int smtp_command_ok P_((struct mailer_info_1 *I, 
			       char response_code[3]));
static int smtp_command_ok(I,response_code)
     struct mailer_info_1 *I;
     char response_code[3];
{
    int status = 0;

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok: mailer info, I=%p\n",
		     I));
    response_code[0] = '\0';
    
    if (smtp_error == I->state   ||
	smtp_closing == I->state) {
    failure:
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    if (smtp_command          != I->state &&
	smtp_reply_read     != I->state &&
	smtp_need_data        != I->state &&
	smtp_reply_incomplete != I->state &&
	smtp_wait_reply       != I->state) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ok",
	      "Bad state",0);
    }
   
    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: writing data to server...\n"));
	
	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_write_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: ... write ended\n"));
	if (smtp_error == I->state ||
	    smtp_closing == I->state) 
	    goto failure;

    }

    /* If nothing read, wait data for now */
    if (smtp_wait_reply  == I->state) {
      
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: waiting reply from server...\n"));

	while ( smtp_wait_reply  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: ... wait ended\n"));
	if (smtp_error == I->state ||
	    smtp_closing == I->state) 
	    goto failure;
	
	if (!I->RA.stream) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok: stream not open\n"));
	    status = 0;
	    goto clean;
	}
    }

    if (smtp_reply_incomplete == I->state ||
	smtp_reply_read     == I->state ||
	smtp_need_data        == I->state) {
	memcpy(response_code,I->response_code,3);
	
	if ('2' == response_code[0] ||
	    '3' == response_code[0])
	    status = 1;
    }
    
 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok=%d\n",status));
    return status;
}


/* Returns response lines one at time and waits that all is returned */
static char * smtp_command_response P_((struct mailer_info_1 *I,
					int *idx));

static char * smtp_command_response(I,idx)
     struct mailer_info_1 *I;
     int *idx;
{
    char * status = NULL;

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response: mailer info, I=%p, *idx=%d\n",
		     I,*idx));
        
    if (*idx < 0) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_response",
	      "Negative *idx",0);
    }

    if (smtp_reply_incomplete != I->state &&
	smtp_reply_read     != I->state &&
	smtp_need_data        != I->state) {
	
	failure:
	DPRINT(Debug,12,(&Debug,
			 "smtp_command_response: Bad state\n"));
	
	status = NULL;
	goto clean;
    }

    if (*idx >= I-> response_text_len &&
	smtp_reply_incomplete == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_response: waiting more data from server...\n"));
	while (*idx >= I-> response_text_len &&
	       smtp_reply_incomplete == I->state &&
	       NULL         != I->RA.stream) {
            WaitStreamFor(I->RA.stream,SS_read_act);
            /* TODO:    Should user to have some way to interrupt progress ? */
        }
	DPRINT(Debug,7,(&Debug,
			"smtp_command_response: ... wait ended\n"));
	if (smtp_reply_incomplete != I->state &&
	    smtp_reply_read     != I->state &&
	    smtp_need_data        != I->state)
	    goto failure;
    }

    if (*idx < I-> response_text_len) {
	status = I-> response_text[(*idx)++];
    }

 clean:

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response=%s\n",
		     status ? status : "<NULL>"));

    return status;
}


/* Waits that command is completed and resets data then */
static void smtp_command_reset P_((struct mailer_info_1 *I));
static void smtp_command_reset(I)
     struct mailer_info_1 *I;
{
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset: mailer info, I=%p\n",I));
    
    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: writing data to server...\n"));
	
	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_write_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: ... write ended\n"));
    }

    if (smtp_wait_reply  == I->state ||
	smtp_reply_incomplete == I->state) {

	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: waiting reply from server...\n"));

	while ( (smtp_wait_reply       == I->state || 
		 smtp_reply_incomplete == I->state) &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* TODO:    Should user to have some way to interrupt progress ? */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: ... wait ended\n"));
    }

    if (smtp_error == I->state   ||
	smtp_closing == I->state) {
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_reset: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
    }

    free_response_text(I);

    if (smtp_error     != I->state  &&
	smtp_closing   != I->state &&
	smtp_need_data != I->state)
	I->state = smtp_idle;
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset: done%s\n",
		     smtp_idle == I->state ? " (state: smtp_idle)" : ""));
}

static void smtp_run_ehlo P_((struct mailer_info_1 *I));
static void smtp_run_ehlo(I)
     struct mailer_info_1 *I;
{
    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo: mailer info, I=%p\n",I));

    if (smtp_start_command(I,"EHLO")) {
	char response_code[3];

	smtp_command_arg(I,hostfullname);
	smtp_command_ready(I,0);

	if (smtp_command_ok(I,response_code)) {

	    if (0 == memcmp(response_code,"250",3)) {
		int i;
		char * line;
		
		I->ESMPT_keyword_len = 0;

		for (i = 0; NULL != (line = smtp_command_response(I,&i)); ) {
		    if (i > 0) { /* First response line is greeting */

			I->ESMPT_keywords = 
			    safe_realloc(I->ESMPT_keywords,
					 (I->ESMPT_keyword_len+1) *
					 sizeof (*(I->ESMPT_keywords)));
			I->ESMPT_keywords[I->ESMPT_keyword_len++] =
			    safe_strdup(line);			
		    }
		}		
	    } else {
		DPRINT(Debug,14,(&Debug,
				 "smtp_run_ehlo: not expected response code: %.3s\n",response_code));
		goto fail;
	    }
	} else {
	fail:
	    /* If EHLO fails try HELO ... (ignoring result) */
	    smtp_command_reset(I);
	    if (smtp_start_command(I,"HELO")) {
		smtp_command_arg(I,hostfullname);
		smtp_command_ready(I,0);

		I->ESMPT_keyword_len = 0;
	    }
	}
	smtp_command_reset(I);
    }
    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo: done\n",I));
}

/* common to submit and sendmail-bs mailer -------------------------------- */


S_(mailer_set_option default_set_option)
static int default_set_option(M,X,value) 
     struct mailer_config *M;
     struct mailer_option_list *X;
     char *value;
{
    int r ;

    r = X->value_func(M,X,&value,1);
    return r;
}

static char * default_from P_((void));
static char * default_from(void)
{
    char * ret = NULL;
    if (NULL != strpbrk(username,
                        ":\"\\!#%&()=?',;.:<>")) {
	ret = elm_message(FRM("<%Q@%s>"),
			   username,hostfullname);
    } else {
	ret = elm_message(FRM("<%s@%s>"),
			   username,hostfullname);
    }

    return ret;
}

S_(mailer_backend_hook smtp_common_mailer_backend)
int smtp_common_mailer_backend(M,C,encoding_top,title,sm)
     struct mailer_config *M;
     struct mail_send_state **C;
     int encoding_top;
     char * title;
     sending_message_func *sm;
{
    char buffer [ 32 * 1024 ];
    int l;
    int have_dsn = 0;

    char * from;
    char * f1;
   
    if ((*C)->magic != MAILER_magic)
    	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_backend",
	      "Bad magic number (mail send state)",0);

    f1 = (*C)->mail_from;
    if (!f1)
	f1 = default_from();

    if ('<' == f1[0]) {
	from = elm_message(FRM("FROM:%s"),
			   f1);
    } else {
	from = elm_message(FRM("FROM:<%s>"),
			   f1);
    }
    
    if (f1 != (*C)->mail_from) {
	free(f1);
	f1 = NULL;
    }

    /* With b)ounce command EHLO/HELO is not issued before that */
    if (-1 == (*C)->head->p.p1->ESMPT_keyword_len)
	smtp_run_ehlo((*C)->head->p.p1);

    if ((*C)->dsn) {
	int i;
	for (i = 0; i < (*C)->head->p.p1->ESMPT_keyword_len; i++)
	    if (0 == istrcmp("DSN",(*C)->head->p.p1->ESMPT_keywords[i])) {
		have_dsn = 1;
	    }
    }
    
    if (smtp_start_command((*C)->head->p.p1,"MAIL")) {
	char response_code[3];
	smtp_command_arg((*C)->head->p.p1,from);

	if (encoding_top == ENCODING_8BIT) {
	    int i;
	    for (i = 0; i < (*C)->head->p.p1->ESMPT_keyword_len; i++)
		if (0 == istrcmp("8BITMIME",
				 (*C)->head->p.p1->ESMPT_keywords[i])) {
		    smtp_command_arg((*C)->head->p.p1,"BODY=8BITMIME");
		    break;
		}
	}

#if 0
	if (encoding_top == ENCODING_BINARY) {
	    int i;
	    for (i = 0; i < (*C)->head->p.p1->ESMPT_keyword_len; i++)
		if (0 == istrcmp("BINARYMIME",
				 (*C)->head->p.p1->ESMPT_keywords[i])) {
		    smtp_command_arg((*C)->head->p.p1,"BODY=BINARYMIME");
		    break;
		}
	}
#endif

	smtp_command_ready((*C)->head->p.p1,0);
	if (!smtp_command_ok((*C)->head->p.p1,response_code)) {	    
	    if (response_code[0]) {
		int idx = 0;
		char * r = 
		    smtp_command_response((*C)->head->p.p1,&idx);
		if (r) {
		    lib_error(FRM("%s server rejected MAIL with %3s %s"),
			      ((*C)->head->p.p1->SMTP_flags & 
			       SMTP_submission) ?
			      "Submission" : "SMTP",
			      response_code,r);
		    while (r = smtp_command_response((*C)->head->p.p1,&idx)) {
			lib_error(FRM("%s server: %s"),
				  ((*C)->head->p.p1->SMTP_flags & 
				   SMTP_submission) ?
				  "Submission" : "SMTP",
				  r);
		    }
		} else
		    lib_error(FRM("%s server rejected MAIL with code %3s"),
			      ((*C)->head->p.p1->SMTP_flags & 
			       SMTP_submission) ?
			      "Submission" : "SMTP",
			      response_code);
	    } else
		lib_error(FRM("Connection to %s server failed after MAIL"),
			  ((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? 
			  "Submission" : "SMTP");

	    free(from); from = NULL;
	
	    smtp_command_reset((*C)->head->p.p1);
	    free_mail_send_state(C);
	    return 0;
	}
    } else {
	lib_error(FRM("Connection to %s server failed"),
		  ((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? 
		  "Submission" : "SMTP");

	free(from); from = NULL;

	free_mail_send_state(C);
	return 0; /* FAIL */
    }
    smtp_command_reset((*C)->head->p.p1);

    free(from); from = NULL;

    if ((*C)->addrs) {
	int i;
	int fail = 0;

	for (i = 0; (*C)->addrs[i]; i++) {
	    char * R = (*C)->addrs[i];

	    /* Make sure that addresses are inside of <> */
	    if ('<' != (*C)->addrs[i][0])
		R = kludge_addr( &((*C)->addrs[i]));

	    if (smtp_start_command((*C)->head->p.p1,"RCPT")) {
		char response_code[3];
		char * rcpt = safe_strdup("TO:");

		rcpt = strmcat(rcpt,R);
		smtp_command_arg((*C)->head->p.p1,rcpt);
		free(rcpt); rcpt = NULL;

		if (have_dsn) {
		    if ((*C)->dsn & DSN_FULL) 
			smtp_command_arg((*C)->head->p.p1,"RET=FULL");
		    else if ((*C)->dsn & DSN_HDRS)
			smtp_command_arg((*C)->head->p.p1,"RET=HDRS");
     
		    if ((*C)->dsn & DSN_NEVER)
			smtp_command_arg((*C)->head->p.p1,"NOTIFY=NEVER");
		    
		    else if ((*C)->dsn & (DSN_SUCCESS|DSN_FAILURE|DSN_DELAY)) {
			char * A = safe_strdup("NOTIFY=");
			int x = 0;
			if ((*C)->dsn & DSN_SUCCESS) {
			    A = strmcat(A,"SUCCESS"); x++;
			}
			if ((*C)->dsn & DSN_FAILURE) {
			    if (x) A = strmcat(A,",");
			    A = strmcat(A,"FAILURE"); x++;
			}
			if ((*C)->dsn & DSN_DELAY) {
			    if (x) A = strmcat(A,",");
			    A = strmcat(A,"DELAY"); 
			}
			smtp_command_arg((*C)->head->p.p1,A);
			free(A); A = NULL;
		    }
		}

		smtp_command_ready((*C)->head->p.p1,0);
		if (!smtp_command_ok((*C)->head->p.p1,response_code)) {
		    if (response_code[0]) {
			int idx = 0;
			char * r = 
			    smtp_command_response((*C)->head->p.p1,&idx);
			if (r) {
			    lib_error(FRM("%s server rejected %s with %3s %s"),
				      ((*C)->head->p.p1->SMTP_flags & 
				       SMTP_submission) ?
				      "Submission" : "SMTP",
				      (*C)->addrs[i],
				      response_code,r);
			    while (r = smtp_command_response((*C)->head->p.p1,&idx)) {
				lib_error(FRM("%s server: %s"),
					  ((*C)->head->p.p1->SMTP_flags & 
					   SMTP_submission) ?
					  "Submission" : "SMTP",
					  r);
			    }
			} else
			    lib_error(FRM("%s server rejected %s with code %3s"),
				      ((*C)->head->p.p1->SMTP_flags & 
				       SMTP_submission) ?
				      "Submission" : "SMTP",
				      (*C)->addrs[i],
				      response_code);
		    } 
		    fail++;
		} 

	    } else
		fail++;
	    smtp_command_reset((*C)->head->p.p1);
	}

	if (fail) {
	    lib_error(FRM("%d recipients from %d failed"),
		      fail,i);
	    if (fail >= i) {
		free_mail_send_state(C);
		return 0; /* FAIL */
	    }
	}

    } else {
	DPRINT(Debug,13,(&Debug,
			 "smtp_common_mailer_backend: No recipient addresses...\n"));
	
	if (smtp_start_command((*C)->head->p.p1,"RSET")) {
	    smtp_command_ready((*C)->head->p.p1,0);
	    smtp_command_reset((*C)->head->p.p1);
	}

	free_mail_send_state(C);
	return 0; /* NO addresses ... */
    }


    if (smtp_start_command((*C)->head->p.p1,"DATA")) {
	char response_code[3];
	smtp_command_ready((*C)->head->p.p1,0);
	if (!smtp_command_ok((*C)->head->p.p1,response_code)) {
	    if (response_code[0])
		lib_error(FRM("%s server rejected DATA with code %3s"),
			  ((*C)->head->p.p1->SMTP_flags &
			   SMTP_submission) ?
			  "Submission" : "SMTP",
			  response_code);
	    else
		lib_error(FRM("Connection to %s server failed after MAIL"),
			  ((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? 
			  "Submission" : "SMTP");
	    
	fail_1:
	    if (smtp_start_command((*C)->head->p.p1,"RSET")) {
		smtp_command_ready((*C)->head->p.p1,0);
		smtp_command_reset((*C)->head->p.p1);
	    }

	    free_mail_send_state(C);
	    return 0; /* NO addresses ... */	    
	}
    } else
	goto fail_1;
	
    out_state_fseek (&((*C)->OUT), 0);

    /* may be incorrect ... 
     * (but here should not be other \r or \n than \r\n) 
     */
    
    while (l = mail_gets(buffer,sizeof buffer,
			 out_state_FILE(&((*C)->OUT)))) {
	
	if (!smtp_push_data((*C)->head->p.p1,buffer,l,1))
	    break;
	
    }
    
    if (!smtp_push_data((*C)->head->p.p1,".",1,0)) {
	lib_error(FRM("Connection to %s server failed during DATA"),
		  ((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? 
		  "Submission" : "SMTP");
	
    } else {
	char response_code[3];
	smtp_command_ready((*C)->head->p.p1,1);   /* End data */


	if (!smtp_command_ok((*C)->head->p.p1,response_code)) {	    
	    if (response_code[0]) {
		int idx = 0;
		char * r = 
		    smtp_command_response((*C)->head->p.p1,&idx);
		if (r) {
		    lib_error(FRM("%s server rejected message with %3s %s"),
			      ((*C)->head->p.p1->SMTP_flags & 
			       SMTP_submission) ?
			      "Submission" : "SMTP",
			      response_code,r);
		    while (r = smtp_command_response((*C)->head->p.p1,&idx)) {
			lib_error(FRM("%s server: %s"),
				  ((*C)->head->p.p1->SMTP_flags & 
				   SMTP_submission) ?
				  "Submission" : "SMTP",
				  r);
		    }
		} else
		    lib_error(FRM("%s server rejected message with code %3s"),
			      ((*C)->head->p.p1->SMTP_flags & 
			       SMTP_submission) ?
			      "Submission" : "SMTP",
			      response_code);
	    } else
		lib_error(FRM("Connection to %s server failed after message"),
			  ((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? 
			  "Submission" : "SMTP");
	    
	    free_mail_send_state(C);
	    return 0;
	}

	{ 
	    union any_fd FD;
	    struct run_state RS;

	    FD.mail_fd         = *C;
	    RS.save_errno      = 0;

	    /* Fake ... */
	    (*C)->orig_func(FD,title,&RS,1,0);
	}
    }
       
    free_mail_send_state(C);
    return 1;
}



S_(mailer_init_hook smtp_common_mailer_init)
static int smtp_common_mailer_init(M,C,I)
     struct mailer_config *M;
     struct mail_send_state *C;
     struct mailer_info *I;
{
    FILE * F;

    if (C->magic != MAILER_magic)
    	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_init",
	      "Bad magic number (mail send state)",0);

    if (!I) {
	lib_error(FRM("SMTP mailer requires info parameter"));
	return 0;
    }

    if (I->p.p1->SMTP_flags & SMTP_reserve) {
	panic("MAILER PANIC",__FILE__,__LINE__,
	      "smtp_common_mailer_init",
	      "SMTP transaction already on progress",0);
    }
	
    I->p.p1->SMTP_flags |= SMTP_reserve;


    if (smtp_start_command(I->p.p1,"NOOP")) {
	smtp_command_ready(I->p.p1,0);
	smtp_command_reset(I->p.p1);
    }

    
    C->fname = tempnam(temp_dir, "snd.");
    if (!C->fname) {
	DPRINT(Debug,1,(&Debug,   
			"couldn't make temp file nam! (mail)\n"));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotMakeTemp,
			  "Sorry - couldn't make temp file name."));
	return 0;
    }

    F = safeopen_rdwr(C->fname);

    if (!F) {
	int err = errno;
	DPRINT(Debug,1, (&Debug,
			 "Attempt to open file %s for writing failed! (write_header_info)\n",
			 C->fname));
	DPRINT(Debug,1, (&Debug, "** %s **\n\n", error_description(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorTryingToWrite,
			  "Error %s encountered trying to write to %s."), 
		  error_description(err), C->fname);
	
	return 0;
    }

    out_state_clear(& (C->OUT), STATE_out_file);
    set_out_state_file(F,& (C->OUT));

    /* SMTP wants CRLF as eoln */
    C->OUT.EOLN_is_CRLF = 1;

    return 1;
}


S_(mailer_close_hook smtp_common_mailer_close)
static void smtp_common_mailer_close(M,C)
     struct mailer_config *M;
     struct mail_send_state *C;
{

    if (! C->head)
	return;   /* Oh no ... */


    switch (C->head->p.p1->state) {
    case smtp_need_data:
    case smtp_data:
	DPRINT(Debug,13,(&Debug,
			 "smtp_common_mailer_close: OH NO -- middle of data transfer\n"));
	C->head->p.p1->state = smtp_error;
	break;
    }

    if (smtp_start_command(C->head->p.p1,"RSET")) {
	smtp_command_ready(C->head->p.p1,0);
	smtp_command_reset(C->head->p.p1);
    }

    /* C->fname is free'ed by caller ( free_mail_send_state() ) */

    if (C->OUT.magic) {
	FILE *F = out_state_FILE(& (C->OUT));
	
	if (F) {
	    out_state_destroy(& (C->OUT));
	    fclose(F);
	}
    }

    C->head->p.p1->SMTP_flags &= ~SMTP_reserve;
}

S_(mailer_info_query_hook smtp_common_mailer_info_query)
static int smtp_common_mailer_info_query(M,I,query)
     struct mailer_config *M;
     struct mailer_info *I;
     enum MI_query query;
{
    if (-1 == I->p.p1->ESMPT_keyword_len)
	smtp_run_ehlo(I->p.p1);

    switch(query) {
	int i;

    case MI_HAVE_8BITMIME:
	for (i = 0; i < I->p.p1->ESMPT_keyword_len; i++)
	    if (0 == istrcmp("8BITMIME",I->p.p1->ESMPT_keywords[i]))
		return 1;
	DPRINT(Debug,10,(&Debug,"8BITMIME not found from EHLO response\n"));
	return 0;
    case MI_HAVE_BINARYMIME:
	return 0;       /* This code does not support BINARYMIME */
    case MI_HAVE_DSN:
	for (i = 0; i < I->p.p1->ESMPT_keyword_len; i++)
	    if (0 == istrcmp("DSN",I->p.p1->ESMPT_keywords[i]))
		return 1;
	DPRINT(Debug,10,(&Debug,"DSN not found from EHLO response\n"));
	return 0;
    case MI_DONT_ADD_FROM:
	return (M -> mailer_bits & MB_DONT_ADD_FROM) ? 1 : 0;
    case MI_USE_DOMAIN:
	return (M -> mailer_bits & MB_USE_DOMAIN) ? 1 : 0;
    }

    return 0; /* Not available */
}

S_(mailer_info_verify_addr smtp_tail_verify_addr)
static int smtp_tail_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     CONST char *text;
     struct addr_item *result;
{
    int i;
    int skip_enchanced = 0;
   
    if (-1 == I->p.p1->ESMPT_keyword_len)
	smtp_run_ehlo(I->p.p1);

    if (I->p.p1->SMTP_flags & SMTP_noverify) {
	DPRINT(Debug,13,(&Debug, "-- Verify temporary disabled\n"));
	return -2;       /* Cannot VRFY user */		
    }

    for (i = 0; i < I->p.p1->ESMPT_keyword_len; i++)
	if (0 == istrcmp("ENHANCEDSTATUSCODES",I->p.p1->ESMPT_keywords[i]))
	    skip_enchanced++;
    
    if (skip_enchanced) {
	DPRINT(Debug,13,(&Debug,
			 "-- Need skip enchanced status codes from responses\n"));
    }

    if (smtp_start_command(I->p.p1,"VRFY")) {
	int succeed = 0;
	char response_code[3];

	smtp_command_arg(I->p.p1,text);
	smtp_command_ready(I->p.p1,0);

	if (smtp_command_ok(I->p.p1,response_code)) {
	    if (0 == memcmp(response_code,"252",3)) {
		DPRINT(Debug,13,(&Debug,
				 " verify response %.3s -- cannot verify user\n",
				 response_code));
		succeed = -2;       /* Cannot VRFY user (relaying?) */
	    } else if (0 == memcmp(response_code,"250",3)) {
		int idx;
		char * text = NULL;
		char * buffer = NULL;

		for (idx = 0; NULL != (text = smtp_command_response(I->p.p1,
								    &idx));) {
		    buffer = text;
		}
		
		DPRINT(Debug,13,(&Debug," %d verify responses, result=%s\n",
				 idx,buffer ? buffer : "<NULL>"));

		/* If there was exactly one answer we can use it */
		if (1 == idx) {
		    struct addr_item * X;
		    char *x;
		    
		    if (skip_enchanced && buffer[0] == '2' &&
			(x = strchr(buffer,' '))) {
			buffer = x+1;
			DPRINT(Debug,13,(&Debug," strimmed result=%s\n",
					 buffer));
		    }

		    /* FIXME: Posibble wrong charset */			
		    X = break_down_address(buffer,
					   0,
					   system_charset);
		    if (!X[0].addr || X[1].addr)
			succeed = -1;   /* User OK, but caller should fill data */
		    else {
			result->addr = safe_strdup(X[0].addr);
			result->fullname = 
			    X[0].fullname ? dup_string(X[0].fullname) : NULL;
			result->comment  =
			    X[0].comment ? dup_string(X[0].comment) : NULL;
			succeed = result->fullname ? 1 : -1;
			DPRINT(Debug,13,(&Debug," filled addr=%s\n",
					 result->addr));
		    }
		} else
		    succeed = -1;   /* User OK, but caller should fill data */
	    }
	} else {
	    /* Failed -- see if response code indicates user unknown */
	    if (0 == memcmp(response_code,"553",3) /* Ambiguous */ ||
		0 == memcmp(response_code,"550",3) /* mailbox unavailable */
		) { 
		succeed = 0;		
	    } else if ('4' == response_code[0])
		succeed = -2;       /* Cannot VRFY user (temporray error) */
	    else {
		succeed = -2;       /* Cannot VRFY user ... */
		DPRINT(Debug,13,(&Debug, "-- Disabling verify...\n"));
		I->p.p1->SMTP_flags |= SMTP_noverify;
	    }
	}

	smtp_command_reset(I->p.p1);	
	DPRINT(Debug,13,(&Debug, 
			 "smtp_tail_verify_addr: %s = %d\n",
			 text,succeed));
	return succeed;
    }
    return -2;   /* caller does job */
}

S_(mailer_info_gen_def_ef smtp_common_mailer_info_gen_def_ef)
static void smtp_common_mailer_info_gen_def_ef(M,I,X)
     struct mailer_config *M;
     struct mailer_info *I;
     struct mailer_env_from *X; 
{
    if (X->mail_from)
	free(X->mail_from);
    X->mail_from = default_from();
}

S_(mailer_info_set_ef smtp_common_mailer_info_set_ef)
static  void smtp_common_mailer_info_set_ef(M,I,X,value)
     struct mailer_config *M;
     struct mailer_info *I;
     struct mailer_env_from *X;
     CONST char *value;
{
    if (0 == strcmp(value,"<>")) {
	/* Null envelope sender is only valid as sender (not as recipient)
	   so do not verify it
	*/
	X->mail_from = strmcpy(X->mail_from,value);
	return;
    }

    if (-1 == I->p.p1->ESMPT_keyword_len)
	smtp_run_ehlo(I->p.p1);

    if (I->p.p1->SMTP_flags & SMTP_noverify) {
	DPRINT(Debug,4,(&Debug, "-- Verify temporary disabled - smtp_common_mailer_info_set_ef assuming ok \n"));
	goto succeed;       /* Cannot VRFY user */		
    }

    if (smtp_start_command(I->p.p1,"VRFY")) {
	int succeed;
	char response_code[3];

	succeed  = 0;
	smtp_command_arg(I->p.p1,value);
	smtp_command_ready(I->p.p1,0);

	if (smtp_command_ok(I->p.p1,response_code)) {
	    if (0 == memcmp(response_code,"252",3)) {
		DPRINT(Debug,4,(&Debug,
				 " verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				 response_code));
		succeed = 1; /* Cannot VRFY user (relaying?) */
	    } else if (0 == memcmp(response_code,"250",3)) {
		succeed = 1;
	    } else {
		DPRINT(Debug,4,(&Debug,
				 " verify response %.3s -- smtp_common_mailer_info_set_ef assuming ok \n",
				 response_code));
		succeed = 1; /* unexpect OK response */
	    }
	} else {
	    /* Failed -- see if response code indicates user unknown */
	    if (0 == memcmp(response_code,"553",3) /* Ambiguous */ ||
		0 == memcmp(response_code,"550",3) /* mailbox unavailable */
		) { 
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef failed\n",
				response_code));
		    goto failed;
	    } else if ('4' == response_code[0])
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				response_code));
	    else {		
		DPRINT(Debug,13,(&Debug, "-- Disabling verify...\n"));
		I->p.p1->SMTP_flags |= SMTP_noverify;
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				response_code));
	    }
	    succeed = 1;
	}
        
	if (succeed) {
	succeed:
	    X->mail_from = strmcpy(X->mail_from,value);
	} else {
	failed:
	    lib_error(FRM("Changing of envelope sender failed: %s"),
		      value);
	}
	smtp_command_reset(I->p.p1);	
    } else {
	DPRINT(Debug,4,(&Debug,
			"smtp_common_mailer_info_set_ef failed to run verify command\n"));
	lib_error(FRM("Changing of envelope sender failed: %s"),
		  value);
    }
}



/* submission mailer ------------------------------------------------------ */

#include <sys/socket.h>

static char *submission_server = "localhost";

S_(mailer_info_init_hook submission_mailer_info_init)
static int submission_mailer_info_init(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    struct service_entry *se;
    char response_code[3];

    PORTS ports[] = { PORT_submission, PORT_smtp, PORT_end };

    int got;  /* port number */


    int status = 0;   

    DPRINT(Debug,13,(&Debug,
		     "submission_mailer_info_init: mailer info, I=%p\n",I));

    se = give_service_entry(submission_server, STFLAG_smtp_like);
    if (!se)
	return 0;

    I->p.p1 = safe_malloc(sizeof (* (I->p.p1)));

    /* bzero is defined on hdrs/defs.h */
    bzero((void *)I->p.p1,sizeof (* (I->p.p1)));

    zero_remote_account(& (I->p.p1->RA));

    /* There is no connection caching currently */

    I->p.p1->RA.host = strmcpy(I->p.p1->RA.host,se->official_name);
    		 
    zero_Read_Buffer(& (I->p.p1->read_buffer));
    zero_Write_Buffer(& (I->p.p1->write_buffer));
    I->p.p1->response_code[0]  = '\0';
    I->p.p1->response_text     = NULL;
    I->p.p1->response_text_len = 0;
    I->p.p1->SMTP_flags        = 0;

    I->p.p1->incomplete_command     = NULL;
    I->p.p1->incomplete_command_len = 0;

    I->p.p1->ESMPT_keywords        = NULL;
    I->p.p1->ESMPT_keyword_len     = -1;   /* -1 == EHLO not issued yet */
    I->p.p1->state                 = smtp_error;

    if (!connect_remote_account(&(I->p.p1->RA),&got,se,ports))
	goto clean;

    if (SUBMISSION_SERVICE == se->service)
	    goto is_submission;
    else if (SMTP_SERVICE == se->service)
	goto is_smtp;
    else switch (got) {
    case PORT_submission:
          is_submission:
    
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: Submission service\n"));
	I->p.p1->SMTP_flags    = SMTP_submission;
	break;

    case PORT_smtp:
        is_smtp:

	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: smtp service\n"));
	    
    default:
	DPRINT(Debug,1,(&Debug,
			"submission_mailer_info_init: Unknown sercide!\n"));
    }

    I->p.p1->state = smtp_wait_reply;

    ConfigStream(I->p.p1->RA.stream,
		 smtp_read_action,ss_noaction_routine,smtp_timeout_action,
		 SMTP_TIMEOUT,I->p.p1);



    if (!smtp_command_ok(I->p.p1,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_init: Failed to get greeting from server\n"));

	/* Caller will call sendmailbs_mailer_info_close ... */
	status =  0;
	goto clean;
    }
    smtp_command_reset(I->p.p1);
    status = 1;

 clean:

    free_temporary_service_entry(&se);

    DPRINT(Debug,12,(&Debug,
		     "submission_mailer_info_init=%d\n",status));
    return status;
}

S_(mailer_info_close_hook submission_mailer_info_close)
static void submission_mailer_info_close(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{

    DPRINT(Debug,13,(&Debug,
		     "submission_mailer_info_close: mailer info, I=%p\n",I));

    if (I->p.p1) {
	if (smtp_idle == I->p.p1->state) {
	    int status = smtp_start_command(I->p.p1,"QUIT");

	    if (status) {
		smtp_command_ready(I->p.p1,0);
		
		/* Wait for QUIT */	       
		smtp_command_reset(I->p.p1);
	    }
	}

	free_remote_account(& (I->p.p1->RA));	
	free_Read_Buffer(& (I->p.p1->read_buffer));
	free_Write_Buffer(& (I->p.p1->write_buffer));

	free_response_text(I->p.p1);

	if (I->p.p1->incomplete_command)
	    free(I->p.p1->incomplete_command);

	I->p.p1->incomplete_command     = NULL;
	I->p.p1->incomplete_command_len = 0;

	if (I->p.p1->ESMPT_keywords) {
	    int i;
	    for (i = 0; i < I->p.p1->ESMPT_keyword_len; i++) {
		if (I->p.p1->ESMPT_keywords[i])
		    free(I->p.p1->ESMPT_keywords[i]);
		I->p.p1->ESMPT_keywords[i] = NULL;
	    }
	    free(I->p.p1->ESMPT_keywords);
	}
	I->p.p1->ESMPT_keywords        = NULL;
	I->p.p1->ESMPT_keyword_len     = -1;   /* -1 == EHLO not issued yet */

	/* bzero is defined on hdrs/defs.h */
	bzero((void *)I->p.p1,sizeof (* (I->p.p1)));
	free(I->p.p1);
	I->p.p1 = NULL;
    } 
}

static int submission_verify_address = 0;

S_(mailer_info_verify_addr submission_mailer_info_verify_addr)
static int submission_mailer_info_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     CONST char *text;
     struct addr_item *result;
{
    if (!submission_verify_address) {
	DPRINT(Debug,13,(&Debug, 
			 "submission_mailer_info_verify_addr=-2 (Verify disabled)\n"));
	return -2;       /* Cannot VRFY user */		
    }

    return smtp_tail_verify_addr(M,I,text,result);
}

static int MO_submission_verify P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_submission_verify (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	     submission_verify_address = 1;
	else if (0 == strcmp(*value,no_yes[0]))
	    submission_verify_address = 0;
	else if (0 == strcmp(*value,no_yes[1])) 
	    submission_verify_address = 1;
	else {
	    lib_error(FRM("Bad verify=%s for submission"),
		      *value);
	    return 0;
	}	
	value_set = 1;
    } else {
	if (submission_verify_address) {
	    *value = no_yes[1];
	    return 1;
	}
	*value = no_yes[0]; /* Only show if verify=no or
			       verify=yes is given at least once 
			    */
    }    
    return value_set;
}


static struct mailer_option_list SUBMISSION_options[] = {
    { "server",   &submission_server,                MO_default_value, 0 },
    { "verify",   NULL,                              MO_submission_verify, 0 },
    { "allow-set-sender",NULL,                       MO_allow_set_sender, 0 },
    { NULL, NULL,                                    MO_default_value, 0 }
};

/* sendmail-bs mailer ----------------------------------------------------- */

#if defined(BACKGROUD_PROCESSES) 
#define HAVE_SENDMAILBS 1
#endif

#if HAVE_SENDMAILBS

static char * sendmail_mailer_path    = SENDMAIL_MAILER_PATH;



S_(mailer_info_init_hook sendmailbs_mailer_info_init)
static int sendmailbs_mailer_info_init(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    CONST char * argv[3];
    char response_code[3];

    int options = SY_ENV_SHELL | SY_NOTTY;
    int ret,r;
    int pair[2];

    I->p.p1 = safe_malloc(sizeof (* (I->p.p1)));

    /* bzero is defined on hdrs/defs.h */
    bzero((void *)I->p.p1,sizeof (* (I->p.p1)));

    DPRINT(Debug,13,(&Debug,
		     "sendmailbs_mailer_info_init: mailer info, I=%p\n",I));

    zero_remote_account(& (I->p.p1->RA));	
    
    r = socketpair(PF_UNIX,SOCK_STREAM,AF_UNIX    /* ?? */, pair);
    
    if (-1 == r) {
	int err = errno;
	lib_error(FRM("socketpair failed: %s"),
		  error_description(err));

	free(I->p.p1);
	I->p.p1 = NULL;
	return 0;
    }

    argv[0] = sendmail_mailer_path;
    argv[1] = "-bs";
    argv[2] = NULL;
	
    /* Connect standard imput and output to other end of socket ... */

    ret = start_run( & (I->p.p1->RS),options,argv,pair[1],pair[1]);

    /* Close other end of socket so it is used only by sendmail */
    close(pair[1]);
    pair[1] = -1;

    if (!ret) {
	if (I->p.p1->RS.save_errno)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailErrno,
			      "Failed: %.30s: %.40s"),
		      argv[0],error_description(I->p.p1->RS.save_errno));
	else
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCantStart,
			      "Can't start %.30s"),
		      argv[0]);

	close(pair[0]); 
	pair[0] = -1;

	free(I->p.p1);
	I->p.p1 = NULL;
	return 0;
    }

    if (transaction_file) {
	time_t tmp = time(NULL);
	struct tm * zz = localtime(&tmp);
	fprintf(transaction_file,
		"%d [%d] %02d:%02d:%02d === CONNECT command %s, pid %d\n",
		getpid(),pair[0] ,
		zz ? zz->tm_hour : 00,
		zz ? zz->tm_min  : 00,
		zz ? zz->tm_sec  : 00,
		sendmail_mailer_path,
		I->p.p1->RS.pid);
    }

    if (-1 == fcntl(pair[0],F_SETFL,O_NONBLOCK)) {
	int err = errno;
	DPRINT(Debug,12,(&Debug,
			 "sendmailbs_mailer_info_init: fcntl [O_NONBLOCK] failed: %s\n",
			 error_description(err)));
    }

    I->p.p1->RA.stream = returnSimpleStream(pair[0]);
    I->p.p1->state = smtp_wait_reply;

    ConfigStream(I->p.p1->RA.stream,
		 smtp_read_action,ss_noaction_routine,smtp_timeout_action,
		 SMTP_TIMEOUT,I->p.p1);
		 
    zero_Read_Buffer(& (I->p.p1->read_buffer));
    zero_Write_Buffer(& (I->p.p1->write_buffer));
    I->p.p1->response_code[0]  = '\0';
    I->p.p1->response_text     = NULL;
    I->p.p1->response_text_len = 0;
    I->p.p1->SMTP_flags        = SMTP_submission;

    I->p.p1->incomplete_command     = NULL;
    I->p.p1->incomplete_command_len = 0;

    I->p.p1->ESMPT_keywords        = NULL;
    I->p.p1->ESMPT_keyword_len     = -1;   /* -1 == EHLO not issued yet */

    if (!smtp_command_ok(I->p.p1,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "sendmailbs_mailer_info_init: Failed to get greeting from sendmail -bs\n"));

	/* Caller will call sendmailbs_mailer_info_close ... */
	return 0;
    }
    smtp_command_reset(I->p.p1);

    return 1;
}

S_(mailer_info_close_hook sendmailbs_mailer_info_close)
static void sendmailbs_mailer_info_close(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{

    DPRINT(Debug,13,(&Debug,
		     "sendmailbs_mailer_info_close: mailer info, I=%p\n",I));

    if (I->p.p1) {
	int ret,exit_code;

	ret = run_already_done(& (I->p.p1->RS),&exit_code);

	if (0 == ret && smtp_idle == I->p.p1->state) {
	    int status = smtp_start_command(I->p.p1,"QUIT");

	    if (status) {
		smtp_command_ready(I->p.p1,0);
		
		/* Wait for QUIT */	       
		smtp_command_reset(I->p.p1);
	    }
	}

	free_remote_account(& (I->p.p1->RA));	
	
	if (0 == ret) 
	    ret = wait_end(&(I->p.p1->RS),&exit_code);

	if (ret < 0) {
	    DPRINT(Debug,4,(&Debug,
			    "%s died on signal %d\n",
			    sendmail_mailer_path,-ret));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === DIED command %s, pid %d, signal %d\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid,
			-ret);
	    }
	} else if (0 == ret) {
	    DPRINT(Debug,4,(&Debug,
			    "%s lost?\n",
			    sendmail_mailer_path));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === LOST command %s, pid %d\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid);
	    }
	} else {
	    DPRINT(Debug,4,(&Debug,
			    "%s exited with %d%s\n",
			    sendmail_mailer_path,
			    exit_code, exit_code == 0 ? " (OK)" : ""));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === EXITED command %s, pid %d, exit code %d%s\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid,
			exit_code,
			exit_code == 0 ? " (OK)" : "");
	    }
	}

	free_Read_Buffer(& (I->p.p1->read_buffer));
	free_Write_Buffer(& (I->p.p1->write_buffer));

	free_response_text(I->p.p1);

	if (I->p.p1->incomplete_command)
	    free(I->p.p1->incomplete_command);

	I->p.p1->incomplete_command     = NULL;
	I->p.p1->incomplete_command_len = 0;

	if (I->p.p1->ESMPT_keywords) {
	    int i;
	    for (i = 0; i < I->p.p1->ESMPT_keyword_len; i++) {
		if (I->p.p1->ESMPT_keywords[i])
		    free(I->p.p1->ESMPT_keywords[i]);
		I->p.p1->ESMPT_keywords[i] = NULL;
	    }
	    free(I->p.p1->ESMPT_keywords);
	}
	I->p.p1->ESMPT_keywords        = NULL;
	I->p.p1->ESMPT_keyword_len     = -1;   /* -1 == EHLO not issued yet */

	/* bzero is defined on hdrs/defs.h */
	bzero((void *)I->p.p1,sizeof (* (I->p.p1)));
	free(I->p.p1);
	I->p.p1 = NULL;
    } 
}

static int sendmailbs_verify_address = 0;

S_(mailer_info_verify_addr sendmailbs_mailer_info_verify_addr)
static int sendmailbs_mailer_info_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     CONST char *text;
     struct addr_item *result;
{
    if (!sendmailbs_verify_address) {
	DPRINT(Debug,13,(&Debug, 
			 "sendmailbs_mailer_info_verify_addr=-2 (Verify disabled)\n"));
	return -2;       /* Cannot VRFY user */		
    }

    return smtp_tail_verify_addr(M,I,text,result);
}

static int MO_sendmailbs_verify P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_sendmailbs_verify (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	     sendmailbs_verify_address = 1;
	else if (0 == strcmp(*value,no_yes[0]))
	    sendmailbs_verify_address = 0;
	else if (0 == strcmp(*value,no_yes[1])) 
	    sendmailbs_verify_address = 1;
	else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSendmailBadVerify,
			      "Bad verify=%s for sendmail"),
		      *value);
	    return 0;
	}	
	value_set = 1;
    } else {
	if (sendmailbs_verify_address) {
	    *value = no_yes[1];
	    return 1;
	}
	*value = no_yes[0]; /* Only show if verify=no or
			       verify=yes is given at least once 
			    */
    }    
    return value_set;
}


static struct mailer_option_list SENDMAILBS_options[] = {
    { "path",     &sendmail_mailer_path,             MO_default_value, 0 },
    { "verify",   NULL,                              MO_sendmailbs_verify, 0 },
    { "allow-set-sender",NULL,                       MO_allow_set_sender,  0 },
    { NULL, NULL,                                    MO_default_value, 0 }
};

#endif

static struct mailer_config MAILERS[] = {
    { "submission", &(SUBMISSION_options[0]), default_set_option, 
      smtp_common_mailer_init, smtp_common_mailer_close, 
      smtp_common_mailer_backend,
      NULL, 
      submission_mailer_info_init, submission_mailer_info_close,
      smtp_common_mailer_info_query, submission_mailer_info_verify_addr,
      smtp_common_mailer_info_gen_def_ef, smtp_common_mailer_info_set_ef,
      
      MB_USE_DOMAIN|MB_REQ_DEFAULT_SENDER
    },
#if HAVE_SENDMAILBS
    { "sendmail-bs", &(SENDMAILBS_options[0]), default_set_option, 
      smtp_common_mailer_init, smtp_common_mailer_close, 
      smtp_common_mailer_backend,
      &sendmail_mailer_path, 
      sendmailbs_mailer_info_init, sendmailbs_mailer_info_close,
      smtp_common_mailer_info_query, sendmailbs_mailer_info_verify_addr,
      smtp_common_mailer_info_gen_def_ef, smtp_common_mailer_info_set_ef,
      
      MB_USE_DOMAIN|MB_REQ_DEFAULT_SENDER
    },
#endif

};

#if ANSI_C
provides_shared_MCF_f provides_shared_MCF;
#endif
struct mailer_config * provides_shared_MCF(count,s_size,opt_size)
     int *count; 
     size_t *s_size; 
     size_t *opt_size;
{
    *s_size    = sizeof (MAILERS[0]);
    *opt_size  = sizeof (SENDMAILBS_options[0]);
    *count  = (sizeof MAILERS) / sizeof (MAILERS[0]);
    

    return &(MAILERS[0]);
}


#if 0

/* elmrc options ---------------------------------------------------------- */

#include "../../hdrs/rc_imp.h"
#include "../../hdrs/save_opts.h"

#endif


#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
