static char rcsid[] = "@(#)$Id: delete.c,v 1.12 2003/07/29 18:46:19 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.12 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *
 *****************************************************************************/

/**  Delete or undelete files: just set flag in header record! 
     Also tags specified message(s)...

**/

#include "def_elm.h"
#include "s_elm.h"

void delete_msg(real_del, update_screen, mailbox)
     int real_del, update_screen;
          struct MailboxView *mailbox;
{
    /** Delete current message.  If real-del is false, then we're
	actually requested to toggle the state of the current
	message... 
    **/
    int current = 0;

    if (real_del) {

	if (inalias) {
	    current = alias_current;

	    if (aliases[alias_current-1]->type & SYSTEM)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoDelSysAlias,
				  "Can't delete a system alias!"));
	    else
		setit(aliases[alias_current-1]->status, DELETED);

	} else {
	    struct header_rec * hdr;

	    current = get_current(mailbox);
	    hdr = give_header(mailbox,current-1);
	    
	    if (hdr)
		setit(hdr->status, DELETED);
	}
    } else {

	if (inalias) {
	    current = alias_current;

	    if (aliases[alias_current-1]->type & SYSTEM)
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoDelSysAlias,
				  "Can't delete a system alias!"));
	    else if (ison(aliases[alias_current-1]->status, DELETED))
		clearit(aliases[alias_current-1]->status, DELETED);
	    else
		setit(aliases[alias_current-1]->status, DELETED);

	} else {
	    struct header_rec * hdr;

	    current = get_current(mailbox);
	    hdr = give_header(mailbox,current-1);

	    if (hdr) {
		if (ison(hdr->status, DELETED))
		    clearit(hdr->status, DELETED);
		else
		    setit(hdr->status, DELETED);
	    }
	}
    }

    if (update_screen)
	show_msg_status(current-1,mailbox);
}

void unread_msg(update_screen, mailbox)
     int update_screen;
     struct MailboxView *mailbox;
{
    int current = get_current(mailbox);
    struct header_rec * hdr = give_header(mailbox,current-1);

  	/** change the message flag to unread **/
   
    if (hdr) {
 	setit(hdr->status, UNREAD);
	hdr->status_chgd = TRUE;
    }

    if (update_screen)
	show_msg_status(current-1,mailbox);
}

void undelete_msg(update_screen,mailbox)
     int update_screen;
     struct MailboxView *mailbox;
{
    int current = 0;

    /** clear the deleted message flag **/
    
    if (inalias) {
	current = alias_current;
	clearit(aliases[alias_current-1]->status, DELETED);
    } else {
	struct header_rec *hdr;

	current = get_current(mailbox);
	hdr = give_header(mailbox,current-1);

	if (hdr) {
	    clearit(hdr->status, DELETED);
	    hdr->status_chgd = TRUE;
	}
    }

    if (update_screen)
	show_msg_status(current-1,mailbox);
}

void show_msg_status(msg, mailbox)
     int msg;
     struct MailboxView *mailbox;
{
    /** show the status of the current message only.  **/

    char tempbuf[3];
    int current = 0;

    if (inalias) {
	current = alias_current;
	strfcpy(tempbuf,show_alias_status(aliases[msg]), sizeof tempbuf);
    } else {
	struct header_rec *hdr = give_header(mailbox,msg);

	current = get_current(mailbox);
	if (hdr)
	    strfcpy(tempbuf,show_message_status(hdr), sizeof tempbuf);
	else
	    tempbuf[0] = '\0';
    }

    if (on_page(msg, mailbox)) {
	MoveCursor(((compute_visible(msg+1, mailbox)-1) 
		    % headers_per_page) + 4, 2);
	if (msg+1 == current && !arrow_cursor) {
	    StartBold();
	    Writechar( tempbuf[0] );
	    EndBold();
	}
	else
	    Writechar( tempbuf[0] );
    }
}

int tag_message(update_screen, mailbox)
     int update_screen;
     struct MailboxView *mailbox;
{
    /** Tag current message and return TRUE.
	If already tagged, untag it and return FALSE. **/
    
	int istagged;
	int current = 0;

	if (inalias) {
	    current = alias_current;

	    if (ison(aliases[alias_current-1]->status, TAGGED)) {
		clearit(aliases[alias_current-1]->status, TAGGED);
		istagged = FALSE;
	    } else {
		setit(aliases[alias_current-1]->status, TAGGED);
		istagged = TRUE;
	    }
		
	} else {
	    struct header_rec *hdr;
	    current = get_current(mailbox);
	    
	    hdr = give_header(mailbox,current-1);

	    if (hdr) {
		if (ison(hdr->status, TAGGED)) {
		    clearit(hdr->status, TAGGED);
		    hdr->status_chgd = TRUE;
		    istagged = FALSE;
		} else {
		    setit(hdr->status, TAGGED);
		    hdr->status_chgd = TRUE;
		    istagged = TRUE;
		}
	    } else
		istagged = FALSE;
	}

	if(update_screen)
	    show_msg_tag(current-1,mailbox);
	return(istagged);
}

void show_msg_tag(msg, mailbox)
     int msg;
     struct MailboxView *mailbox;
{
    /** show the tag status of the current message only.  **/

    int current =  
	inalias ? alias_current : get_current(mailbox);

    if (on_page(msg, mailbox)) {

	MoveCursor(((compute_visible(msg+1, mailbox)-1) % 
		      headers_per_page) + 4, 4);

	if (msg+1 == current && !arrow_cursor) {

	    StartBold();

	    if (inalias) {
		Writechar(ison(aliases[msg]->status, TAGGED)? '+' : ' ');
	    } else {
		struct header_rec *hdr = give_header(mailbox,msg);
		if (hdr) {
		    Writechar(ison(hdr->status, TAGGED)? '+' : ' ');
		}
	    }

	    EndBold();
	} else {
	    if (inalias) {
		Writechar(ison(aliases[msg]->status, TAGGED)? '+' : ' ');
	    } else {
		struct header_rec *hdr = give_header(mailbox,msg);
		if (hdr) {
		    Writechar(ison(hdr->status, TAGGED)? '+' : ' ');
		}
	    }
	}
    }
}

void show_new_status(msg,mailbox)
     int msg;
     struct MailboxView *mailbox;
{
    /** If the specified message is on this screen, show
	the new status (could be marked for deletion now,
	and could have tag removed...)
    **/

    if (on_page(msg, mailbox)) {
	char * s = "";
	char   c = ' ';
	int current =  0;
    
	if (inalias) {  
	    current = alias_current;

	    s = show_alias_status(aliases[msg]);
	    c = ison(aliases[msg]->status, TAGGED ) ? '+' : ' ';

	} else {
	    struct header_rec *hdr = give_header(mailbox,msg);
	    current = get_current(mailbox);
	    
	    if (hdr) {
		s = show_message_status(hdr);
		c = ison(hdr->status, TAGGED ) ? '+' : ' ';
	    }
	}

	if (msg+1 == current && !arrow_cursor) {
	    StartBold();
	    
	    PutLineX(((compute_visible(msg+1,mailbox)-1) 
		      % headers_per_page) + 4,
		     2, FRM("%s%c"), 
		     s,c);

	    EndBold();
	  }
	  else
	    PutLineX(((compute_visible(msg+1,mailbox)-1) 
		      % headers_per_page) + 4,
		     2, FRM("%s%c"), 
		     s,c);
	
    }
}

int DeleteTagged(mailbox)
     struct MailboxView *mailbox;
{

    /** delete all tagged messages.  return TRUE if the current message was
	deleted. 
    **/

    int iindex, count = 0;
       
    if (!inalias) {
	int mc      = get_message_count(mailbox);
	int current = get_current(mailbox);
	struct header_rec * hdr;

	for (iindex=0; iindex < mc; iindex++) {
	    struct header_rec * hdr = give_header(mailbox,iindex);
		
	    if (hdr && (hdr->status & TAGGED)) {
		hdr->status |= DELETED;
		count++;
		show_new_status(iindex,mailbox);
	    }
	}

	if (count == 1)
	    print_format_center(elm_LINES,
				CATGETS(elm_msg_cat, ElmSet, ElmMsgDeleted,
					"Message deleted."));
	else
	    print_format_center(elm_LINES,
				CATGETS(elm_msg_cat, ElmSet, ElmMsgsDeleted,
					"%d messages deleted."), 
				count);
	  
	hdr = give_header(mailbox,current-1);
	
	if (hdr)
	    return(hdr->status & DELETED ? 1 : 0);
  }

  return FALSE;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */

