static char rcsid[] = "@(#)$Id: limit.c,v 1.20 2003/09/03 09:16:29 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.20 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This stuff is inspired by MH and dmail and is used to 'select'
    a subset of the existing mail in the folder based on one of a
    number of criteria.  The basic tricks are pretty easy - we have
    as status of VISIBLE associated with each header stored in the
    (er) mind of the computer (!) and simply modify the commands to
    check that flag...the global variable `selected' is set to the
    number of messages currently selected, or ZERO if no select.
**/

#include "def_elm.h"
#include "s_elm.h"
#include "s_aliases.h"

DEBUG_VAR(Debug,__FILE__,"mbox");

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

#define TO		1
#define FROM		2
#define CC              3

char *shift_lower();
void ClearSelection();

static void limit_selection P_((int based_on, struct string *pattern, 
				int additional_criteria,
				struct MailboxView *mailbox));
static void limit_by_string_subject P_((const struct string *pattern,
					int additional_criteria,
					struct MailboxView *mailbox));
static int limit_alias_selection P_((int based_on, char *pattern, 
				     int additional_criteria));
static int SelectTaggedMessages P_((struct MailboxView *mailbox));
static int TallySelection P_((struct MailboxView *mailbox));


static void limit_by_header P_((char *header, char *value,
				int additional_criteria,
				struct MailboxView *mailbox));


int limit(mailbox)
     struct MailboxView *mailbox;
    
{
    /** returns non-zero if we changed selection criteria = need redraw **/
	
    char criteria[STRING], first[STRING], rest[STRING];
    static char *prompt = NULL;
    int  last_selected, all;
    
    last_selected = ifmain(get_selected(mailbox),alias_selected);

	all = 0;
	if (prompt == NULL) {
	  prompt = catgets(elm_msg_cat, ElmSet, ElmLimitEnterCriteria,
		"Enter criteria or '?' for help: ");
	}

	if (ifmain(get_selected(mailbox),alias_selected)) {
	    PutLineX(elm_LINES-2, 0, 
		     CATGETS(elm_msg_cat, ElmSet, ElmLimitAlreadyHave,
			     "Already have selection criteria - add more? (%c/%c) %c%c"),
		     *def_ans_yes, *def_ans_no, *def_ans_no, BACKSPACE);
	    criteria[0] = ReadCh(0);
	    if (
#ifdef ASCII_CTYPE
		isascii(criteria[0]) &&
#endif
		tolower((unsigned char)criteria[0]) == *def_ans_yes) {
		Write_to_screen(CATGETS(elm_msg_cat, ElmSet, ElmYesWord, "Yes."));
		PutLineX(elm_LINES-3, elm_COLUMNS-30, 
			 CATGETS(elm_msg_cat, ElmSet, ElmLimitAdding,
				 "Adding criteria..."));
	  } else {
	      Write_to_screen(CATGETS(elm_msg_cat, ElmSet, ElmNoWord, "No."));
	      PutLineX(elm_LINES-3, elm_COLUMNS-30, 
		       CATGETS(elm_msg_cat, ElmSet, ElmLimitChanging,
			       "Change criteria..."));

	      if (inalias)
		  alias_selected = 0;
	      else
		  set_selected(mailbox,0);
	  }
	}

	while(1) {
	  PutLine0(elm_LINES-2, 0, prompt);
	  CleartoEOLN();

	  criteria[0] = '\0';
	  if (0 != optionally_enter(criteria, elm_LINES-2, strlen(prompt), 0,
				    sizeof criteria))
	    return TRUE;

	  clear_error();
	  
	  if (strlen(criteria) == 0) {
	    /* no change */

	      if (inalias)
		  alias_selected = last_selected;
	      else
		  set_selected(mailbox,last_selected);

	    return(FALSE);	
	  }

	  /* NOTE: split_word lowercases string when copying */
	  split_word(criteria, first, rest);

	  if (inalias) {
	      if (equal(first, "?")) {
		  if (last_selected)
		      lib_transient(CATGETS(elm_msg_cat, AliasesSet, 
				  AliasesEnterLastSelected,
				  "Enter:{\"name\",\"alias\"} [pattern] OR {\"person\",\"group\",\"user\",\"system\"} OR \"all\""));
		  else
		      lib_transient(CATGETS(elm_msg_cat, AliasesSet, 
					    AliasesEnterSelected,
					    "Enter: {\"name\",\"alias\"} [pattern] OR {\"person\",\"group\",\"user\",\"system\"}"));
		  continue;
	      }
	      else if (equal(first, "all")) {
		  all++;

		  alias_selected = 0;
		  
	      }
	      else if (equal(first, "name"))
		  alias_selected = limit_alias_selection(BY_NAME, rest, alias_selected);
	      else if (equal(first, "alias"))
		  alias_selected = limit_alias_selection(BY_ALIAS, rest, alias_selected);
	      else if (equal(first, "person"))
		  alias_selected = limit_alias_selection(PERSON, rest, alias_selected);
	      else if (equal(first, "group"))
		  alias_selected = limit_alias_selection(GROUP, rest, alias_selected);
	      else if (equal(first, "user"))
		  alias_selected = limit_alias_selection(USER, rest, alias_selected);
	      else if (equal(first, "system"))
		  alias_selected = limit_alias_selection(SYSTEM, rest, alias_selected);
	      else {
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmLimitNotValidCriterion,
				    "\"%s\" not a valid criterion."), 
			    first);
		  continue;
	      }
	      break;
	  }
	  else {
	    if (equal(first, "?")) {
	      if (last_selected)
	        lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmEnterLastSelected,
				      "Enter: {\"subject\",\"to\",\"from\",\"cc\",\"header\"} [pattern] OR {\"thread\",\"tagged\",\"all\"}"));
	      else
	        lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmEnterSelected,
				      "Enter: {\"subject\",\"to\",\"from\",\"header\"} [pattern] OR \"thread\",\"tagged\"}"));
	      continue;
	    }
	    else if (equal(first, "all")) {
	      all++;
	      set_selected(mailbox,0);
	    }
	    else if (equal(first, "tagged"))
		set_selected(mailbox,SelectTaggedMessages(mailbox));
	    else {
	      char *c, *p, buf[STRING], buf2[STRING];
	      int complx = 0, err = 0, iter = 0;

	      if (!get_selected(mailbox))
		  ClearSelection(mailbox);
		
	      /* this next bit of code assumes that the input string is a
		 list of "and"ed or "or"ed items.  NOTE: it assumes that
		 if the first operator is "and", then all operators are "and",
		 and vice-versa. */
	      strfcpy(buf, first, sizeof buf);
	      strfcat(buf, " ", sizeof buf);
	      strfcat(buf, rest, sizeof buf);
	      p = buf;
	      if (index(p, '&') != NULL)
		complx = 1;
	      while ((c = strtok(p, "|&")) != NULL) {
		while (*c && isspace(*c)) ++c;
		p = c;
		while (*p && !isspace(*p)) ++p;
		while (*p && isspace(*p)) ++p;
		strfcpy(buf2, p, sizeof buf2);
		remove_possible_trailing_spaces(buf2);
		if (strncmp(c, "subj", 4) == 0) {
		    struct string *X = new_string2(display_charset,s2us(buf2));

		    limit_selection(SUBJECT, X, get_selected(mailbox)||(complx&&iter!=0),
				    mailbox);

		    free_string(&X);
		} else if (strncmp(c, "to", 2) == 0) {
		    struct string *X = new_string2(display_charset,s2us(buf2));

			limit_selection(TO, X, get_selected(mailbox)||(complx&&iter!=0),
					mailbox);

		    free_string(&X);
		} else if (strncmp(c, "cc", 2) == 0) {
		    struct string *X = new_string2(display_charset,s2us(buf2));
		    
		    limit_selection(CC, X, get_selected(mailbox)||(complx&&iter!=0),
				    mailbox);
		    free_string(&X);
		} else if (strncmp(c, "from", 4) == 0) {
		    struct string *X = new_string2(display_charset,s2us(buf2));

		    limit_selection(FROM, X, get_selected(mailbox)||(complx&&iter!=0),
				    mailbox);

		    free_string(&X);

		} else if (strncmp(c, "header", 6) == 0) {
		    char * A = strchr(buf2,':');

		    if (A) {
			*A = '\0';
			A++;
			while (*A && isspace(*A)) ++A;
		    }

		    limit_by_header(buf2,A,
				    get_selected(mailbox)||(complx&&iter!=0),
				    mailbox);

		} else if (strncmp(c, "thread", 6) == 0) {
		    int current = get_current(mailbox);
		    struct header_rec * hdr = give_header(mailbox,current-1);

		    struct string * temp;

		    if (!hdr)
			continue;   
		    if (!hdr->subject)
			break;
		    
		    temp = 
			skip_ascii_head_from_string(hdr->
						    subject,
						    s2us("Re: "),1);

		    limit_by_string_subject(temp,get_selected(mailbox)||(complx&&iter!=0),
					    mailbox);
		    free_string(&temp);
		  }
		else {
		    if ((p = index(c, ' ')) != NULL)
			*p = '\0';
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmLimitNotValidCriterion,
				    "\"%s\" not a valid criterion."), 
			    c);
		  err = 1;
		  break;
		}
		p = NULL;
		++iter;
	      }
	      if (err)
		continue;
	      set_selected(mailbox,TallySelection(mailbox));
	    }
	    break;
	  }
        }

	if (all && last_selected)
	  lib_error(CATGETS(elm_msg_cat, ElmSet, ElmLimitReturnToUnlimited,
			    "Returned to unlimited display."));
	else {
	  if (inalias) {
	    if (alias_selected > 1)
	      lib_error(CATGETS(elm_msg_cat, AliasesSet,
				AliasesLimitMessagesSelected, 
				"%d aliases selected."),
			alias_selected);
	    else if (alias_selected == 1)
	      lib_error(CATGETS(elm_msg_cat, AliasesSet,
				AliasesLimitMessageSelected, 
				"1 alias selected."));
	    else
	      lib_error(CATGETS(elm_msg_cat, AliasesSet,
				AliasesLimitNoMessagesSelected, 
				"No aliases selected."));
	  }
	  else {
	    if (get_selected(mailbox) > 1)
	      lib_error(CATGETS(elm_msg_cat, ElmSet,
				ElmLimitMessagesSelected, 
				"%d messages selected."), 
			get_selected(mailbox));
	    else if (get_selected(mailbox) == 1)
	      lib_error(CATGETS(elm_msg_cat, ElmSet, 
				ElmLimitMessageSelected,
				"1 message selected."));
	    else
	      lib_error(CATGETS(elm_msg_cat, ElmSet,
				ElmLimitNoMessagesSelected, 
				"No messages selected."));
	  }
	}
       
	/* we need a redraw if there had been a selection or there is now. */
	if (last_selected || ifmain(get_selected(mailbox),
				    alias_selected)) {

	    /* if current message won't be on new display, 
	       go to first message */	    
	    if (inalias) { 
		if (alias_selected && !(aliases[alias_current-1]->status & VISIBLE))
		    alias_current = visible_to_index(1,NULL)+1;

	    } else {
		int current = get_current(mailbox);
		struct header_rec * hdr = give_header(mailbox,current-1);

		if (!hdr || get_selected(mailbox) && 
		    !(hdr->status & VISIBLE)) {
		    current = visible_to_index(1,mailbox)+1;		
		    set_current(mailbox,current);
		}
	    }
	    
	    return(TRUE);
	} else {
	    return(FALSE);
	}
}

static void limit_by_string_subject(pattern,additional_criteria,mailbox)
     CONST struct string *pattern;
     int additional_criteria;
     struct MailboxView *mailbox;
{
    int iindex;	
    int mc = get_message_count(mailbox);

    for (iindex = 0 ; iindex < mc ; iindex++) {
	struct header_rec * hdr = give_header(mailbox,iindex);
	int match = 0;
	
	if (hdr && hdr->subject)
	    match = find_pattern_from_string(hdr->subject,
					     pattern,1);
	

	if (additional_criteria) {
	    if (hdr->status & VISIBLE) {
		if (!match)
		    hdr->status &= ~VISIBLE;
	    }
	}
	else if (match)
	    hdr->status |= VISIBLE;
    }
}

static void limit_by_header(header,value, additional_criteria, mailbox)
     char *header; 
     char *value;
     int additional_criteria;
     struct MailboxView *mailbox;
{
    int message_number;
    int mc = get_message_count(mailbox);

    for (message_number = 0 ; message_number < mc ; message_number++) {
	struct header_rec *hdr;
	FILE *ZZ;

	int val = (message_number+1) / (float) mc * 100;
	
	if ((message_number+1) % readmsginc == 0 ||
	    val % readdatapercentinc       == 0) {
	    
	    PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
					   ElmSearchingInMessage,
					   "Searching %d (%02d%%) "),
		     (message_number+1), val);
	    
	    CleartoEOLN();
	    
	}

	if (give_message_data(mailbox,message_number,
			      &hdr,&ZZ,NULL,NO_mime_parse)) {

	    int match = 0;
	    header_list_ptr all_headers;

	    if (skip_envelope(hdr, ZZ) == -1) 
		continue;

	    all_headers = file_read_headers(ZZ,
					    RHL_CHECK_HEADER);

	    if (all_headers) {
		header_list_ptr walk;

		for (walk = locate_header_by_name(all_headers,header);
		     walk;
		     walk = walk->next_this_header) {

		    if (in_string(shift_lower(walk->body),value))
			match++;
		}
		delete_headers(&all_headers);
	    }

	    if (additional_criteria) {
		if (hdr->status & VISIBLE) {
		    if (!match)
			hdr->status &= ~VISIBLE;
		}
	    }
	    else if (match)
		hdr->status |= VISIBLE;
	    
	}
    }

    PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
				   ElmSearchingInMessageDone,
				   "Searching %d. Done "),
	     (message_number+1));

}

static void limit_selection(based_on, pattern, additional_criteria,mailbox)
     int based_on, additional_criteria;
     struct string *pattern;
     struct MailboxView *mailbox;
{
    int iindex;
    int mc = get_message_count(mailbox);

    for (iindex = 0 ; iindex < mc ; iindex++) {
	struct header_rec * hdr = give_header(mailbox,iindex);
	
	int match;
	
	switch (based_on) {
	    
	case FROM:
	    match = from_matches(hdr,pattern);
	    break;
	case TO:
	    match = to_matches(hdr,pattern);
	    break;
	case CC:
	    match = cc_matches(hdr,pattern);
	    break;
	case SUBJECT:
	default:
	    match = subject_matches(hdr,pattern);
	    break;
	}

	if (additional_criteria) {
	    if (hdr->status & VISIBLE) {
		if (!match)
		    hdr->status &= ~VISIBLE;
	    }
	}
	else if (match)
	    hdr->status |= VISIBLE;
    }
}

static int limit_alias_selection(based_on, pattern, additional_criteria)
     int based_on, additional_criteria;
     char *pattern;
{
	/** Given the type of criteria, and the pattern, mark all
	    non-matching aliases as ! VISIBLE.  If additional_criteria,
	    don't mark as visible something that isn't currently!
	**/

	register int iindex, count = 0;

	DPRINT(Debug,2,(&Debug,  
			"\n\n\n**limit on %d - '%s' - (%s) **\n\n",
			based_on, pattern, 
			additional_criteria?"add'tl":"base"));

	if (based_on == BY_NAME) {
	  for (iindex = 0; iindex < alias_message_count; iindex++)
	    if (! in_string(shift_lower(aliases[iindex]->name), pattern))
	      clearit(aliases[iindex]->status, VISIBLE);
	    else if (additional_criteria && 	
		     !(aliases[iindex]->status & VISIBLE))
	      clearit(aliases[iindex]->status, VISIBLE);	/* shut down! */
	    else { /* mark it as readable */
	      setit(aliases[iindex]->status, VISIBLE);
	      count++;
	      DPRINT(Debug,5,(&Debug, 
			      "  Alias %d (%s, %s) marked as visible\n",
			      iindex, aliases[iindex]->alias,
			      aliases[iindex]->name));
	    }
	}
	else if (based_on == BY_ALIAS) {
	  for (iindex = 0; iindex < alias_message_count; iindex++)
	    if (! in_string(shift_lower(aliases[iindex]->alias), pattern))
	      clearit(aliases[iindex]->status, VISIBLE);
	    else if (additional_criteria && 	
		     !(aliases[iindex]->status & VISIBLE))
	      clearit(aliases[iindex]->status, VISIBLE);	/* shut down! */
	    else { /* mark it as readable */
	      setit(aliases[iindex]->status, VISIBLE);
	      count++;
	      DPRINT(Debug,5,(&Debug, 
			      "  Alias %d (%s, %s) marked as visible\n",
			      iindex, aliases[iindex]->alias,
			      aliases[iindex]->name));
	    }
	}
	else {
	  for (iindex = 0; iindex < alias_message_count; iindex++)
	    if (! (based_on & aliases[iindex]->type))
	      clearit(aliases[iindex]->status, VISIBLE);
	    else if (additional_criteria && 	
		     !(aliases[iindex]->status & VISIBLE))
	      clearit(aliases[iindex]->status, VISIBLE);	/* shut down! */
	    else { /* mark it as readable */
	      setit(aliases[iindex]->status, VISIBLE);
	      count++;
	      DPRINT(Debug,5,(&Debug, 
			      "  Alias %d (%s, %s) marked as visible\n",
			      iindex, aliases[iindex]->alias,
			      aliases[iindex]->name));
	    }
	}

	DPRINT(Debug,4,(&Debug,  
			"\n** returning %d selected **\n\n\n", count));

	return(count);
}

int next_message(iindex, skipdel, mailbox)
     int iindex, skipdel;
     struct MailboxView *mailbox;
{
	/** Given 'iindex', this routine will return the actual iindex into the
	    array of the NEXT message, or '-1' iindex is the last.
	    If skipdel, return the iindex for the NEXT undeleted message.
	    If selected, return the iindex for the NEXT message marked VISIBLE.
	**/

	int remember_for_debug, stat;
	int mc;

	if (iindex < 0) return(-1);	/* invalid argument value! */

	remember_for_debug = iindex;

	mc = ifmain(get_message_count(mailbox),alias_message_count);
	for(iindex++;
	    iindex < mc; 
	    iindex++) {

	    if (inalias) {
		stat = aliases[iindex]->status;
	    } else {
		struct header_rec *hdr = give_header(mailbox,iindex);
		if (hdr)
		    stat = hdr->status;
		else 
		    stat = 0;
	    }
	    
	    if (((stat & VISIBLE) || (!ifmain(get_selected(mailbox),alias_selected)))
		&& (!(stat & DELETED) || (!skipdel))) {
		DPRINT(Debug,9,(&Debug,  
				"[Next%s%s: given %d returning %d]\n", 
				(skipdel ? " undeleted" : ""),
				(ifmain(get_selected(mailbox),alias_selected) ? " visible" : ""),
				remember_for_debug+1, iindex+1));
		return(iindex);
	    }
	}
	return(-1);
}

int prev_message(iindex, skipdel, mailbox)
     int iindex, skipdel;
     struct MailboxView *mailbox;
{
    /** Like next_message, but the PREVIOUS message. **/
    
    int remember_for_debug, stat;    
    int mc = ifmain(get_message_count(mailbox),alias_message_count);
    
    if (iindex >= mc) 
	return(-1);	/* invalid argument value! */
    
    remember_for_debug = iindex;
    
    for(iindex--; iindex >= 0; iindex--) {
	
	if (inalias) {
	    stat = aliases[iindex]->status;
	} else {
	    struct header_rec *hdr = give_header(mailbox,iindex);
	    if (hdr)
		stat = hdr->status;
	    else 
		stat = 0;
	}
	
	if (((stat & VISIBLE) || (!ifmain(get_selected(mailbox),alias_selected)))
	    && (!(stat & DELETED) || (!skipdel))) {
	    DPRINT(Debug,4,(&Debug, 
			    "[Previous%s%s: given %d returning %d]\n", 
			    (skipdel ? " undeleted" : ""),
			    (ifmain(get_selected(mailbox),alias_selected) ? " visible" : ""),
			    remember_for_debug+1, iindex+1));
	    return(iindex);
	}
    }
    return(-1);
}

int compute_visible(message,mailbox)
     int message;
     struct MailboxView *mailbox;
{
	/** return the 'virtual' iindex of the specified message in the
	    set of messages - that is, if we have the 25th message as
	    the current one, but it's #2 based on our limit criteria,
	    this routine, given 25, will return 2.
	**/

	int iindex, count = 0;

	if (! ifmain(get_selected(mailbox),alias_selected)) 
	    return(message);

	if (message < 1) message = 1;	/* normalize */

	for (iindex = 0; iindex < message; iindex++) {
	    int stat;

	    if (inalias) {
		stat = aliases[iindex]->status;
	    } else {
		struct header_rec *hdr = give_header(mailbox,iindex);
		if (hdr)
		    stat = hdr->status;
		else 
		    stat = 0;
	    }

	    if (stat & VISIBLE)
		count++;
	}
	
	DPRINT(Debug,4,(&Debug, 
			"[compute-visible: displayed message %d is actually %d]\n",
			count, message));
	
	return(count);
}

int visible_to_index(message,mailbox)
     int message;
     struct MailboxView *mailbox;
{
	/** Given a 'virtual' iindex, return a real one.  This is the
	    flip-side of the routine above, and returns (message_count+1)
	    if it cannot map the virtual iindex requested (too big) 
	**/

	int iindex = 0, count = 0;
	int mc = ifmain(get_message_count(mailbox),alias_message_count);

	for (iindex = 0; iindex < mc; iindex++) {
	    int stat;

	    if (inalias) {
		stat = aliases[iindex]->status;
	    } else {
		struct header_rec *hdr = give_header(mailbox,iindex);
		if (hdr)
		    stat = hdr->status;
		else 
		    stat = 0;
	    }

	    if (stat & VISIBLE)
		count++;
	    if (count == message) {
		DPRINT(Debug,4,(&Debug, 			       
				"visible-to-index: (up) index %d is displayed as %d\n",
				message, iindex));
		return(iindex);
	    }
	}
	
      	DPRINT(Debug,4,(&Debug,  "index %d is NOT displayed!\n", message));
	
	return mc+1;
}

void ClearSelection(mailbox) 
     struct MailboxView *mailbox;
{
    int iindex;
    int mc = get_message_count(mailbox);
    
    for (iindex = 0; iindex < mc; iindex++) {
	struct header_rec *hdr = give_header(mailbox,iindex);
	if (hdr)
	    hdr->status &= ~VISIBLE;
    }
}

static int TallySelection(mailbox) 
     struct MailboxView *mailbox;
{
    int iindex, count = 0;
    int mc = get_message_count(mailbox);
  
    for (iindex = 0 ; iindex < mc ; iindex++) {
	struct header_rec *hdr = give_header(mailbox,iindex);
	if (hdr &&
	    hdr->status & VISIBLE)
	    ++count;
    }
    return(count);
}

static int SelectTaggedMessages(mailbox)
     struct MailboxView *mailbox;
{
    int iindex, count = 0;
    int mc = get_message_count(mailbox);


    for (iindex = 0; iindex < mc; iindex++) {
	struct header_rec *hdr = give_header(mailbox,iindex);

	if (hdr) {
	    if (hdr->status & TAGGED) {
		hdr->status |= VISIBLE;
		++count;
	    } else
		hdr->status &= ~VISIBLE;
	}
    }
    return(count);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
