static char rcsid[] = "@(#)$Id: pattern.c,v 1.25 2003/09/03 09:16:30 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.25 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 ****************************************************************************/

/**    General pattern matching for the ELM mailer.     

**/

#include "def_elm.h"
#include "s_elm.h"

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

static CONST unsigned char *cs2us P_((const char *str));
static CONST unsigned char *cs2us(str) 
     CONST char *str;
{
    return (CONST unsigned char *)str;
}


#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

char *error_description(), *shift_lower();
static char *tag_word = NULL;
static char *tagged_word = NULL;
static char *Tagged_word = NULL;
static char *delete_word = NULL;
static char *mark_delete_word = NULL;
static char *Mark_delete_word = NULL;
static char *undelete_word = NULL;
static char *undeleted_word = NULL;
static char *Undeleted_word = NULL;

static int DoAll P_((int func,struct MailboxView *mailbox));
static int name_matches P_((int message_number, struct string *pat));
static int alias_matches P_((int message_number, struct string *pat));
static int comment_matches P_((int message_number, struct string *pat));
static int address_matches P_((int message_number, struct string *pat));

DEBUG_VAR(Debug,__FILE__,"mbox");

int meta_match(function, mailbox)
     int function;
     struct MailboxView *mailbox;
{
    char	ch, tagmsg[SLEN];
    
    /** Perform specific function based on whether an entered string 
	matches either the From or Subject lines.. 
	Return TRUE if the current message was matched, else FALSE.
    **/
    
    int i, tagged=0, count=0, curtag = 0;
    char		msg[SLEN];

    struct string * meta_pattern = NULL;
    int status;
    
    if (delete_word == NULL) {
	tag_word = catgets(elm_msg_cat, ElmSet, ElmTag, "Tag");
	tagged_word = catgets(elm_msg_cat, ElmSet, ElmTagged, "tagged");
	Tagged_word = catgets(elm_msg_cat, ElmSet, ElmCapTagged, "Tagged");
	delete_word = catgets(elm_msg_cat, ElmSet, ElmDelete, "Delete");
	mark_delete_word = catgets(elm_msg_cat, ElmSet, ElmMarkDelete, "marked for deletion");
	Mark_delete_word = catgets(elm_msg_cat, ElmSet, ElmCapMarkDelete, "Marked for deletion");
	undelete_word = catgets(elm_msg_cat, ElmSet, ElmUndelete, "Undelete");
	undeleted_word = catgets(elm_msg_cat, ElmSet, ElmUndeleted, "undeleted");
	Undeleted_word = catgets(elm_msg_cat, ElmSet, ElmCapUndeleted, "Undeleted");
    }
    
    PutLineX(elm_LINES-3, strlen(Prompt), CATGETS(elm_msg_cat, ElmSet, 
						  ElmMessagesMatchPattern,
						  "%s %s that match pattern..."), 
	     function==TAGGED?tag_word: function==DELETED?delete_word:undelete_word, items);
    
    if (function == TAGGED) {	/* are messages already tagged??? */
	int mc = ifmain(get_message_count(mailbox), alias_message_count);
	
	for (i=0; i < mc; i++) {
	    int stat = 0;
	    
	    if (inalias)
		stat = aliases[i]->status;
	    else {
		struct header_rec *hdr = give_header(mailbox,i);
		if (hdr)
		    stat = hdr->status;
	    }
	    
	    if (ison(stat,TAGGED))
		tagged++;
	}

	if (tagged) {
	    if (tagged > 1) {
		elm_sfprintf(tagmsg, sizeof tagmsg,
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmSomeMessagesATagged,
				     "Some %s are already tagged."), 
			     items);
		elm_sfprintf(msg, sizeof msg,
			     CATGETS(elm_msg_cat, ElmSet, ElmRemoveTags,
				     "%s Remove Tags? (%c/%c) "),
			     tagmsg, *def_ans_yes, *def_ans_no);
	    } else {
		elm_sfprintf(tagmsg, sizeof tagmsg,
			     CATGETS(elm_msg_cat, ElmSet, 
				     ElmAMessageATagged,
				     "One %s is already tagged."), 
			     item);
		elm_sfprintf(msg, sizeof msg,
			     CATGETS(elm_msg_cat, ElmSet, ElmRemoveTag,
				     "%s Remove Tag? (%c/%c) "),
			     tagmsg, *def_ans_yes, *def_ans_no);
	    }
	    
	    ch = want_to(msg, *def_ans_yes, elm_LINES-2, 0);
	    if (ch != *def_ans_no) {	/* remove tags... */
		for (i=0; i < mc; i++) {
		    if (inalias)
			clearit(aliases[i]->status,TAGGED);
		    else {
			struct header_rec *hdr = give_header(mailbox,i);
			if (hdr)
			    clearit(hdr->status,TAGGED);
		    }
		    show_new_status(i, mailbox);
		}
	    }
	}
    }
	

    status = optionally_enter2(&meta_pattern, elm_LINES-2, 0,
			       OE_REDRAW_MARK, 
			       CATGETS(elm_msg_cat, ElmSet, ElmEnterPattern, 
				       "Enter pattern: "));

    while (status == REDRAW_MARK) {
	status = optionally_enter2(&meta_pattern, elm_LINES-2, 0,
				   OE_REDRAW_MARK|OE_APPEND_CURRENT,
				   CATGETS(elm_msg_cat, ElmSet, 
					   ElmEnterPattern, 
					   "Enter pattern: "));
    }
    
    if (status != 0 || !meta_pattern ||
	string_len(meta_pattern) == 0) {

	if (meta_pattern)
	    free_string( & (meta_pattern));

	ClearLine(elm_LINES-2);
	return(curtag);
    }
    

    if (function == DELETED && 
	string_matches_ascii(meta_pattern, s2us("tagged"))) {

	if (meta_pattern)
	    free_string( & (meta_pattern));
	
	return(DeleteTagged(mailbox));
    }
    if (inalias) {

	for (i = 0; i < alias_message_count; i++) {
	    if (name_matches(i, meta_pattern) ||
		alias_matches(i, meta_pattern)) {
		if (! alias_selected || (alias_selected && 
					 (aliases[i]->status & VISIBLE))) {
		    if (function == UNDELETE)
			clearit(aliases[i]->status, DELETED);
		    else
			if ((function == DELETED) && (aliases[i]->type & SYSTEM)) {
			    if(i == alias_current - 1) curtag--;
			    count--;
			}
			else
			    setit(aliases[i]->status, function);
		    show_new_status(i,NULL);
		    if(i == alias_current - 1) curtag++;
		    count++;
		}
	    }
	}
    } else if (string_matches_ascii(meta_pattern, s2us("all")))
	count = DoAll(function, mailbox);
    else {
	int mc       = get_message_count(mailbox);
	int current  = get_current(mailbox);
	int selected = get_selected(mailbox);

	for (i = 0; i < mc; i++) {
	    struct header_rec *hdr = give_header(mailbox,i);
	    
	    if (hdr && (
			from_matches(hdr, meta_pattern) ||
			subject_matches(hdr, meta_pattern))) {
		if (! selected || (selected && 
				   (hdr->status & VISIBLE))) {
		    
		    if (function == UNDELETE)
			clearit(hdr->status, DELETED);
		    else
			setit(hdr->status, function);
		    show_new_status(i,mailbox);
		    if(i == current - 1) curtag++;
		    count++;
		}
	    }
	}	    
	
    }

    ClearLine(elm_LINES-2);	/* remove "pattern: " prompt */
    
    if (count > 1) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTaggedMessages,
			  "%s %d %s."), 
		  function==TAGGED? Tagged_word : 
		  function==DELETED? Mark_delete_word : Undeleted_word,
		  count, items);
    } else if (count == 1) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTaggedMessage,
			  "%s 1 %s."), 
		  function==TAGGED? Tagged_word : 
		  function==DELETED? Mark_delete_word : Undeleted_word, item);
	} else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmNoMatchesNoTags,
			      "No matches. No %s %s."), items,
		      function==TAGGED? tagged_word : 
		      function==DELETED? mark_delete_word: undeleted_word);
	}
    
    if (meta_pattern)
	free_string( & (meta_pattern));

    return(curtag);
}

int pattern_match(mailbox, redraw)
     struct MailboxView *mailbox;
     int *redraw;
{
    /** Get a pattern from the user and try to match it with the
	from/subject lines being displayed.  If matched (ignoring
	case), move current message pointer to that message, if
	not, error and return ZERO **/
    
    int i;
    
    struct string * pattern = NULL;
    struct string * alt_pattern = NULL;
    int ret = 0 ;

    int anywhere = FALSE;
    int matched;
    struct string * buffer = NULL;
    struct string *buffer1 = NULL;
    int val;

    buffer = format_string(CATGETS(elm_msg_cat, ElmSet,
				   ElmMatchPatternInEntire, 
				   "Match pattern (in entire %s):"),
			   item);

    PutLineX(elm_LINES-3,40, CATGETS(elm_msg_cat, ElmSet, ElmMatchAnywhere,
				     "/ = Match anywhere in %s."), 
	     items);
	
    buffer1 = format_string(CATGETS(elm_msg_cat, ElmSet, ElmMatchPattern,
				   "Match pattern:"));



    val = pattern_enter(&pattern, &alt_pattern, elm_LINES-1,0, 
			buffer, buffer1, redraw);

    /* 0 == pattern given
       1 == alt_pattern given
       -1 == EOF ...
 
    */

    if (val < 0) {
	ret = EOF;
	goto out;
    }

    if (1 == val) {
	if (! alt_pattern ||
	    string_len(alt_pattern) == 0) {
	    
	    ret = 1;
	    goto out;
	  }

	  anywhere = TRUE;
	  pattern = alt_pattern;
	  alt_pattern = NULL;
    }

    if (0 == val) {
	if (! pattern ||
	    string_len(pattern) == 0) {
	    ret = 0;
	    goto out;
	}
    }

    if (pattern) {
	if (inalias) {
	  for (i = alias_current; i < alias_message_count; i++) {
	    matched = name_matches(i, pattern) || alias_matches(i, pattern);
	    if (! matched && anywhere) {	/* Look only if no match yet */
		matched = comment_matches(i, pattern) ||
		    address_matches(i, pattern);
	    }
	    if (matched) {
		if (!alias_selected || 
		    (alias_selected && aliases[i]->status & VISIBLE)) {
		    alias_current = ++i;
		    ret = 1;
		    goto out;
		}
	    }
	  }
	}
	else {
	  if (anywhere) {
	      ret = match_in_message(pattern, mailbox);
	  }
	  else {
	      int current = get_current(mailbox);
	      int mc = get_message_count(mailbox);
	      int selected = get_selected(mailbox);
	      
	      for (i = current; i < mc; i++) {
		  struct header_rec *hdr = give_header(mailbox,i);
		  
		  if (hdr && (from_matches(hdr, pattern) || 
			      subject_matches(hdr, pattern))) {
		      if (!selected || (selected && 
					hdr->status & VISIBLE)) {
			  current = ++i;
			  set_current(mailbox,current);
			  ret = 1;
			  goto out;
		      }
		  }
	      }
	  }
	}	
    }

 out:
	if (pattern)
	    free_string(&pattern);
        if (alt_pattern)
	    free_string(&alt_pattern);
	if (buffer)
	    free_string(&buffer);
	if (buffer1)
	    free_string(&buffer1);
		

	return(ret);
}

int from_matches(mss, pattern)
     struct header_rec *mss;
     struct string *pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the from line of the indicated message **/

    struct addr_item *p;
    int match = 0;

    if (mss) 
	for (p = mss->from;
	     p && p->fullname && p->addr;
	     p++) {

	    struct string *s = new_string2(ASCII_SET,s2us(p->addr));
	    
	    if (find_pattern_from_string(p->fullname,pattern,1))
		match++;
	    if (find_pattern_from_string(p->comment,pattern,1))
		match++;

	    if (find_pattern_from_string(s,pattern,1))
		match++;

	    free_string(&s);

	}

    return match;
}


int to_matches(mss, pattern)
     struct header_rec *mss;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the TO line of the indicated message **/

    struct addr_item *p;
    int match = 0;
    
    if (mss) 
	for (p = mss->to;
	     p && p->fullname && p->addr;
	     p++) {
	    struct string *s = new_string2(ASCII_SET,s2us(p->addr));
	    
	    if (find_pattern_from_string(p->fullname,pattern,1))
		match++;
	    if (find_pattern_from_string(p->comment,pattern,1))
		match++;

	    if (find_pattern_from_string(s,pattern,1))
		match++;

	    free_string(&s);
	}

    return match;
}

int cc_matches(mss, pattern)
     struct header_rec *mss;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the CC line of the indicated message **/
    
    struct addr_item *p;
    int match = 0;
    
    if (mss) 
	for (p = mss->cc;
	     p && p->fullname && p->addr;
	     p++) {
	    struct string *s = new_string2(ASCII_SET,s2us(p->addr));

	    if (find_pattern_from_string(p->fullname,pattern,1))
		match++;
	    if (find_pattern_from_string(p->comment,pattern,1))
		match++;

	    if (find_pattern_from_string(s,pattern,1))
		match++;

	    free_string(&s);
	}

    return match;
}

int subject_matches(mss, pattern)
     struct header_rec *mss;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the subject line of the indicated message **/
        
    int ret = 0;

    if (mss && mss->subject)
	ret = find_pattern_from_string(mss->subject,
				       pattern,1);    
    return ret;
}

static int name_matches(message_number, pattern)
     int message_number;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the proper name of the indicated alias **/
    
    int r = 0;

    struct string *s = new_string2(display_charset,
				   s2us(aliases[message_number]->name));

    if (find_pattern_from_string(s,pattern,1))
	r++;

    free_string(&s);

    return r;
}

int alias_matches(message_number, pattern)
     int message_number;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the alias name of the indicated alias **/
    
    int r = 0;
    
    struct string *s = new_string2(display_charset,
				   s2us(aliases[message_number]->alias));
    
    if (find_pattern_from_string(s,pattern,1))
	r++;
    
    free_string(&s);

    return r;
}

int comment_matches(message_number, pattern)
     int message_number;
     struct string * pattern;
{
    /** Returns true iff the pattern occurs in it's entirety
	in the comment field of the indicated alias **/

    int r = 0;
    
    struct string *s = new_string2(display_charset,
				   s2us(aliases[message_number]->comment));
    
    if (find_pattern_from_string(s,pattern,1))
	r++;
    
    free_string(&s);

    return r;

}

static int address_matches(message_number,pattern)
     int message_number;
     struct string * pattern;
{
    struct addr_item *result;
    int dummy;
    int res = 0;
        
    /** Returns true iff the pattern occurs in it's entirety
	in the fully expanded address of the indicated alias **/
    
    result = get_alias_address_l(aliases[message_number]->alias,TRUE,
				 &dummy, NULL);
    if (result != NULL) {
	struct addr_item *walk;
	for (walk = result; 
	     walk->addr && walk ->fullname && walk->comment && !res;
	     walk++) {

	    struct string *s = new_string2(ASCII_SET,s2us(walk->addr));
	    
	    if (find_pattern_from_string(walk->fullname,pattern,1))
		res = 1;
	    if (find_pattern_from_string(walk->comment,pattern,1))
		res = 1;

	    if (find_pattern_from_string(s,pattern,1))
		res = 1;
	    
	    free_string(&s);
	}
	free_addr_items(result);	
    }
      
    return(res);
}

struct search_mes {
    struct string     * pat;
    int                 found;
};

S_(mw_init_handler mw_init_search)
static void mw_init_search(hdl)
    struct walk_handler *hdl;
{
    hdl->u.search_mes = safe_malloc(sizeof (* (hdl->u.search_mes)));
    
    hdl->u.search_mes -> pat     = NULL;
    hdl->u.search_mes -> found   = 0;
}

S_(mw_free_handler mw_free_search)
static void mw_free_search(hdl)
     struct walk_handler *hdl;
{

    free(hdl->u.search_mes);
    hdl->u.search_mes = NULL;
}

S_(mw_action_handler mw_action_search)
static void mw_action_search(hdl,ptr,state_in,fp,state_out) 
     struct walk_handler *hdl; 
     mime_t *ptr;
     in_state_t *state_in; 
     FILE *fp;                    /* Same than state_in */
     out_state_t *state_out;   /* NULL */
{
    charset_t res = NULL;
    char buffer[1024];
    int l;

    if (ptr->description &&
	find_pattern_from_string(ptr->description,
				 hdl->u.search_mes -> pat,1)) {

	hdl->u.search_mes -> found = 1;
	DPRINT(Debug,3,(&Debug, "%S found from description: %S\n",
			hdl->u.search_mes -> pat, ptr->description));
    }

    mime_get_charset(&res, ptr->type_opts,NULL,
		     default_mimetext_charset /* FIXME:  No correct */);
    
    while (0 < (l =  state_getl(buffer,sizeof buffer,state_in))) {
	struct string *x =  new_string(res);

	add_streambytes_to_string(x,l,s2us(buffer));

	if (find_pattern_from_string(x,hdl->u.search_mes -> pat,1)) {
	    hdl->u.search_mes -> found = 1;

	    DPRINT(Debug,3,(&Debug, "%S found from: %S\n",
			    hdl->u.search_mes -> pat, x));

	}

	free_string(&x);
    }

}

static struct handler_type search_mech_type = {
    WALKTYPE_magic,
    mw_init_search,
    mw_free_search,
    mw_action_search
};

int match_in_message(pattern,mailbox)
     struct string * pattern;
     struct MailboxView *mailbox;
{
    /** Match a string INSIDE a message...starting at the current 
	message read each line and try to find the pattern.  As
	soon as we do, set current and leave! 
	Returns 1 if found, 0 if not
    **/
    
    char buffer[VERY_LONG_STRING];
    int  message_number, lines, line, line_len, err;
    int mc;
    int ret = 0;
    int selected = get_selected(mailbox);

    struct walk_handler * walk = malloc_walk_handler(&search_mech_type);

    message_number = get_current(mailbox)-1;
    
    walk->u.search_mes -> pat    = pattern;

    lib_transient(CATGETS(elm_msg_cat, ElmSet, ElmSearchingFolderPattern,
			  "Searching folder for pattern..."));
    
   

    /* TODO: Match should be done agaist situation where
       MIME encoding is *undone*
    */
    
    mc = get_message_count(mailbox);
    while (message_number < mc) {
	
	struct header_rec *hdr = give_header(mailbox,message_number);
	
	if (!selected || (selected && 
			  hdr->status & VISIBLE)) {
	    
	    FILE *ZZ;
	    
	    in_state_t  state_in;
	    int val = (message_number+1) / (float) mc * 100;
	    
	    if ((message_number+1) % readmsginc == 0 ||
		val % readdatapercentinc       == 0) {
		
		PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
					       ElmSearchingInMessage,
					       "Searching %d (%02d%%) "),
			 (message_number+1), val);
		
		CleartoEOLN();
		
	    }

	    /* Search first from subject and from */
	    if (from_matches(hdr, pattern) || 
		subject_matches(hdr, pattern)) {
		
		int current = message_number+1; 
		set_current(mailbox,current);
		
		clear_error();
		ret = 1;
		
		PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
					       ElmSearchingInMessageFound,
					       "Searching %d (%02d%%) FOUND"),
			 (message_number+1), val);
		
		CleartoEOLN();
		
		goto out;		
	    }
	    
	    if (!give_message_data(mailbox,message_number,
				   &hdr,&ZZ,NULL,mime_parse_routine)) {
		err = errno;
		
		DPRINT(Debug,1,(&Debug, 
				"Error: seek %ld bytes into file failed. errno %d (%s)\n",
				hdr ? hdr->offset : -1, err, 
				"match_in_message"));
	    failXX:
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmMatchSeekFailed,
				  "ELM [match] failed looking %ld bytes into file (%s)."),
			  hdr ? hdr->offset : -1, 
			  error_description(err));		
		ret = 1;	/* fake it out to avoid replacing error message */
		goto out;
	    }
	    
	    if (!hdr->mime_parsed) {
		DPRINT(Debug,1,(&Debug,
				"mime_parse_routine was not called\n"));
		mime_parse_routine(NULL,hdr,ZZ);
	    }
	    show_msg_status(message_number,mailbox);


	    /* Copy some headers from part 1 */
	    if (skip_envelope(hdr, ZZ) != -1) {
	
		int found  = 0;

		header_list_ptr all_headers = file_read_headers(ZZ,
								RHL_MARK_FOLDING);
		header_list_ptr next_hdr;
		
		for (next_hdr = all_headers; 
		     next_hdr; 
		     next_hdr = next_hdr -> next_header) {
		    CONST char * hdr_name = give_header_name(next_hdr->header_name);
		    struct string *s = new_string2(ASCII_SET,cs2us(hdr_name));
	  
		    struct string *buffer = 
			give_decoded_header(next_hdr,
					    !(hdr->status & 
					      NOHDRENCODING),
					    hdr -> header_charset);


		    if (find_pattern_from_string(buffer,pattern,1))
			found++;
		    if (find_pattern_from_string(s,pattern,1))
			found++;

		    free_string(&buffer);
		    free_string(&s);
		}

		if (all_headers)
		    delete_headers(&all_headers);

		if (found) {
		    int current = message_number+1; 
		    set_current(mailbox,current);
		    
		    clear_error();
		    ret = 1;
		    
		    PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
						   ElmSearchingInMessageFound,
						   "Searching %d (%02d%%) FOUND"),
			     (message_number+1), val);
		    
		    CleartoEOLN();
		    
		    goto out;				   
		}

	    } else
		goto failXX;
	    
	    in_state_clear(&state_in,  STATE_in_file);
	    set_in_state_file(ZZ,&state_in);
	    
	    simple_mime_walk(&(hdr->mime_rec), walk, &state_in, NULL);
	    
	    in_state_destroy(&state_in); 
	    
	    if (walk->u.search_mes ->found) {
	    
		int current = message_number+1; 
		set_current(mailbox,current);
		
		clear_error();
		ret = 1;
		
		PutLineX(elm_LINES, 0, CATGETS(elm_msg_cat, ElmSet, 
					       ElmSearchingInMessageFound,
					       "Searching %d (%02d%%) FOUND"),
			 (message_number+1), val);
		
		CleartoEOLN();
		
		goto out;
	    }	    
	}
	/** now we've passed the end of THIS message...increment and 
	    continue the search with the next message! **/
	
	message_number++;
    }

 out:
    free_walk_handler(&walk);
    return(ret);
}

/* This will tag, delete, or undelete all or the currently visible messages */
static int DoAll(func, mailbox)
     int func;
     struct MailboxView *mailbox;
{
    int i, count = 0;
    int mc = get_message_count(mailbox);
    
    for (i=0; i < mc; i++) {
	struct header_rec * hdr = give_header(mailbox,i);
	
	if (hdr) {
	    int selected = get_selected(mailbox);

	    if (!selected ||
		(selected && ison(hdr->status,VISIBLE))) {
		if (func == TAGGED)
		    hdr->status |= TAGGED;
		else if (func == DELETED)
		    hdr->status |= DELETED;
		else if (func == UNDELETE)
		    hdr->status &= ~DELETED;
		show_new_status(i,mailbox);
		++count;
	    }
	}
    }
    return(count);
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
