static char rcsid[] = "@(#)$Id: showmsg.c,v 1.27 2003/07/29 18:46:21 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.27 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This file contains all the routines needed to display the specified
    message.
**/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"ui");

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif       

VOLATILE int pipe_abort = FALSE; /* not used anymore, but won't compile without it */


/* Returns 0 on quit */

struct prompt_list {
    struct mimeinfo *structure;
    char   * printable_command;
    int    selected;
    int    col;
};

static void update_prompt P_((struct prompt_list  ** list, int  *list_len,
			      int *current_offset, int idx));

static void update_prompt (list,list_len,current_offset,idx)
     struct prompt_list  ** list; 
     int  *list_len;
     int *current_offset;
     int idx;
{

    if (idx < 0) {
	int i;

	*current_offset = 0;
	ClearScreen();

	print_format_center(1,
			    CATGETS(elm_msg_cat, ElmSet, ElmMailcapMenu,
				    "Mailcap program selection"));

	for (i = *current_offset; 
	     i >= 0 && 
		 i < *current_offset - 2 + elm_LINES / 3  &&
		 i < *list_len;
	     i++)
	    update_prompt(list,list_len,current_offset,i);
		 
    } else if (idx >= 0 &&
	       idx >= *current_offset &&
	       idx < *current_offset - 2 + elm_LINES / 3 &&
	       idx < *list_len
	       ) {
	
	int X = (idx - *current_offset) * 3 + 3;
	media_type_t            T =  (*list)[idx].structure->TYPE;

	struct string * S = 
	    format_string(CATGETS(elm_msg_cat, ElmSet, ElmUseMailcap,
			 "%2d Use '%s' to show %s/%s"),
			  idx,
			  (*list)[idx].printable_command ? 
			  (*list)[idx].printable_command : "???",
			  get_major_type_name(T), get_subtype_name(T));
	int pos = 0;
	int visible_len = 0;
	struct string * S1 = curses_printable_clip(S,&pos,string_len(S),
						   &visible_len,
						   elm_COLUMNS-10);

	PutLineX(X,0,FRM("%S ? "),S1);
	CleartoEOLN();
		 
	(*list)[idx].col = visible_len+3;
	
	if ((*list)[idx].structure->handler_data->use_entry ||
	    (*list)[idx].selected) {
	    
	    if ((*list)[idx].structure->handler_data->use_entry)
		PutLineX(X,(*list)[idx].col,
			 CATGETS(elm_msg_cat, ElmSet, ElmYesWord, 
				 "Yes."));
	    else
		PutLineX(X,(*list)[idx].col,
			 CATGETS(elm_msg_cat, ElmSet, ElmNoWord, 
				 "No."));
	}
	    
	free_string(&S1);
	free_string(&S);
    }

}

static void prompt_programs_1 P_((struct mimeinfo *structure,
				 struct prompt_list  ** list, 
				 int  *list_len,
				 int *current_offset));

static void prompt_programs_1(structure, list, list_len, current_offset)
     struct mimeinfo *structure;
     struct prompt_list  ** list; 
     int  *list_len;
     int *current_offset;
{

    if (structure->mime_flags) {
       
	if (structure->handler_data &&
	    structure->handler_data->entry) {
	    
	    char * cmd = mailcap_view_command(structure->handler_data->entry,
					      structure);

	    if (cmd) {
		int i;
		char *x;

		*list = safe_realloc(*list,
				     sizeof ((*list)[0]) * (*list_len +1));
		
		bzero((void *) &( (*list)[*list_len] ),
		      sizeof    ( (*list)[*list_len] ));
		
		if (strlen(cmd) > 15 && (x = strrchr(cmd,'/')) && x != cmd) {

		    char * s = elm_message(FRM("...%s"),x);

		    free(cmd);
		    cmd = s;
		}

		(*list)[*list_len].structure          = structure;
		(*list)[*list_len].printable_command  = cmd;
		(*list)[*list_len].selected           = 0;
		(*list)[*list_len].col                = elm_COLUMNS-7;
		
		i = (*list_len)++;

		update_prompt(list,list_len,current_offset,i);
	    }
	}
	
	/* Is alternative -- just process it */
	if (structure->handler_data &&
	    structure->handler_data->selected_alternative) {

	    if (structure->handler_data->selected_alternative->magic != MIME_magic)
		mime_panic(__FILE__,__LINE__,"prompt_programs_1",
			   "Bad magic number (alternatice)");

	    prompt_programs_1(structure->handler_data->selected_alternative,
			      list,list_len,current_offset);

	} 

	/* Process subparts recursively */
	else if (structure->parser_data) {
	    int count = mime_parser_subparts(structure->parser_data);
	    int i;

	    for (i = 0; i < count; i++) {
		mime_t *att = 	mime_parser_index(structure->parser_data,i);

		if (att->magic != MIME_magic)
		    mime_panic(__FILE__,__LINE__,"prompt_programs_1",
			       "Bad magic number (subpart)");

		prompt_programs_1(att,
				  list,list_len,current_offset);
	    
	    }

	}
	
    }
}

/* return TRUE if metamail is used instead */
static int prompt_programs P_((struct mimeinfo *structure));
static int prompt_programs(structure)
     struct mimeinfo *structure;
{
    int r = TRUE;

    struct prompt_list  * list     = NULL;
    int                   list_len = 0;
    int                   offset   = 0;

    update_prompt(&list,&list_len,&offset,-1);
    
    prompt_programs_1(structure,&list,&list_len,&offset);

    if (list_len > 0) {
	int q = 0;
	int ch;


	r = FALSE;
	PutLineX(elm_LINES-1,1,
		 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
			 "Answer questions, use m)etamail or q)uit menu: "));

	do {
	    int defans = *def_ans_no;


	    if (q >= list_len) {

		PutLineX(elm_LINES-1,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		defans = 'q';

	    } else {	       		
		int X = (q - offset) * 3 + 3;

		if (X <= 1) {
		    offset = q-1;
		    if (offset < 0)
			offset = 0;
		    ch = REDRAW_MARK;
		    goto redraw;
		}

		if (X >= elm_LINES-1) {
		    offset = q;
		    ch = REDRAW_MARK;
		    goto redraw;
		}

		MoveCursor(X,list[q].col);
	    }

	    
	    CleartoEOLN();   /* Clear current answer */
	    
	    Write_to_screen(FRM("%c%c"),defans,BACKSPACE);
	    FlushBuffer();


	    ch = ReadCh(REDRAW_MARK|READCH_CURSOR);
	    
	    if (ch == '\n') 
		ch = defans;
	    
	    switch (ch) {
	    case REDRAW_MARK:
	    redraw:
		update_prompt(&list,&list_len,&offset,-1);
		
		PutLineX(elm_LINES-1,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		break;
		
	    case '-':
	    case LEFT_MARK:
	    case PAGEUP_MARK:
		offset -= elm_LINES / 3 - 2;
		if (offset < 0)
		    offset = 0;
		q = offset;
		goto redraw;

	    case '+':
	    case RIGHT_MARK:
	    case PAGEDOWN_MARK:

		offset += elm_LINES / 3 - 2;
		if (offset >= list_len);
		    offset = list_len-1;
		q = offset;
		goto redraw;

	    case DOWN_MARK:
		if (q < list_len) {
		    update_prompt(&list,&list_len,&offset,q);
		    q++;
		}
		break;

	    case UP_MARK:
		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q);
		if (q > 0)
		    q--;
		break;

	    case 'q':
		ch = 0;

		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q);

		PutLineX(elm_LINES-1,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));

		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmQuit,
					"Quit"));

		break;

	    case 'm':

		if (q < list_len) 
		    update_prompt(&list,&list_len,&offset,q);

		PutLineX(elm_LINES-1,1,
			 CATGETS(elm_msg_cat, ElmSet, ElmMailcapQuit,
				 "Answer questions, use m)etamail or q)uit menu: "));
		
		Write_to_screen(CATGETS(elm_msg_cat, ElmSet,
					ElmMetamail,
					"Metamail"));

		if (have_metamail()) {
		    ch = 0;
		    r = TRUE;
		}

		break;
	    }

	    if (q < list_len) {
		
		if (ch == *def_ans_yes) {
		    list[q].selected = 1;
		    list[q].structure->handler_data->use_entry = 1;
		    update_prompt(&list,&list_len,&offset,q);
		    q++;
		} else if (ch == *def_ans_no) {
		    
		    list[q].selected = 1;
		    list[q].structure->handler_data->use_entry = 0;
		    update_prompt(&list,&list_len,&offset,q);
		    q++;

		}
	    } 

	    FlushBuffer();
		
	} while(ch);
    }

 failure:

    if (list) {
	int i;

	for (i = 0; i < list_len; i++) {
	    if (list[i].printable_command) {
		free(list[i].printable_command);
		list[i].printable_command = NULL;
	    }
	}

	free(list);
	list = NULL;
    }
    list_len = 0;

    return r;
}


int need_meta (hdr)
     struct header_rec *hdr;
{
    /* Determine whether or not we need to call metamail to display the
     * message contents.
     */
    int result = 0;

    if ((hdr->status & (MIME_MESSAGE)) && (hdr->status & MIME_UNSUPPORTED))
	result = have_metamail();

    /* Do not call metamail if Disposition is not inline */
    else if ((hdr->status & MIME_MESSAGE) && 
	     hdr->mime_rec.disposition != DISP_INLINE)
	result = FALSE;  
    else if (hdr->status & (MIME_MESSAGE | PRE_MIME_CONTENT)) {

	int X = hdr->mime_rec.mime_flags;  /* NOTPLAIN_need_metamail           0x01
					      NOTPLAIN_need_mailcap            0x02
					      NOTPLAIN_canuse_mailcap          0x04
					   */
		
	if (0 != (X & NOTPLAIN_need_metamail) && have_metamail())
	    result = TRUE;
	else if (0 != (X & NOTPLAIN_canuse_mailcap) ||
		 0 != (X & NOTPLAIN_need_mailcap)) {

	    result = prompt_programs (& (hdr->mime_rec));

	}

    }

    DPRINT(Debug,9,(&Debug, "need_meta=%d\n",result));
    return result;
}

int show_msg(current_header, infile, current,message_count)
     struct header_rec *current_header;
     FILE *infile;
     int current, message_count;
{
	/*** Display number'th message.  Get starting and ending lines
	     of message from headers data structure, then fly through
	     the file, displaying only those lines that are between the
	     two!

	     Return 0 to return to the index screen or a character entered
	     by the user to initiate a command without returning to
	     the index screen (to be processed via process_showmsg_cmd()).
	***/



	int is_need_meta = 0;

	DPRINT(Debug,4,(&Debug, 
			"displaying %d lines from message #%d using %s\n", 
			current_header->lines, 
			current, pager));


	if(ison(current_header->status, NEW)) {
	    clearit(current_header->status, NEW);   /* it's been read now! */
	    current_header->status_chgd = TRUE;
	}
	if(ison(current_header->status, UNREAD)) {
	  clearit(current_header->status, UNREAD);   /* it's been read now! */
	  current_header->status_chgd = TRUE;
	}



	is_need_meta = need_meta(current_header);

        if (is_need_meta) {
	        char fname[STRING], Cmd[SLEN];
		int err;
		FILE *fpout;

		elm_sfprintf(fname, sizeof fname,
			     FRM("%semm.%d.%d"), temp_dir, getpid(), getuid());

		if ((fpout = safeopen_rdwr(fname)) == NULL) {
		    err = errno;
		    DPRINT(Debug,1,(&Debug, 
				    "Error: open of temporary file %s, errno %s (show_message)\n",
				    fname, error_description(err)));
		  lib_error(CATGETS(elm_msg_cat, ElmSet, 
				    ElmCantOpenAsOutputFile,
				    "Can't open \"%s\" as output file! (%s)."),
			    fname, error_description(err));
		  return(0);
		}
		/* Let metamail decode it! 
		 * (Now CM_DECODE also decodes MIME and PGP) -KEH
		 *
		 * CM_LF: Metamail can not cope with CRLF
		 */
		if (!copy_message_f(infile,current_header,
				    "", fpout, CM_LF, NULL)) {
		    /* FAIL */
		    fclose (fpout);
		    unlink (fname);
		    return 0;
		}
		(void) fclose (fpout);
		elm_sfprintf(Cmd, sizeof Cmd,
			     FRM("%s -p -z -m Elm %s"), 
			     metamail_path, fname);
		Raw(OFF);
		ClearScreen();
		printf ("Executing metamail...\n");
		system_call(Cmd, SY_ENAB_SIGINT|SY_ENV_METAMAIL, NULL);
		Raw(ON | NO_TITE);	/* Raw on but don't switch screen */
		(void) unlink (fname);
		PutLineX(elm_LINES,0, 
			 CATGETS(elm_msg_cat, ElmSet, 
				 ElmPressAnyKeyIndex,
				 "Press any key to return to index."));
		(void) ReadCh(0);
		printf("\r\n");
		Raw(OFF | NO_TITE); /* Raw off so raw on takes effect */
		Raw(ON); /* Finally raw on and switch screen */
		return(0);
	} else {
	    
	    return (metapager (infile, current_header, TRUE, 
			       current, message_count));
	}
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */

