static char rcsid[] = "@(#)$Id: elmlibregister.c,v 1.2 2003/05/25 11:19:57 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.2 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************
 * Some code based on ../src/save_optc.c
 * That code was following copyright:
 *
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#include "elmutil.h"
#include "s_me.h"
#include "s_elm.h"
#include "s_elmrc.h"

#include "rc_imp.h"
#include "save_opts.h"

#include "reghelper.h"

DEBUG_VAR(Debug,__FILE__,"util");

extern char *optarg;		
extern int   optind;		

int    register_fd     = -1; 
char * program_name    = "elmrc-write"; 

int main(argc, argv)
     int argc;
     char *argv[];
{
    int err = 0;
    int global = 0;
    int c;
    int write_default = 0;
    char * targetfile = NULL;
    FILE * commentfile = NULL;

    int x;

#if DEBUG
    init_debugfile("ELMLIBREGISTER");
#endif
    locale_init();

    REGHELPER_INIT(argv[0]);

    while ((c = getopt(argc, argv, "GId:w:C:")) != EOF) {
	
	switch(c) {
	case 'G':
	    global++;
	    break;

	case 'I':
	    write_default++;
	    break;
	    
	case 'd':
#if DEBUG
	    set_debugging(optarg);	 
#else
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmArgsIngoringDebug,
			      "Warning: system created without debugging enabled - request ignored\n"));
#endif
	    break;
	    
	case 'w' : 
	    targetfile = optarg;
	    if (0 != access(targetfile,WRITE_ACCESS)) {
		int errcode = errno;
		
		if (errcode != ENOENT) {
		    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
				      "File %.50s is not writeable: %s"),
			      targetfile, error_description(errcode));
		    err++;
		    goto fail;
		}
	    }
	    break;

	case 'C':
	    if (0 != access(optarg,READ_ACCESS)) {
		int errcode = errno;
		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotReadable,
				  "File %.50s is not readable: %s"),
			  optarg, error_description(errcode));
		err++;
		goto fail;
	    }

	    commentfile = fopen(optarg,"r");
	    if (!commentfile) {
		int errcode = errno;
		lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotReadable,
				  "File %.50s is not readable: %s"),
			  optarg, error_description(errcode));
		err++;
		goto fail;
		
	    }
	    break;

	case '?':
	    err = 1;
	    goto fail;
	}

    }

    user_init();
    init_defaults();

    elm_sfprintf(version_buff, sizeof version_buff,
		 FRM("%s PL%s"), VERSION, PATCHLEVEL);


#ifdef DEBUG
    { 
	int d = panic_dprint("\n\
======================================================\n\
Debug output of the ELMLIBREGISTER program (version %s).\n",
			     version_buff);

	if (d >= 50) {
#if 0	
	    panic_dprint("WARNING: Edit manually out sensitive information from that file!\n");
	    
	    lower_prompt("WARNING: Debug file may include passwords -- edit it!");
	    error_sleep(5+sleepmsg);	    
#endif
	}
    }
#endif

    if (!global)
	read_rc_file(0);

    
    if (optind < argc) {
	test_and_set_shared(& argv[optind] , targetfile || write_default);
    } else {
	/* NO libraries .. only check current .... */
	test_and_set_shared( NULL , targetfile || write_default);
    }

    for (x = 0; x < NUMBER_OF_SAVEABLE_OPTIONS; x++) {
	if (0 == strcmp(save_info[x].name,"use-library"))
	    save_info[x].flags |= FL_CHANGED | FL_LOCAL;
    }

    if (!commentfile) {
	commentfile = fopen(ELMRC_INFO,"r");
	if (!commentfile)
	    lib_error(CATGETS(elm_msg_cat, ElmrcSet, 
			      ElmrcSavingWithoutComments,
			      "Warning: saving without comments! Can't get to %s."),
		      ELMRC_INFO);
    }

    if (targetfile && write_default) {
	lib_error(CATGETS(elm_msg_cat, ElmrcSet, ElmrcNoBothwI,
			  "Can't specify both -w and -I options!"));
	err++;
	goto fail;

    }
    if (write_default)
	targetfile = global ? system_rc_file : user_rc_file;


    if (targetfile) {
	char * tmp = elm_message(FRM("%s.N"),targetfile);
	int errcode = can_open(tmp,"w");
	int x;
	FILE *f;

		
	if (errcode) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			      "File %.50s is not writeable: %s"),
		      tmp, error_description(errcode));
	    
	    err++;
	    free(tmp);
	    goto fail;
	}
	f = fopen(tmp,"w");
	if (!f) {
	    int errcode = errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			      "File %.50s is not writeable: %s"),
		      tmp, error_description(errcode));
	    
	    err++;
	    free(tmp);
	    goto fail;
	}

	elm_fprintf(f, 
		    CATGETS(elm_msg_cat, ElmrcSet, ElmrcSavedAutoLIB,
			    "# Saved automatically by ELMLIBREGISTER %s\n#\n\n"), 
		    version_buff);


	if (commentfile) {
	    char line_buffer[SLEN];
	    int x;
	    int len;
	    int local_value;
	    int skip = 1;

	    while (0  < (len = 
			 mail_gets(line_buffer, sizeof line_buffer, 
				   commentfile))) {

		int y = 0;


		if (line_buffer[len-1] != '\n') {
		    DPRINT(Debug,1,(&Debug, 
				    "Too long line (read len=%d): %s\n",
				    len,line_buffer));
		    break;
		}
		line_buffer[--len] = '\0';
		if (0 == len)
		    continue;

		if ('#' == line_buffer[0] || whitespace(line_buffer[0])) {

		    if (strncmp(line_buffer, "#$HDR", 5) == 0) {
			skip = 0;
			continue;
		    }

		    if (skip)
			continue;

		    /* Copy initial comments */

		    fprintf(f,"%s\n",line_buffer);
		    continue;
		}
		skip = 1;


		for (x = 0; x < NUMBER_OF_SAVEABLE_OPTIONS; x++) {
		    y = strcmp(line_buffer, save_info[x].name);
		    if (y <= 0)
			break;
		}

		if (x >= NUMBER_OF_SAVEABLE_OPTIONS || y != 0) {

		    DPRINT(Debug,15,(&Debug,
				     "Skipping option %s ... \n",
				     line_buffer));
		    
		    continue;
		}

		if (global && 
		    ! (save_info[x].flags & FL_CHANGED) &&
		    ! (save_info[x].flags & FL_SYS)) {
		    DPRINT(Debug,15,(&Debug,
				     "Skipping user option %s ... \n",
				     line_buffer));
		    
		    continue;
		}

		if (!global &&
		    (save_info[x].flags & FL_SYS)) {
		    DPRINT(Debug,15,(&Debug,
				     "Skipping system option %s ... \n",
				     line_buffer));
		    
		    continue;
		}

		fprintf(f,"\n");

		/* Copy comments */
		while (0  < (len = 
			     mail_gets(line_buffer, sizeof line_buffer, 
				       commentfile))) {
		    if ('#' != line_buffer[0])
			break;
		    fputs(line_buffer,f);
		}

		if (global)
		    local_value = save_info[x].flags & FL_CHANGED;
		else
		    local_value = save_info[x].flags & FL_LOCAL;
		if (!valid_rc_type(save_info[x].dt_type))
		    panic("RC PANIC",__FILE__,__LINE__,"main",
			  "Bad config item type",0);

		save_info[x].dt_type->print_value(f,
						  & save_info[x],
						  !local_value);


		save_info[x].flags |= FL_SAVED;
	    }
	}

	for (x = 0; x < NUMBER_OF_SAVEABLE_OPTIONS; x++) {
	    int local_value;
		

	    if (save_info[x].flags & FL_SAVED)
		continue;

	    if (global && 
		! (save_info[x].flags & FL_CHANGED) &&
		! (save_info[x].flags & FL_SYS)) {
		DPRINT(Debug,15,(&Debug,
				 "Skipping user option %s ... \n",
				 save_info[x].name));
		
		continue;
	    }

	    if (!global &&
		(save_info[x].flags & FL_SYS)) {
		DPRINT(Debug,15,(&Debug,
				 "Skipping system option %s ... \n",
				 save_info[x].name));
		
		continue;
	    }

	    fprintf(f,"\n");

	    if (global)
		local_value = save_info[x].flags & FL_CHANGED;
	    else
		local_value = save_info[x].flags & FL_LOCAL;
	    if (!valid_rc_type(save_info[x].dt_type))
		panic("RC PANIC",__FILE__,__LINE__,"main",
		      "Bad config item type",0);
	    
	    save_info[x].dt_type->print_value(f,
					      & save_info[x],
					      !local_value);
	}

	print_local_shared_options(f,global);

	if (EOF == fclose(f)) {
	    int errcode = errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			      "File %.50s is not writeable: %s"),
		      tmp, error_description(errcode));
	    
	    err++;
	    free(tmp);
	    goto fail;
	}
	if (0 != rename(tmp,targetfile)) {
	    int errcode = errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotRenamed,
			      "Failed to rename temporary file to %.50s: %.30s"),
		      targetfile, error_description(errcode));
	    
	    err++;
	    free(tmp);
	    goto fail;
	}

	log_config(targetfile);

	free(tmp);
    }

 fail:
    if (commentfile)
	fclose(commentfile);

    if (err)
	lib_error(CATGETS(elm_msg_cat, MeSet, MeProgFailed,
			  "%s failed; exit code=%d"),
		  argv[0],err);

    return err;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
